﻿using System;
using System.Net;
using System.Collections.Generic;
using System.Reflection;
using System.Data.SqlClient;
using System.Data;
using System.Collections;
using System.IO;
using System.Diagnostics;
using System.Text.RegularExpressions;

namespace IntelliSite.CommonUtil
{
	public enum SQLComparisonOperator : byte
	{
        None = 0,
		Is,
		Not,
		LessThan,
		LessEqual,
		GreaterThan,
		GreaterEqual,
	}

    public enum SQLLogicalOperator : byte
    {
        None = 0,
        And,
        Or,
    }

    public struct WhereClauseEntry
    {
        public string                   fieldName;
        public object                   fieldValue;
        public SQLComparisonOperator    sqlOperator;
        public SQLLogicalOperator       sqlLogOperator;

        public WhereClauseEntry( string fieldNameVal, object fieldValueVal, SQLComparisonOperator sqlOperatorVal, SQLLogicalOperator sqlLogOperatorVal )
        {
            fieldName       = fieldNameVal;
            fieldValue      = fieldValueVal;
            sqlOperator     = sqlOperatorVal;
            sqlLogOperator  = sqlLogOperatorVal;
        }        
    };

    public class CustomDB
    {
        static private HashSet<string>  autoIncrementFields = new HashSet<string>();
   
        private SqlConnection       connection;
        private SqlDataAdapter      adapter;
        private DataSet             dataSet;
        private DataTable           table;        
        private Type                objectType;

        private ArrayList           fieldNames;
        private string              tableName;        
        private ArrayList           whereClauseEntries;
        private string              fieldToFilterOutDefaultRec;

        internal ArrayList          objects;
        public   ArrayList          Objects             { get { return objects; }  }

		internal static LoggerBase	logger;
		public static LoggerBase	Logger				{ get { return logger; } set { logger = value; } }

        internal static string      dbConnectionString;
        public static string        ConnectionString    { get { return dbConnectionString; } set { dbConnectionString = value; } }

        internal string             whereClause;
        public string               WhereClause			{ get { return whereClause; } set { whereClause = value; } }

        internal string             fromClause;
        public string               FromClause			{ get { return fromClause; } set { fromClause = value; } }

        internal string             sortClause;
        public string               SortClause			{ get { return sortClause; } set { sortClause = value; } }

        internal string             groupByClause;
        public string               GroupByClause	    { get { return groupByClause; } set { groupByClause = value; } }

        internal static string      lastErrorText;
        public string               LastErrorText       { get { return lastErrorText; } set { lastErrorText = value; } }

		public static void OutputDebug( string output )
		{
			if( logger != null )
				logger.Write( output );
			else
				Debug.WriteLine( output );
		}

        #region Destructor/Constructor/Close/Cleanup
        ~CustomDB()
        {
            Cleanup();

            fieldNames          = null;
            whereClauseEntries  = null;
            objects             = null;
        }

        /// <summary>        
        /// Constructor initializes the database table and table fields to be retrieved.
        /// </summary>
        /// 
        /// <param name="tableName">The Database table name(s).</param>
        /// <param name="objectType">The type from who's data members are retrieved and used as field names to retrieve from the database.</param>
        /// 
        /// <returns></returns>
        /// 
        public CustomDB( string tableName, Type objectType, string fieldToFilterOutDefaultRec = "" )
        {
            if( dbConnectionString.Length == 0 )
                throw new ArgumentException("Connection string not set.");

            if( autoIncrementFields.Count == 0 )
                LoadDatabaseSchema( dbConnectionString );

            Cleanup();

            this.objectType = objectType;
            this.tableName  = tableName;

            whereClauseEntries  = new ArrayList();
            fieldNames          = new ArrayList();
            objects             = new ArrayList();

            string[] tableNamesArray = tableName.Split(',');

            if( tableNamesArray.GetLength(0) == 1 )
                fromClause = string.Format("FROM {0} ", tableNamesArray[0]);
            else if( tableNamesArray.GetLength(0) == 2 )
                fromClause = string.Format("FROM {0} INNER JOIN {1} ", tableNamesArray[0], tableNamesArray[1]);
            else
                throw new ArgumentException("Invalid table name(s) specified.  If wanting to join more than 2 tables, modify the 'FromClause' and 'WhereClause' properties directly.");

            fieldNames = new ArrayList();
            PropertyInfo[] propertyInfos;
            propertyInfos = objectType.GetProperties( BindingFlags.Instance | BindingFlags.Public | BindingFlags.GetProperty | BindingFlags.SetProperty );

            foreach( PropertyInfo propertyInfo in propertyInfos )
                fieldNames.Add( propertyInfo.Name );

            this.fieldToFilterOutDefaultRec = fieldToFilterOutDefaultRec;
        }

        /// <summary>        
        /// Close Database connection and clean up.
        /// </summary>
        ///
        public void Close()
        {
            Cleanup();

            fieldNames.Clear();
            whereClauseEntries.Clear();
            objects.Clear();
        }

        private void Cleanup()
        {   
            connection          = null;
            adapter             = null;
            dataSet             = null;
            table               = null;
            objectType          = null;
            tableName           = "";
            fromClause          = "";
            sortClause          = "";
            whereClause         = "";
            groupByClause       = "";
        }
        #endregion

        #region Join
        /// <summary>        
        /// Specify two fields to join.  The corresponding table names will have already been specified in the constructor's 'tableName' parameter.
        /// </summary>
        /// 
        /// <param name="joinFields">The two field names to join.</param>
        /// 
        public void Join( string joinFields )
        {
            string[] tableNamesArray = tableName.Split(',');
            string[] joinFieldsArray = joinFields.Split(',');

            if( tableNamesArray.GetLength(0) != 2 || joinFieldsArray.GetLength(0) != 2 )
                throw new ArgumentException("Invalid field names specified.  If wanting to join more than 2 tables, modify the 'FromClause' and 'WhereClause' properties directly.");
            
            fromClause += string.Format("ON {0}.{1} = {2}.{3} ", tableNamesArray[0], joinFieldsArray[0], tableNamesArray[1], joinFieldsArray[1]);
        }
        #endregion

        #region Sort
        /// <summary>        
        /// Constructor initializes the database table and table fields to be retrieved.
        /// </summary>
        /// 
        /// <param name="sortField">Set the sort order.</param>
        /// <param name="ascending">Specifies the sort order (true is ascending, false is descending).</param>
        /// 
        public void Sort( string sortField, bool ascending )
        {
            if( sortClause.Length == 0 )
                sortClause = "ORDER BY ";
            else
                sortClause += ", ";

            sortClause += string.Format("{0} {1} ", sortField, ascending ? "ASC" : "DESC");
        }
        #endregion

        #region Get
        /// <summary>        
        /// Get the data from the database and fill an array of objects.
        /// </summary>
        /// 
        /// <returns>true if the data fetch succeeded.</returns>
        public bool Get()
        {
            return Get( "", null, SQLComparisonOperator.None );
        }

        /// <summary>        
        /// Get the data from the database and fill an array of objects.
        /// </summary>
        /// 
        /// <param name="fields">Only fetch the fields from the database specified in this list.  This overrides the list of fields generated from the Type argument in the constuctor.</param>
        /// 
        /// <returns>true if the data fetch succeeded.</returns>
        public bool GetWithFields( params string[] fields )
        {
            SetFieldsToInclude( fields );
            return Get( "", null, SQLComparisonOperator.None );
        }

        /// <summary>        
        /// Get the data from the database and fill an array of objects.
        /// </summary>
        /// 
        /// <param name="fields">Only fetch the fields from the database NOT specified in this list.  This overrides the list of fields generated from the Type argument in the constuctor.</param>
        /// 
        /// <returns>true if the data fetch succeeded.</returns>
        public bool GetWithoutFields( params string[] fields )
        {
            SetFieldsToExclude( fields );
            return Get( "", null, SQLComparisonOperator.None );
        }

        /// <summary>        
        /// Get the data from the database and fill an array of objects.
        /// </summary>
        /// 
        /// <param name="whereClauseField">The database field to use in the where clause for this query.</param>
        /// <param name="whereClauseFieldValue">The value to be used in conjuction with the where clause field</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>
        /// 
        /// <returns>true if the data fetch succeeded.</returns>
        public bool Get( string whereClauseField, object whereClauseFieldValue, SQLComparisonOperator sqlOperator )
        {
            bool rc = false;

            try
            {
                objects.Clear();

                table = Open(null, whereClauseField, whereClauseFieldValue, sqlOperator );

				if( table.Rows.Count == 0 )
					return false;
                
                foreach( DataRow row in table.Rows )
                {
                    object newObject = objectType.Assembly.CreateInstance(objectType.FullName);
                    foreach( string fieldName in fieldNames )
                    {
                        if( !row[fieldName].Equals(DBNull.Value) )
                            newObject.GetType().GetProperty(fieldName).GetSetMethod().Invoke( newObject, new object[]{ row[fieldName] } );
                    }

                    objects.Add( newObject );
                }

                rc = true;
            }
            catch( ArgumentException ae )
            {
                lastErrorText = ae.ToString();
				OutputDebug( "CustomDB Get Error (" + tableName + ") " + lastErrorText );
            }
            catch( SqlException se)
            {
                lastErrorText = se.ToString();
				OutputDebug( "CustomDB Get Error (" + tableName + ") " + lastErrorText );
            }

            return rc;
        }

        /// <summary>        
        /// Get the data from the database and fill an array of objects.
        /// </summary>
        /// <param name="whereClauseField">The database field to use in the where clause for this query.</param>
        /// <param name="whereClauseFieldValue">The value to be used in conjuction with the where clause field</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>
        /// <param name="fields">Only fetch the fields from the database specified in this list.  This overrides the list of fields generated from the Type argument in the constuctor.</param>
        /// 
        /// <returns>true if the data fetch succeeded.</returns>
        public bool GetWithFields( string whereClauseField, object whereClauseFieldValue, SQLComparisonOperator sqlOperator, params string[] fields )
        {
            SetFieldsToInclude( fields );
            return Get( whereClauseField, whereClauseFieldValue, sqlOperator );
        }

        /// <summary>        
        /// Get the data from the database and fill an array of objects.
        /// </summary>
        /// <param name="whereClauseField">The database field to use in the where clause for this query.</param>
        /// <param name="whereClauseFieldValue">The value to be used in conjuction with the where clause field</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>
        /// <param name="fields">Only fetch the fields from the database NOT specified in this list.  This overrides the list of fields generated from the Type argument in the constuctor.</param>
        /// 
        /// <returns>true if the data fetch succeeded.</returns>
        public bool GetWithoutFields( string whereClauseField, object whereClauseFieldValue, SQLComparisonOperator sqlOperator, params string[] fields )
        {
            SetFieldsToExclude( fields );
            return Get( whereClauseField, whereClauseFieldValue, sqlOperator );
        }

        /// <summary>        
        /// Get the data from the database and fill an array of objects.
        /// </summary>
        /// <param name="whereClauseField">The database field to use in the where clause for this query.</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>
        /// <param name="dataObject">This object has a data member that matches the 'whereClauseField'.  Get the value for the where clause from the dataObject's current value for this member.</param>
        /// 
        /// <returns>true if the data fetch succeeded.</returns>
        public object Get( string whereClauseField, SQLComparisonOperator sqlOperator, object dataObject )
        {
            if( dataObject.GetType() != objectType )
            {
                lastErrorText = "Data type of object supplied does not match the type used for this instance of customDB";
				OutputDebug( "CustomDB Get Error (" + tableName + ") " + lastErrorText );
                return null;
            }

            object newObject = null;

            try
            {
                objects.Clear();

                table = Open(dataObject, whereClauseField, null, sqlOperator);

                if( table.Rows.Count == 1 )
                {
                    newObject = objectType.Assembly.CreateInstance(objectType.FullName);

                    DataRow row = table.Rows[0];
                    foreach( string fieldName in fieldNames )
                    {
                        if( !row[fieldName].Equals(DBNull.Value) )
                            objectType.GetProperty(fieldName).GetSetMethod().Invoke( newObject, new object[]{ row[fieldName] } );
                    }
                }
            }
            catch( ArgumentException ae )
            {
                lastErrorText = ae.ToString();
				OutputDebug( "CustomDB Get Error (" + tableName + ") " + lastErrorText );
            }
            catch( SqlException se)
            {
                lastErrorText = se.ToString();
				OutputDebug( "CustomDB Get Error (" + tableName + ") " + lastErrorText );
            }

            return newObject;
        }

        /// <summary>        
        /// Get the data from the database and fill an array of objects.
        /// </summary>
        /// <param name="whereClauseField">The database field to use in the where clause for this query.</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>
        /// <param name="dataObject">This object has a data member that matches the 'whereClauseField'.  Get the value for the where clause from the dataObject's current value for this member.</param>
        /// <param name="fields">Only fetch the fields from the database specified in this list.  This overrides the list of fields generated from the Type argument in the constuctor.</param>
        /// 
        /// <returns>true if the data fetch succeeded.</returns>
        public object GetWithFields( string whereClauseField, SQLComparisonOperator sqlOperator, object dataObject, params string[] fields )
        {
            SetFieldsToInclude( fields );
            return Get( whereClauseField, sqlOperator, dataObject );
        }

        /// <summary>        
        /// Get the data from the database and fill an array of objects.
        /// </summary>
        /// <param name="whereClauseField">The database field to use in the where clause for this query.</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>
        /// <param name="dataObject">This object has a data member that matches the 'whereClauseField'.  Get the value for the where clause from the dataObject's current value for this member.</param>
        /// <param name="fields">Only fetch the fields from the database NOT specified in this list.  This overrides the list of fields generated from the Type argument in the constuctor.</param>
        /// 
        /// <returns>true if the data fetch succeeded.</returns>
        public object GetWithoutFields( string whereClauseField, SQLComparisonOperator sqlOperator, object dataObject, params string[] fields )
        {
            SetFieldsToExclude( fields );
            return Get( whereClauseField, sqlOperator, dataObject );
        }

        private DataTable Open( object dataObject, string whereClauseField, object whereClauseFieldValue, SQLComparisonOperator sqlOperator )
        {
            table = null;

            connection = new SqlConnection(dbConnectionString);
            adapter = new SqlDataAdapter();

            string tableOrTables = tableName;
            string[] tableNamesArray = tableName.Split(',');
            if( tableNamesArray.GetLength(0) > 1 )
                tableOrTables = "JoinedTable";

            adapter.TableMappings.Add("Table", tableOrTables );

            connection.Open();

            string sqlCommand = "SELECT ";
            foreach( string fieldName in fieldNames )
                sqlCommand += (fieldName + ", ");

            SqlCommand command = new SqlCommand( sqlCommand, connection );
            sqlCommand = sqlCommand.Remove( sqlCommand.LastIndexOf(','), 1 );
            sqlCommand += fromClause; 
            
            if( whereClauseField.Length != 0 )
                AddWhereClauseEntry( whereClauseField, whereClauseFieldValue, sqlOperator, SQLLogicalOperator.None );

            sqlCommand += BuildWhereClause( command, dataObject );
            sqlCommand += groupByClause;
            sqlCommand += sortClause;

            command.CommandType = CommandType.Text;
            command.CommandText = sqlCommand;
            adapter.SelectCommand = command;

            dataSet = new DataSet(tableOrTables);
            adapter.Fill(dataSet);

            table = dataSet.Tables[tableOrTables];
            return table;
        }
        #endregion

        #region Set
        /// <summary>        
        /// Save data from the data object to the database.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used to udpate the corresponding database fields.</param>
        /// <param name="whereClauseField">The database field to use in the where clause for this query.  The value to set this field is retrieved from the first parameter 'dataObject'.</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>        
        /// 
        /// <returns>true if the data is successfully updated.</returns>
        public bool Set( object dataObject, string whereClauseField, SQLComparisonOperator sqlOperator )
        {
            bool rc = false;

            try
            {
                connection = new SqlConnection(dbConnectionString);
                connection.Open();

                string sql = string.Format("UPDATE {0} SET ", tableName );
                SqlCommand updateCommand = new SqlCommand( sql, connection );

                if( whereClauseField.Length != 0 )
                    AddWhereClauseEntry( whereClauseField, null, sqlOperator, SQLLogicalOperator.None );

                string whereClause = BuildWhereClause( updateCommand, dataObject );
                string fields = "";
                
                foreach( string fieldName in fieldNames )
                {
                    object value = dataObject.GetType().GetProperty(fieldName).GetGetMethod().Invoke(dataObject, null);
                    if( value != null )
                    {
                        if( !autoIncrementFields.Contains( tableName + "." + fieldName ) )
                        {
                            fields += string.Format("{0} = @{0},", fieldName);
                            updateCommand.Parameters.AddWithValue( "@"+fieldName, value ); 
                        }
                    }
                }

                fields = fields.Remove( fields.Length-1, 1);
                fields += " ";

                sql += fields;
                sql += whereClause;

                updateCommand.CommandText = sql;
                updateCommand.ExecuteNonQuery();

                rc = true;
            }
            catch( ArgumentException ae )
            {
                lastErrorText = ae.ToString();
				OutputDebug( "CustomDB Set Error (" + tableName + ") " + lastErrorText );
            }
            catch( SqlException se)
            {
                lastErrorText = se.ToString();
				OutputDebug( "CustomDB Set Error (" + tableName + ") " + lastErrorText );
            }

            return rc;
        }

        /// <summary>        
        /// Save data from the data object to the database.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used to udpate the corresponding database fields.</param>
        /// <param name="whereClauseField">The database field to use in the where clause for this query.  The value to set this field is retrieved from the first parameter 'dataObject'.</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>        
        /// <param name="fields">Only update the fields from the database specified in this list.  This overrides the list of fields generated from 'dataObject' parameter.</param>
        /// 
        /// <returns>true if the data is successfully updated.</returns>
        public bool SetWithFields( object dataObject, string whereClauseField, SQLComparisonOperator sqlOperator, params string[] fields )
        {
            SetFieldsToInclude( fields );
            return Set( dataObject, whereClauseField, sqlOperator );
        }

        /// <summary>        
        /// Save data from the data object to the database.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used to udpate the corresponding database fields.</param>
        /// <param name="whereClauseField">The database field to use in the where clause for this query.  The value to set this field is retrieved from the first parameter 'dataObject'.</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>        
        /// <param name="fields">Only update the fields from the database NOT specified in this list.  This overrides the list of fields generated from 'dataObject' parameter.</param>
        /// 
        /// <returns>true if the data is successfully updated.</returns>
        public bool SetWithoutFields( object dataObject, string whereClauseField, SQLComparisonOperator sqlOperator, params string[] fields )
        {
            SetFieldsToExclude( fields );
            return Set( dataObject, whereClauseField, sqlOperator );
        }
        #endregion

        #region Add
        /// <summary>        
        /// Add a new record to the database using the supplied data object.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used to udpate the corresponding database fields.</param>
        /// 
        /// <returns>true if the data is successfully updated.</returns>
        public bool Add( object dataObject )
        {
            return Add( ref dataObject );
        }

        /// <summary>        
        /// Add a new record to the database using the supplied data object.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used to udpate the corresponding database fields.</param>
        /// <param name="fields">Only update the fields from the database specified in this list.  This overrides the list of fields generated from 'dataObject' parameter.</param>
        /// 
        /// <returns>true if the data is successfully updated.</returns>
        public bool AddWithFields( object dataObject, params string[] fields )
        {
            SetFieldsToInclude( fields );
            return Add( ref dataObject );
        }

        /// <summary>        
        /// Add a new record to the database using the supplied data object.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used to udpate the corresponding database fields.</param>
        /// <param name="fields">Only update the fields from the database NOT specified in this list.  This overrides the list of fields generated from 'dataObject' parameter.</param>
        /// 
        /// <returns>true if the data is successfully updated.</returns>
        public bool AddWithoutFields( object dataObject, params string[] fields )
        {
            SetFieldsToExclude( fields );
            return Add( ref dataObject );
        }

        /// <summary>        
        /// Add a new record to the database using the supplied data object.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used to udpate the corresponding database fields.</param>
        /// 
        /// <returns>true if the data is successfully updated.</returns>
        public bool Add( ref object dataObject )
        {
            bool rc = false;

            try
            {   
                connection = new SqlConnection(dbConnectionString);
                connection.Open();

                string autoIncrementFieldName = "";
                string sql = string.Format("INSERT INTO {0}", tableName );
                string fields = "(";
                string parms = "VALUES(";
                
                SqlCommand insertCommand = new SqlCommand( sql, connection );

                foreach( string fieldName in fieldNames )
                {
                    object value = dataObject.GetType().GetProperty(fieldName).GetGetMethod().Invoke(dataObject, null);

                    if( value != null )
                    {
                        if( !autoIncrementFields.Contains( tableName + "." + fieldName ) )
                        {
                            fields += (fieldName + ",");
                            parms  += string.Format("@{0},", fieldName);                            
                            insertCommand.Parameters.AddWithValue( "@"+fieldName, value );
                        }
                        else
                            autoIncrementFieldName = fieldName;
                    }
                }

                fields = fields.Remove( fields.Length-1, 1);
                fields += ") ";
                parms = parms.Remove( parms.Length-1, 1);
                parms += ")";

                sql += fields;
                sql += parms;

                insertCommand.CommandText = sql;
                insertCommand.ExecuteNonQuery(); 

                if( autoIncrementFieldName.Length != 0 )
                {
                    Type autoIncrementFieldType = objectType.GetProperty(autoIncrementFieldName).PropertyType;
                    insertCommand.CommandText = "SELECT @@IDENTITY";
                    object autoIncrementFieldValue = insertCommand.ExecuteScalar();

                    if( autoIncrementFieldType.Name == "Int32" )
                        autoIncrementFieldValue = Convert.ToInt32(autoIncrementFieldValue);
                    else if( autoIncrementFieldType.Name == "Int64" )
                        autoIncrementFieldValue = Convert.ToInt64(autoIncrementFieldValue);
                    else if( autoIncrementFieldType.Name == "Int16" )
                        autoIncrementFieldValue = Convert.ToInt16(autoIncrementFieldValue);
                    else if( autoIncrementFieldType.Name == "Decimal" )
                        autoIncrementFieldValue = Convert.ToDecimal(autoIncrementFieldValue);

                    objectType.GetProperty(autoIncrementFieldName).GetSetMethod().Invoke( dataObject, new object[]{ autoIncrementFieldValue } );
                }

                rc = true;
            }
            catch( ArgumentException ae )
            {
                lastErrorText = ae.ToString();
				OutputDebug( "CustomDB Add Error (" + tableName + ") " + lastErrorText );
            }
            catch( SqlException se)
            {
                lastErrorText = se.ToString();
				OutputDebug( "CustomDB Add Error (" + tableName + ") " + lastErrorText );
            }

            return rc;
        }

        /// <summary>        
        /// Add a new record to the database using the supplied data object.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used to udpate the corresponding database fields.</param>
        /// <param name="fields">Only update the fields from the database specified in this list.  This overrides the list of fields generated from 'dataObject' parameter.</param>
        /// 
        /// <returns>true if the data is successfully updated.</returns>
        public bool AddWithFields( ref object dataObject, params string[] fields )
        {
            SetFieldsToInclude( fields );
            return Add( ref dataObject );
        }

        /// <summary>        
        /// Add a new record to the database using the supplied data object.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used to udpate the corresponding database fields.</param>
        /// <param name="fields">Only update the fields from the database NOT specified in this list.  This overrides the list of fields generated from 'dataObject' parameter.</param>
        /// 
        /// <returns>true if the data is successfully updated.</returns>
        public bool AddWithoutFields( ref object dataObject, params string[] fields )
        {
            SetFieldsToExclude( fields );
            return Add( ref dataObject );
        }
        #endregion
        
        #region Delete
        /// <summary>        
        /// Delete records from the database using the specified criteria.
        /// </summary>
        /// <param name="dataObject">The values from this object's data members are used in the where clause of the Database Delete operation.</param>
        /// <param name="whereClauseField">The database field to use in the where clause for this query.  The value to set this field is retrieved from the first parameter 'dataObject'.</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>        
        /// 
        /// <returns>true if the data is successfully deleted.</returns>
        public bool Delete( object dataObject, string whereClauseField, SQLComparisonOperator sqlOperator )
        {
            bool rc = false;

            try
            {   
                connection = new SqlConnection(dbConnectionString);
                connection.Open();

                SqlCommand deleteCommand = new SqlCommand( "", connection );

                if( whereClauseField.Length != 0 )
                    AddWhereClauseEntry( whereClauseField, null, sqlOperator, SQLLogicalOperator.None );

                string sql = string.Format("DELETE FROM {0} {1}", tableName, BuildWhereClause( deleteCommand, dataObject ) );
                deleteCommand.CommandText = sql;
                deleteCommand.ExecuteNonQuery();
                rc = true;
            }
            catch( ArgumentException ae )
            {
                lastErrorText = ae.ToString();
				OutputDebug( "CustomDB Delete Error (" + tableName + ") " + lastErrorText );
            }
            catch( SqlException se)
            {
                lastErrorText = se.ToString();
				OutputDebug( "CustomDB Delete Error (" + tableName + ") " + lastErrorText );
            }

            return rc;
        }
        #endregion

        #region 'Where' Clause
        /// <summary> 
        /// Clear all cached Where clause entries.
        /// </summary>        
        /// 
        public void ClearWhereClauseEntries()
        {
            whereClauseEntries.Clear();
        }

        private void AddWhereClauseEntry( string fieldName, SQLComparisonOperator sqlOperator, SQLLogicalOperator sqlLogOperatorVal )
        {
            whereClauseEntries.Add ( new WhereClauseEntry( fieldName, null, sqlOperator, sqlLogOperatorVal ) );
        }

        /// <summary> 
        /// Add a where clause entry. There is no limit to the number of Where clause entries that may be specified.
        /// </summary>
        /// <param name="fieldName">The name of the database field.</param>
        /// <param name="fieldValue">The value to be used when looking at the 'fieldName'.</param>
        /// <param name="sqlOperator">The comparison operator to use (e.g. 'equals' means field = value)</param>
        /// <param name="sqlLogOperatorVal">The logical operator to be applied.  If there is only one Where clause entry, specify 'None'.  If this is not the first entry, specify AND or OR.</param>
        /// 
        public void AddWhereClauseEntry( string fieldName, object dataObject, SQLComparisonOperator sqlOperator, SQLLogicalOperator sqlLogOperatorVal )
        {
            whereClauseEntries.Add ( new WhereClauseEntry( fieldName, dataObject, sqlOperator, sqlLogOperatorVal ) );
        }

        private string BuildWhereClause( SqlCommand sqlCommand, object dataObject )
        {
            if( whereClause.Length != 0 ) // user has entered the sql manually, nothing to build here.
                return whereClause;

            string clause = "";

            int i = 1; // to avoid duplicate parameter names when a field is used more than once (e.g.  name = @name1 OR name = @name2)
            foreach( WhereClauseEntry entry in whereClauseEntries )
            {
                if( clause.Length == 0 )
                    clause = "WHERE ";
                else
                    clause += ( entry.sqlLogOperator == SQLLogicalOperator.And ? "AND " : "OR " );

                object value = entry.fieldValue;
                if( dataObject != null )
                    value = dataObject.GetType().GetProperty(entry.fieldName).GetGetMethod().Invoke(dataObject, null);

                string parmName = "@"+entry.fieldName+i.ToString();
                sqlCommand.Parameters.AddWithValue( parmName, value );
                clause += string.Format(" {0} {1} {2} ", entry.fieldName, SQLOperatorToString(entry.sqlOperator), parmName );

                i++;
            }

            if( fieldToFilterOutDefaultRec.Length != 0 )
            {
                if( clause.Length == 0 )
                    clause = "WHERE " + fieldToFilterOutDefaultRec + " <> -1 ";
                else
                    clause += ( " AND " + fieldToFilterOutDefaultRec + " <> -1 " );
            }

            return clause;
        }

        private string SQLOperatorToString( SQLComparisonOperator sqlOperator )
        {
            string val = "";

            switch(sqlOperator)
            {
                case SQLComparisonOperator.Is:            val = "=";  break;
                case SQLComparisonOperator.GreaterEqual:  val = ">="; break;
                case SQLComparisonOperator.GreaterThan:   val = ">";  break;
                case SQLComparisonOperator.LessEqual:     val = "<="; break;
                case SQLComparisonOperator.LessThan:      val = "<";  break;
                case SQLComparisonOperator.Not:           val = "<>"; break;
            }

            return val;
        }
        #endregion

        #region Database Schema
        private static void LoadDatabaseSchema( string connectionString )
        {
            autoIncrementFields.Clear();

            dbConnectionString = connectionString;

            try
            {
                using( SqlConnection schemaConnection = new SqlConnection(dbConnectionString) )
                {
                    schemaConnection.Open();

                    DataTable tblTables = schemaConnection.GetSchema(SqlClientMetaDataCollectionNames.Tables, new string[] { null, null, null, "BASE TABLE" });

                    foreach( DataRow rowTable in tblTables.Rows )
                    {
                        string query = string.Format("SELECT * FROM {0} where 1=0", rowTable["table_name"].ToString() );
                       
                        using( SqlConnection cnn = new SqlConnection(dbConnectionString) )
                        {
                            using( SqlCommand cmd = cnn.CreateCommand() )
                            {
                                cmd.CommandText = query;
                                cmd.CommandType = CommandType.Text;
                                cnn.Open();

                                using( SqlDataReader rdr = cmd.ExecuteReader(CommandBehavior.KeyInfo) )
                                {
                                    DataTable tblSchema = rdr.GetSchemaTable();

                                    int numColumns = tblSchema.Columns.Count;
                                    foreach (DataRow dr in tblSchema.Rows)
                                    {
                                        if( ((bool)dr["IsAutoIncrement"]) == true && ((bool)dr["IsKey"]) == true)
                                            autoIncrementFields.Add( rowTable["table_name"].ToString() + "." + dr["ColumnName"] );
                                    }            
                                }
                                cnn.Close();
                            }
                        }
                    }

                    schemaConnection.Close();
                }

            }
            catch( ArgumentException ae )
            {
                lastErrorText = ae.ToString();
				OutputDebug( "CustomDB LoadDatabaseSchema Error: " + lastErrorText );
            }
            catch( SqlException se)
            {
                lastErrorText = se.ToString();
				OutputDebug( "CustomDB LoadDatabaseSchema Error: " + lastErrorText );
            }
        }

        /// <summary> 
        /// Execute a SQL script against the datasource.
        /// </summary>
        /// <param name="connectionString">The connection string for the datasource against which the query will be executed.</param>
        /// <param name="scriptFileName">The fully qualified filename of the sql script.</param>
        /// 
        /// <returns>true if the data is successfully executed.</returns>
        public static bool ExecuteSQLScriptFile( string connectionString, string scriptFileName )
        {
			string sql = null;
			try
			{
				FileInfo fileInfo = new FileInfo(scriptFileName);
				sql = fileInfo.OpenText().ReadToEnd();
			}
			catch( Exception fe )
			{
				lastErrorText = fe.Message;
				OutputDebug( "CustomDB ExecuteSQLScriptFile SQL File Error: " + fe.Message );
				return false;
			}

			try
			{
				return ExecuteSQL( connectionString, sql );
			}
			catch( Exception se )
			{
				lastErrorText = se.Message;
				OutputDebug( "CustomDB ExecuteSQLScriptFile ExecuteSQL Error: " + se.Message );
			}

			return false;
        }

        /// <summary> 
        /// Execute a SQL statement against the datasource.
        /// </summary>
        /// <param name="connectionString">The connection string for the datasource against which the query will be executed.</param>
        /// <param name="sql">The sql statement to execute.</param>
        /// 
        /// <returns>true if the data is successfully executed.</returns>
        public static bool ExecuteSQL( string connectionString, string sql )
        {
            bool rc = false;

            try
            {
				using( SqlConnection connection = new SqlConnection( connectionString ) )
				{

					//	GDT Note: The following code requires the use of a .NET 2.0 assembly.  This causes a mixed mode exception in release mode while running as a service
					//
					//				Server server = new Server(new ServerConnection(connection));
					//				server.ConnectionContext.ExecuteNonQuery(sql);

					connection.Open();
					Regex regex = new Regex( "^GO", RegexOptions.IgnoreCase | RegexOptions.Multiline );
					string[] SqlLine = regex.Split( sql );

					SqlCommand cmd = connection.CreateCommand();
					cmd.Connection = connection;

					foreach( string line in SqlLine )
					{
						if( line.Length > 0 )
						{
							cmd.CommandText = line;
							cmd.CommandType = CommandType.Text;
							try
							{
								cmd.ExecuteNonQuery();
							}
							catch( Exception e )
							{
								lastErrorText = e.ToString();
								OutputDebug( "CustomDB ExecuteSQL Error, line: " + line + ": " + lastErrorText );
								break;
							}
						}
					}
				}

				rc = true;
            }
            catch( Exception e )
            {
                lastErrorText = e.ToString();
				OutputDebug( "CustomDB ExecuteScriptFile Error: " + lastErrorText );
            }

            return rc;
        }

        /// <summary> 
        /// Check for the existance of a database in the supplied data source.
        /// </summary>
        /// <param name="connectionString">The connection string for the datasource to be searched.</param>
        /// <param name="sql">The name of the database who's existance is to be determined.</param>
        /// 
        /// <returns>true if the Database exists.</returns>
        public static bool DatabaseExists( string connectionString, string databaseName )
        {
            bool rc = false;
            databaseName = databaseName.ToUpper();

            using (SqlConnection connection =new SqlConnection(connectionString))
            {
                connection.Open();
                DataTable myData = connection.GetSchema(SqlClientMetaDataCollectionNames.Databases);
                foreach (DataRow row in myData.Rows)
                {
                    if( row[0].ToString().ToUpper() == databaseName ) 
                    {
                        rc = true;
                        break;
                    }
                }

               connection.Close();
            }

            return rc;
        }

        /// <summary> 
        /// Retrieve the name of the local SQL Server service running.  The first instance discovered is returned (wether it be a named instance or not).
        /// </summary>
        /// 
        /// <returns>The name of the instance of a blank string if none are discovered.</returns>
        public static string GetLocalServerInstanceName()
        {
            string serverName = "";
            string localComputerName = Dns.GetHostName().ToUpper();

            DataTable table = System.Data.Sql.SqlDataSourceEnumerator.Instance.GetDataSources();
            DataView view = new DataView(table, string.Empty, "ServerName", DataViewRowState.Added);

            foreach( System.Data.DataRow row in table.Rows )
            {
                string server  = row["ServerName"].ToString().ToUpper();

                if( server.Contains(localComputerName) )
                {
                    serverName = server;
                    if( !row["InstanceName"].Equals(DBNull.Value) )
                    {
                        string instanceName = row["InstanceName"].ToString().ToUpper();
                        if( instanceName.Length != 0 )
                        {
                            if( instanceName.Contains("SQLEXPRESS") )
                                serverName = ".\\" + instanceName; // express
                            else
                                serverName += string.Format("\\{0}", instanceName ); // full sql server
                        }
                    }

                    break;
                }
            }

            return serverName;
        }
        #endregion

        #region Set Include/Exclude Fields
        /// <summary>        
        /// Set the list of fields that will be retrieved from the database.
        /// </summary>
        /// <param name="fields">The list of database fields.  This overrides the list of fields generated from 'objectType' parameter in the constructor.</param>
        ///         
        private void SetFieldsToInclude( string [] fields )
        {
            fieldNames.Clear();
            foreach( string field in fields )
                fieldNames.Add( field );
        }

        /// <summary>        
        /// Set the list of fields that will be retrieved from the database.
        /// </summary>
        /// <param name="fields">Use all fields NOT specified in this list.  This overrides the list of fields generated from 'objectType' parameter in the constructor.</param>
        /// 
        private void SetFieldsToExclude( string [] fields )
        {
            string fieldToExclude = "";
            foreach( string field in fields )
            {
                fieldToExclude = field.ToUpper();

                for( int i = 0; i < fieldNames.Count; i++ )
                {
                    if( fieldToExclude == fieldNames[0].ToString().ToUpper() )
                    {
                        fieldNames.RemoveAt(i);
                        break;
                    }
                }
            }
        }
        #endregion
    }
}
