﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace IntelliSite.CommonUtil.Extensions
{
	public static class CollectionExtensions
	{
		public static void ForEachPair<TFirst, TSecond>( this TFirst[] first, TSecond[] second, Action<TFirst, TSecond> action )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( action == null )
				throw new ArgumentNullException( "action" );

			for( int i = 0; i < first.Length && i < second.Length; i++ )
				action( first[i], second[i] );
		}

		public static void ForEachPair<TFirst, TSecond>( this TFirst[] first, TSecond[] second, Action<TFirst, TSecond, int> action )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( action == null )
				throw new ArgumentNullException( "action" );

			for( int i = 0; i < first.Length && i < second.Length; i++ )
				action( first[i], second[i], i );
		}

		public static void ForEachPair<TFirst, TSecond>( this IEnumerable<TFirst> first, IEnumerable<TSecond> second, Action<TFirst, TSecond> action )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( action == null )
				throw new ArgumentNullException( "action" );

			using( var enum1 = first.GetEnumerator() )
			using( var enum2 = second.GetEnumerator() )
			{
				while( enum1.MoveNext() && enum2.MoveNext() )
					action( enum1.Current, enum2.Current );
			}
		}

		public static void ForEachPair<TFirst, TSecond>( this IEnumerable<TFirst> first, IEnumerable<TSecond> second, Action<TFirst, TSecond, int> action )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( action == null )
				throw new ArgumentNullException( "action" );

			int i = 0;
			using( var enum1 = first.GetEnumerator() )
			using( var enum2 = second.GetEnumerator() )
			{
				while( enum1.MoveNext() && enum2.MoveNext() )
				{
					action( enum1.Current, enum2.Current, i );
					++i;
				}
			}
		}

		public static IEnumerable<TResult> ForEachPair<TFirst, TSecond, TResult>( this TFirst[] first, TSecond[] second, Func<TFirst, TSecond, TResult> func )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( func == null )
				throw new ArgumentNullException( "func" );

			for( int i = 0; i < first.Length && i < second.Length; i++ )
			{
				yield return func( first[i], second[i] );
			}
		}

		public static IEnumerable<TResult> ForEachPair<TFirst, TSecond, TResult>( this TFirst[] first, TSecond[] second, Func<TFirst, TSecond, int, TResult> func )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( func == null )
				throw new ArgumentNullException( "func" );

			for( int i = 0; i < first.Length && i < second.Length; i++ )
			{
				var firstItem = first[i];
				var secondItem = second[i];
				yield return func( firstItem, secondItem, i );
			}
		}

		public static IEnumerable<TResult> ForEachPair<TFirst, TSecond, TResult>( this IEnumerable<TFirst> first, IEnumerable<TSecond> second, Func<TFirst, TSecond, TResult> func )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( func == null )
				throw new ArgumentNullException( "func" );

			using( var enum1 = first.GetEnumerator() )
			using( var enum2 = second.GetEnumerator() )
			{
				while( enum1.MoveNext() && enum2.MoveNext() )
					yield return func( enum1.Current, enum2.Current );
			}
		}

		public static IEnumerable<TResult> ForEachPair<TFirst, TSecond, TResult>( this IEnumerable<TFirst> first, IEnumerable<TSecond> second, Func<TFirst, TSecond, int, TResult> func )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( func == null )
				throw new ArgumentNullException( "func" );

			int i = 0;
			using( var enum1 = first.GetEnumerator() )
			using( var enum2 = second.GetEnumerator() )
			{
				while( enum1.MoveNext() && enum2.MoveNext() )
				{
					yield return func( enum1.Current, enum2.Current, i );
					++i;
				}
			}
		}

		public static T[] ToArrayOrNull<T>( this IEnumerable<T> items )
		{
			if( items == null )
				return null;
			else if( items is List<T> )
				return ( (List<T>)items ).ToArray();

			return ( new List<T>( items ) ).ToArray();
		}

		public static T[] ToArrayOrNull<T>( this List<T> items )
		{
			if( items == null )
				return null;

			return items.ToArray();
		}

		public static T[][] ToArrayOfArraysOrNull<T>( this IEnumerable<IEnumerable<T>> itemsWithItems )
		{
			if( itemsWithItems == null )
				return null;

			var result = new List<T[]>();
			foreach( IEnumerable<T> items in itemsWithItems )
				result.Add( items.ToArrayOrNull() );
			return result.ToArray();
		}

		public static T[][] ToArrayOfArraysOrNull<T>( this List<List<T>> itemsWithItems )
		{
			if( itemsWithItems == null )
				return null;

			var result = new List<T[]>();
			foreach( List<T> items in itemsWithItems )
				result.Add( items.ToArrayOrNull() );
			return result.ToArray();
		}

		public static List<T> ToListOrNull<T>( this IEnumerable<T> items )
		{
			if( items == null )
				return null;

			var result = new List<T>( items );
			return result;
		}

		public static List<T> ToListOrNull<T>( this T[] items )
		{
			if( items == null )
				return null;

			var result = new List<T>( items.Length );
			for( int i = 0; i < items.Length; i++ )
				result.Add( items[i] );
			return result;
		}

		public static List<List<T>> ToListOfListsOrNull<T>( this T[][] itemsWithItems )
		{
			if( itemsWithItems == null )
				return null;

			var result = new List<List<T>>( itemsWithItems.Length );
			for( int i = 0; i < itemsWithItems.Length; i++ )
				result.Add( ToListOrNull( itemsWithItems[i] ) );
			return result;
		}

		public static List<List<T>> ToListOfListsOrNull<T>( this IEnumerable<IEnumerable<T>> itemsWithItems )
		{
			if( itemsWithItems == null )
				return null;

			var result = new List<List<T>>();
			foreach( IEnumerable<T> items in itemsWithItems )
				result.Add( new List<T>( ToListOrNull( items ) ) );
			return result;
		}

		public static IEnumerable<TFirst> WhereZip<TFirst, TSecond>( this TFirst[] first, TSecond[] second, Func<TFirst, TSecond, bool> predicate )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( predicate == null )
				throw new ArgumentNullException( "predicate" );

			for( int i = 0; i < first.Length && i < second.Length; i++ )
			{
				var firstItem = first[i];
				var secondItem = second[i];
				if( predicate( firstItem, secondItem ) )
					yield return firstItem;
			}
		}

		public static IEnumerable<TFirst> WhereZip<TFirst, TSecond>( this TFirst[] first, TSecond[] second, Func<TFirst, TSecond, int, bool> predicate )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( predicate == null )
				throw new ArgumentNullException( "predicate" );

			for( int i = 0; i < first.Length && i < second.Length; i++ )
			{
				var firstItem = first[i];
				var secondItem = second[i];
				if( predicate( firstItem, secondItem, i ) )
					yield return firstItem;
			}
		}

		public static IEnumerable<KeyValuePair<TFirst, TSecond>> EnumerateWith<TFirst, TSecond>( this TFirst[] first, TSecond[] second )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			
			for( int i = 0; i < first.Length && i < second.Length; i++ )
			{
				yield return new KeyValuePair<TFirst, TSecond>( first[i], second[i] );
			}
		}

		public static IEnumerable<KeyValuePair<TFirst, TSecond>> EnumerateWith<TFirst, TSecond>( this IEnumerable<TFirst> first, IEnumerable<TSecond> second )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );

			using( var enum1 = first.GetEnumerator() )
			using( var enum2 = second.GetEnumerator() )
			{
				while( enum1.MoveNext() && enum2.MoveNext() )
				{
					yield return new KeyValuePair<TFirst, TSecond>( enum1.Current, enum2.Current );
				}
			}
		}

		public static IEnumerable<TFirst> WhereZip<TFirst, TSecond>( this IEnumerable<TFirst> first, IEnumerable<TSecond> second, Func<TFirst, TSecond, bool> predicate )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( predicate == null )
				throw new ArgumentNullException( "predicate" );

			using( var enum1 = first.GetEnumerator() )
			using( var enum2 = second.GetEnumerator() )
			{
				while( enum1.MoveNext() && enum2.MoveNext() )
				{
					var firstItem = enum1.Current;
					var secondItem = enum2.Current;
					if( predicate( firstItem, secondItem ) )
						yield return firstItem;
				}
			}
		}

		public static IEnumerable<TFirst> WhereZip<TFirst, TSecond>( this IEnumerable<TFirst> first, IEnumerable<TSecond> second, Func<TFirst, TSecond, int, bool> predicate )
		{
			if( first == null )
				throw new ArgumentNullException( "first" );
			if( second == null )
				throw new ArgumentNullException( "second" );
			if( predicate == null )
				throw new ArgumentNullException( "predicate" );

			int i = 0;
			using( var enum1 = first.GetEnumerator() )
			using( var enum2 = second.GetEnumerator() )
			{
				while( enum1.MoveNext() && enum2.MoveNext() )
				{
					var firstItem = enum1.Current;
					var secondItem = enum2.Current;
					if( predicate( firstItem, secondItem, i ) )
						yield return firstItem;

					++i;
				}
			}
		}
	}
}
