﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;
using System.Text;
using System.Text.RegularExpressions;
using IntelliSite.CommonUtil.FusionApi;

namespace IntelliSite.CommonUtil.Fusion
{
	public static class FusionWrapper
	{
		public static IEnumerable<AssemblyName> EnumCachedAssemblyNames( bool precompiledCache = false, bool globalAssemblyCache = true, bool downloadCache = false )
		{
			var cacheFlags = ASM_CACHE_FLAGS.None;
			if( precompiledCache )
				cacheFlags |= ASM_CACHE_FLAGS.ZAP;
			if( globalAssemblyCache )
				cacheFlags |= ASM_CACHE_FLAGS.GAC;
			if( downloadCache )
				cacheFlags |= ASM_CACHE_FLAGS.DOWNLOAD;

			var displayFlags = ASM_DISPLAY_FLAGS.FULL;

			foreach( var instance in EnumCachedComAssemblyNames( cacheFlags ) )
			{
				yield return ToAssemblyName( instance, displayFlags );
			}
		}

		public static IEnumerable<AssemblyName> EnumCachedAssemblyNames(
			string assemblyName,
			bool setDefaultValues = true,
			bool parseDisplayName = false,
			bool verifyFriendAssemblyName = false,
			bool precompiledCache = false,
			bool globalAssemblyCache = true,
			bool downloadCache = false )
		{
			var nameFlags = CREATE_ASM_NAME_OBJ_FLAGS.None;
			if( setDefaultValues )
				nameFlags |= CREATE_ASM_NAME_OBJ_FLAGS.SET_DEFAULT_VALUES;
			if( parseDisplayName )
				nameFlags |= CREATE_ASM_NAME_OBJ_FLAGS.PARSE_DISPLAY_NAME;
			if( verifyFriendAssemblyName )
				nameFlags |= CREATE_ASM_NAME_OBJ_FLAGS.VERIFY_FRIEND_ASSEMBLYNAME;

			var cacheFlags = ASM_CACHE_FLAGS.None;
			if( precompiledCache )
				cacheFlags |= ASM_CACHE_FLAGS.ZAP;
			if( globalAssemblyCache )
				cacheFlags |= ASM_CACHE_FLAGS.GAC;
			if( downloadCache )
				cacheFlags |= ASM_CACHE_FLAGS.DOWNLOAD;

			var displayFlags = ASM_DISPLAY_FLAGS.FULL;

			foreach( var instance in EnumCachedComAssemblyNames( assemblyName, nameFlags, cacheFlags ) )
			{
				yield return ToAssemblyName( instance, displayFlags );
			}
		}

		internal static IEnumerable<IAssemblyName> EnumCachedComAssemblyNames( ASM_CACHE_FLAGS cacheFlags )
		{
			FusionApi.IAssemblyEnum comAssemblyEnum = null;
			FusionApi.IAssemblyName comAssemblyName = null;

			FusionApi.Native.CreateAssemblyEnum( out comAssemblyEnum, dwFlags: cacheFlags );
			if( comAssemblyEnum == null )
				throw new InvalidOperationException( "Unable to create assembly cache enumerator." );

			do
			{
				try
				{
					if( comAssemblyName != null )
					{
						Marshal.ReleaseComObject( comAssemblyName );
						comAssemblyName = null;
					}

					comAssemblyEnum.GetNextAssembly( ppName: out comAssemblyName );
				}
				catch
				{
					if( comAssemblyName != null )
						Marshal.ReleaseComObject( comAssemblyName );
					Marshal.ReleaseComObject( comAssemblyEnum );
					throw;
				}

				if( comAssemblyName == null )
					break;

				yield return comAssemblyName;
			} while( true );

			Marshal.ReleaseComObject( comAssemblyEnum );
		}

		internal static IEnumerable<IAssemblyName> EnumCachedComAssemblyNames( string assemblyName, CREATE_ASM_NAME_OBJ_FLAGS nameFlags, ASM_CACHE_FLAGS cacheFlags )
		{
			FusionApi.IAssemblyEnum comAssemblyEnum = null;
			FusionApi.IAssemblyName comAssemblyName = null;

			FusionApi.Native.CreateAssemblyNameObject( out comAssemblyName, assemblyName, nameFlags );
			if( comAssemblyName == null )
				throw new InvalidOperationException( "Unable to create assembly name to query from." );

			try
			{
				FusionApi.Native.CreateAssemblyEnum( out comAssemblyEnum, pName: comAssemblyName, dwFlags: cacheFlags );
				if( comAssemblyEnum == null )
					throw new InvalidOperationException( "Unable to create assembly cache enumerator." );
			}
			catch
			{
				if( comAssemblyName != null )
					Marshal.ReleaseComObject( comAssemblyName );
				throw;
			}

			do
			{
				try
				{
					if( comAssemblyName != null )
					{
						Marshal.ReleaseComObject( comAssemblyName );
						comAssemblyName = null;
					}

					comAssemblyEnum.GetNextAssembly( ppName: out comAssemblyName );
				}
				catch
				{
					if( comAssemblyName != null )
						Marshal.ReleaseComObject( comAssemblyName );
					Marshal.ReleaseComObject( comAssemblyEnum );
					throw;
				}

				if( comAssemblyName == null )
					break;

				yield return comAssemblyName;
			} while( true );

			Marshal.ReleaseComObject( comAssemblyEnum );
		}

		internal static IEnumerable<IDENTITY_ATTRIBUTE> EnumerateAttributes( this IEnumIDENTITY_ATTRIBUTE enumerator, bool release = true )
		{
			if( enumerator == null )
				throw new ArgumentNullException( "enumerator" );

			try
			{
				enumerator.Reset();
				IDENTITY_ATTRIBUTE[] attrData = new IDENTITY_ATTRIBUTE[1];
				while( enumerator.Next( 1, attrData ) == 1 )
					yield return attrData[0];
			}
			finally
			{
				if( release )
					Marshal.ReleaseComObject( enumerator );
			}
		}

		internal static T GetAssemblyIdentityFromFile<T>( string filePath ) where T : class
		{
			Type type = typeof( T );
			Guid iid = type.GUID;
			object identity;
			FusionApi.Native.GetAssemblyIdentityFromFile( filePath, ref iid, out identity );
			return (T)identity;
		}

		internal static void GetAssemblyIdentityFromFile<T>( string filePath, out T identity ) where T : class
		{
			identity = default( T );
			identity = GetAssemblyIdentityFromFile<T>( filePath );
		}

		public static string GetAssemblyPath( AssemblyName assemblyName, ProcessorArchitecture? processor = null )
		{
			var arch = processor ?? assemblyName.ProcessorArchitecture;
			return GetAssemblyPath( assemblyName.FullName, arch );
		}

		public static string GetAssemblyPath( string assemblyName, ProcessorArchitecture processor = ProcessorArchitecture.None )
		{
			IAssemblyCache cache;
			FusionApi.Native.CreateAssemblyCache( out cache );
			try
			{
				return FusionWrapper.QueryAssemblyInfo( cache, assemblyName, processor: processor ).AssemblyPath;
			}
			finally
			{
				Marshal.ReleaseComObject( cache );
			}
		}

		public static AssemblyInfo QueryAssemblyInfo( AssemblyName assemblyName, bool validate = true, bool getSize = false, ProcessorArchitecture? processor = null )
		{
			var arch = processor ?? assemblyName.ProcessorArchitecture;
			return QueryAssemblyInfo( assemblyName.FullName, validate, getSize, arch );
		}

		public static AssemblyInfo QueryAssemblyInfo( string assemblyName, bool validate = true, bool getSize = false, ProcessorArchitecture processor = ProcessorArchitecture.None )
		{
			IAssemblyCache cache;
			FusionApi.Native.CreateAssemblyCache( out cache );
			try
			{
				return QueryAssemblyInfo( cache, assemblyName, validate, getSize, processor );
			}
			finally
			{
				Marshal.ReleaseComObject( cache );
			}
		}

		internal static AssemblyInfo QueryAssemblyInfo( IAssemblyCache cache, string assemblyName, bool validate = true, bool getSize = false, ProcessorArchitecture processor = ProcessorArchitecture.None )
		{
			if( cache == null )
				throw new InvalidComObjectException();

			string fullname = assemblyName;

			var arch = processor;
			if( arch == ProcessorArchitecture.None )
				arch = SystemInfo.Current.ProcessorArchitecture;

			const string archTermPrefix = "processorArchitecture=";
			int archIndex;
			if( ( archIndex = assemblyName.IndexOf( archTermPrefix, StringComparison.OrdinalIgnoreCase ) ) < 0 )
			{
				fullname += ", " + archTermPrefix + ToQueryProcessorArchitectureValue( arch );
			}
			else if( processor != ProcessorArchitecture.None )
			{
				const string pattern = @"\bprocessorArchitecture=\w+\b";
				var replacement = archTermPrefix + ToQueryProcessorArchitectureValue( arch );
				fullname = Regex.Replace( fullname, pattern, replacement, RegexOptions.IgnoreCase );
			}

			// Get assembly info
			var flags = QueryAsmInfo_Flag.None;
			if( validate )
				flags |= QueryAsmInfo_Flag.Validate;
			if( getSize )
				flags |= QueryAsmInfo_Flag.GetSize;

			var assemblyInfo = new ASSEMBLY_INFO( 512 );
			cache.QueryAssemblyInfo( flags, fullname, ref assemblyInfo );

			return new AssemblyInfo( fullname, assemblyInfo );
		}

		internal static AssemblyName ToAssemblyName( IAssemblyName instance, ASM_DISPLAY_FLAGS displayFlags )
		{
			if( instance == null )
				throw new ArgumentNullException( "instance" );

			const int bufferSize = 1024;
			StringBuilder stringBuffer = new StringBuilder( bufferSize );
			uint stringBufferSize = (uint)stringBuffer.Capacity;
			instance.GetDisplayName( stringBuffer, ref stringBufferSize, displayFlags );

			var name = new AssemblyName( assemblyName: stringBuffer.ToString() );

			return name;
		}

		public static string ToQueryProcessorArchitectureValue( ProcessorArchitecture processor )
		{
			switch( processor )
			{
				case ProcessorArchitecture.Amd64:
					return "AMD64";
				case ProcessorArchitecture.IA64:
					return "IA64";
				case ProcessorArchitecture.MSIL:
					return "MSIL";
				case ProcessorArchitecture.X86:
					return "x86";
			}

			return null;
		}
	}

	public class AssemblyCache : DisposableBase
	{
		IAssemblyCache _comAssemblyCache;

		public AssemblyCache()
		{
			// Get IAssemblyCache pointer
			IAssemblyCache assemblyCache = null;
			Native.CreateAssemblyCache( out assemblyCache );
			_comAssemblyCache = assemblyCache;
		}

		~AssemblyCache()
		{
			Dispose( false );
		}

		protected override void Dispose( bool disposing )
		{
			// release unmanaged resources
			if( _comAssemblyCache != null )
			{
				Marshal.ReleaseComObject( _comAssemblyCache );
				_comAssemblyCache = null;
			}

			if( disposing )
				GC.SuppressFinalize( this );
		}

		public string GetAssemblyPath( AssemblyName assemblyName, ProcessorArchitecture? processor = null )
		{
			return GetAssemblyPath( assemblyName.FullName, processor ?? assemblyName.ProcessorArchitecture );
		}

		public string GetAssemblyPath( string assemblyName, ProcessorArchitecture processor = ProcessorArchitecture.None )
		{
			return QueryAssemblyInfo( assemblyName, processor: processor ).AssemblyPath;
		}

		public AssemblyInfo QueryAssemblyInfo( AssemblyName assemblyName, bool validate = true, bool getSize = false, ProcessorArchitecture? processor = null )
		{
			return QueryAssemblyInfo( assemblyName.FullName, validate, getSize, processor ?? assemblyName.ProcessorArchitecture );
		}

		public AssemblyInfo QueryAssemblyInfo( string assemblyName, bool validate = true, bool getSize = false, ProcessorArchitecture processor = ProcessorArchitecture.None )
		{
			return FusionWrapper.QueryAssemblyInfo( _comAssemblyCache, assemblyName, validate, getSize, processor );
		}
	}

	public class AssemblyInfo
	{
		protected AssemblyInfo() { }

		internal AssemblyInfo( string assemblyName, ASSEMBLY_INFO info )
		{
			AssemblyName = assemblyName;
			AssemblyPath = info.currentAssemblyPath;
			SizeInKB = info.assemblySizeInKB;
			Installed = info.assemblyFlags.HasFlag( AssemblyInfo_Flag.Installed );
			PayloadResident = info.assemblyFlags.HasFlag( AssemblyInfo_Flag.PayloadResident );
		}

		public string AssemblyName { get; protected set; }
		public string AssemblyPath { get; protected set; }
		public long SizeInKB { get; protected set; }
		public bool Installed { get; protected set; }
		public bool PayloadResident { get; protected set; }
	}

	public class DefinitionIdentity
	{
		protected DefinitionIdentity() { }

		internal DefinitionIdentity( IDefinitionIdentity identity )
		{
			Attributes = FusionWrapper.EnumerateAttributes( identity.EnumAttributes() )
				.Select( a => new IdentityAttribute( a ) )
				.ToArray();
		}

		public IdentityAttribute[] Attributes { get; protected set; }

		public static DefinitionIdentity FromAssemblyFile( string filepath )
		{
			IDefinitionIdentity identity = null;
			try
			{
				FusionWrapper.GetAssemblyIdentityFromFile( filepath, out identity );
				var result = new DefinitionIdentity( identity );
				return result;
			}
			finally
			{
				if( identity != null )
					Marshal.ReleaseComObject( identity );
			}
		}
	}

	public class ReferenceIdentity
	{
		protected ReferenceIdentity() { }

		internal ReferenceIdentity( IReferenceIdentity identity )
		{
			Attributes = FusionWrapper.EnumerateAttributes( identity.EnumAttributes() )
				.Select( a => new IdentityAttribute( a ) )
				.ToArray();
		}

		public IdentityAttribute[] Attributes { get; protected set; }

		public static ReferenceIdentity FromAssemblyFile( string filepath )
		{
			IReferenceIdentity identity = null;
			try
			{
				FusionWrapper.GetAssemblyIdentityFromFile( filepath, out identity );
				var result = new ReferenceIdentity( identity );
				return result;
			}
			finally
			{
				if( identity != null )
					Marshal.ReleaseComObject( identity );
			}
		}
	}

	public class IdentityAttribute
	{
		public IdentityAttribute() { }

		public IdentityAttribute( string _namespace, string name, string value )
		{
			Namespace = _namespace;
			Name = name;
			Value = value;
		}

		internal IdentityAttribute( IDENTITY_ATTRIBUTE attr )
		{
			Namespace = attr.Namespace;
			Name = attr.Name;
			Value = attr.Value;
		}

		public string Namespace { get; set; }
		public string Name { get; set; }
		public string Value { get; set; }
	}
}
