using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;

namespace IntelliSite.CommonUtil
{
	public class LogMessageArgs : EventArgs
	{
		public string Message { get; private set; }
		public DateTime? Timestamp { get; private set; }

		public LogMessageArgs( string message )
			: base()
		{
			Message = message;
		}

		public LogMessageArgs( string message, DateTime timestamp )
			: this( message )
		{
			Timestamp = timestamp;
		}

		public override string ToString()
		{
			return "{LogMessageArgs: Message=" + ( Message ?? "{null}" ) + "}";
		}
	}

	/// <summary>
	/// Summary description for ISLogger.
	/// </summary>
	public class LoggerBase : DisposableBase
	{
		private static Encoding _defaultEncoding = new UTF8Encoding( encoderShouldEmitUTF8Identifier: false, throwOnInvalidBytes: true );

		public static Encoding DefaultEncoding { get { return _defaultEncoding; } }

		private string          filename;
		private Queue<string>	queue;
		private AutoResetEvent	writeEvent;
		private RegisteredWaitHandle writeWaitHandle;

		public Encoding         Encoding         { get; set; }
		public bool             IncludeTimestamp { get; set; }
		public string           Filename         { get { lock( filename ) return filename; } set { lock( filename ) filename = value; } }
		public bool             DebugWrite       { get; set; }
		public NewLineOptions   NewLineHandling  { get; set; }

		[Flags]
		public enum NewLineOptions
		{
			TranslateToCurrentPlatform = 1 << 0, // translate \n and \r\n to System.Environment.NewLine
			AppendNewLine = 1 << 1,

			None = 0,
			Default = TranslateToCurrentPlatform | AppendNewLine
		}

		
		protected event EventHandler<LogMessageArgs> WriteMessage;

		protected LoggerBase()
		{
		}

		/// <summary>
		/// Create a logger
		/// </summary>
		/// <param name="filename">A file path to write the log to.</param>
		/// <param name="encoding">Encoding to use. (if null: UTF-8 without BOM)</param>
		public LoggerBase( string filename, Encoding encoding = null, bool includeTimestamp = true )
		{
			if( filename == null )
				throw new ArgumentNullException( "filename" );

			this.filename    = filename;
			WriteMessage     = null;
			Encoding         = encoding ?? _defaultEncoding;
			IncludeTimestamp = includeTimestamp;
			DebugWrite       = true;
			NewLineHandling  = NewLineOptions.Default;

			queue           = new Queue<string>();
			writeEvent      = new AutoResetEvent( false );
			writeWaitHandle = ThreadPool.RegisterWaitForSingleObject( writeEvent, WriteCallback, state: this, millisecondsTimeOutInterval: 3000, executeOnlyOnce: false );
		}

		public LoggerBase( EventHandler<LogMessageArgs> writeMessageHandler, Encoding encoding = null, bool includeTimestamp = true )
		{
			WriteMessage    += writeMessageHandler;
			Encoding         = encoding ?? _defaultEncoding;
			IncludeTimestamp = includeTimestamp;
			DebugWrite       = true;
			NewLineHandling  = NewLineOptions.AppendNewLine;
		}

		~LoggerBase()
		{
			Dispose( false );
		}

		protected override void Dispose( bool disposing )
		{
			if( writeWaitHandle != null )
			{
				writeWaitHandle.Unregister( writeEvent );
				writeWaitHandle = null;
			}

			if( disposing )
			{
				if( writeEvent != null )
				{
					writeEvent.Dispose();
					writeEvent = null;
				}

				GC.SuppressFinalize( this );
			}
		}

		public void Write( object logEntry )
		{
			Write( logEntry.ToString() );
		}

		public void Write( string logEntry )
		{
			if( String.IsNullOrWhiteSpace( logEntry ) )
				return;

			DateTime timestamp = DateTime.Now;
			string entry = logEntry;

			if( DebugWrite )
			{
				if( NewLineHandling.HasFlag( NewLineOptions.AppendNewLine ) )
					Debug.WriteLine( logEntry );
				else
					Debug.Write( logEntry );
			}

			if( IncludeTimestamp )
				entry = timestamp.ToString( "yyyy/MM/dd HH:mm:ss.FFF" ) + ": " + entry;

			EventHandler<LogMessageArgs> handler = WriteMessage;
			if( handler != null )
			{
				handler( this, new LogMessageArgs( entry, timestamp ) );
			}
			else if( writeWaitHandle != null )
			{
				lock( queue )
					queue.Enqueue( entry );
				writeEvent.Set();
			}
		}

		public void Write( string format, params object[] args )
		{
			Write( String.Format( format, args: args ) );
		}

		public void WriteCallback( object state, bool timedOut )
		{
			if( !timedOut )
			{
				lock( filename )
				{
					NewLineOptions options = NewLineHandling;
					bool translateNewLines = options.HasFlag( NewLineOptions.TranslateToCurrentPlatform );
					bool appendNewLine = options.HasFlag( NewLineOptions.AppendNewLine );

					int tries;
					for( tries = 4; tries > 0; --tries )
					{
						try
						{
							using( var stream = new FileStream( filename, FileMode.Append, FileAccess.Write, FileShare.Read ) )
							using( var writer = new StreamWriter( stream, this.Encoding ) )
							{
								while( true )
								{
									string[] entries;
									lock( queue )
									{
										entries = queue.ToArray();
										queue.Clear();
									}

									if( entries.Length <= 0 )
										return;
									
									if( translateNewLines )
										for( int i = 0, len = entries.Length; i < len; ++i )
											entries[i] = Regex.Replace( entries[i], @"\r?\n", Environment.NewLine );

									if( appendNewLine )
										foreach( var entry in entries )
											writer.WriteLine( entry );
									else
										foreach( var entry in entries )
											writer.Write( entry );

									writer.Flush();
									stream.Flush();
								}
							}
						}
						catch( IOException ) { }

						Thread.Sleep( millisecondsTimeout: 250 );
					}

					if( tries == 0 )
						Debug.WriteLine( "Failed to write to log." );
				}
			}
		}
	}
}
