using System;
using System.IO;

namespace IntelliSite.CommonUtil
{
	/// <summary>
	/// MessageAttribute: This attribute needs to be applied to a MessageBase derivitive class to instruct MessageSocket that the class is a message definition.
	/// </summary>
	[AttributeUsage(AttributeTargets.Class, AllowMultiple = false)]
	public class MessageAttribute : System.Attribute
	{
	}

	public class MessageBase
	{
		public const short MessageBaseStart	= 0;

		public static int MessageBlockSize	= 512;
		public static int MessageHeaderSize	= 14;

		public short   Type                    { get; set; }
		public byte [] Data                    { get; set; }
		public int     DataLen                 { get { if( Data != null ) return Data.Length; else return 0; } }
		public byte [] Header                  { get; private set; }
		public int     HeaderSize              { get; private set; }
		public int     Transferred             { get; internal set; }
		public int     Received                { get; private set; }
		
		public bool    NetworkTransferComplete { get { return Transferred == DataLen; } }

		public Int32   TotalSizeFromHeader     { get { return BitConverter.ToInt32( Header, 0 ); } }
		public Int32   ChunkSizeFromHeader     { get { return BitConverter.ToInt32( Header, 4 ); } }
		public Int32   OffsetFromHeader        { get { return BitConverter.ToInt32( Header, 8 ); } }
		public short   TypeFromHeader          { get { return BitConverter.ToInt16( Header, 12 ); } }

		virtual public object  Target          { get { return null; } }

		public MessageBase( short type, byte [] data )
		{
			Type       = type;
			Data       = data;
			Header     = new byte[MessageHeaderSize];
			HeaderSize = 0;
			Received   = 0;
		}
	
		public MessageBase( short type )
		{
			Type       = type;
			Data       = new byte[0];
			Header     = new byte[MessageHeaderSize];
			HeaderSize = 0;
			Received   = 0;
		}
		
		public MessageBase( MessageBase message )
		{
			Type       = message.Type;
			Data       = message.Data;
			Header     = new byte[MessageHeaderSize];
			HeaderSize = 0;
			Received   = 0;
		}
		
		public MessageBase()
		{
			Type       = 0;
			Data       = null;
			Header     = new byte[MessageHeaderSize];
			HeaderSize = 0;
			Received   = 0;
		}
		
		public int TransferSize
		{
			get
			{
				if( DataLen <= ( MessageBlockSize - MessageHeaderSize ) )
					return DataLen;
				else if( ( DataLen - Transferred ) < ( MessageBlockSize - MessageHeaderSize ) )
					return DataLen - Transferred;
				else
					return MessageBlockSize - MessageHeaderSize;
			}
		}

	//	Fill the header of any part of a message to be transferred.
		public void FillHeader( int size, byte [] buffer )
		{
			MemoryStream memStream = new MemoryStream( buffer );
			BinaryWriter writer = new BinaryWriter( memStream );
			writer.Write( (int)DataLen );
			writer.Write( size );
			writer.Write( Transferred );
			writer.Write( Type );
		}

	//	Get the first part of a message to be transferred.
		public void GetNetworkBuffer( out int size, byte [] buffer )
		{
			Transferred = 0;
			size = TransferSize;
			FillHeader( size, buffer );
			if( size > 0 )
				Array.Copy( Data, 0, buffer, MessageHeaderSize, size );
			Transferred = size;
			size += MessageHeaderSize;
		}
		
	//	Get the subsequent parts of a multi-part message to be transferred.		
		public void GetNextNetworkBuffer( out int size, byte [] buffer )
		{
			size = TransferSize;
			FillHeader( size, buffer );
			if( size > 0 )
				Array.Copy( Data, Transferred, buffer, MessageHeaderSize, size );
			Transferred += size;
			size += MessageHeaderSize;
		}

	//	Returns true if the transfer is complete.
		public bool ReceiveNetworkBuffer( byte [] buffer, int bytes, ref int used )
		{
			while( used < bytes )
			{
			//	If I don't already have a complete header, fill it
				if( HeaderSize != MessageHeaderSize )
				{
					int len = Math.Min( MessageHeaderSize - HeaderSize, bytes - used );
					Array.Copy( buffer, used, Header, HeaderSize, len );
					HeaderSize += len;
					used += len;

					if( HeaderSize == MessageHeaderSize )
					{
						Type = TypeFromHeader;
						Data = new byte[TotalSizeFromHeader];
						Received = 0;
					}
				}

				if( HeaderSize == MessageHeaderSize )
				{
					if( DataLen > 0 )
					{
						int len = Math.Min( Data.Length - Received, bytes - used );
						Array.Copy( buffer, used, Data, Received, len );
						Received += len;
						used += len;
					}

					if( Received == DataLen )
					{
						HeaderSize = 0;
						return true;
					}
				}
			}
			
			return false;
		}

		#region ToString		
		override public string ToString()
		{
			switch( Type )
			{
				default:
					return "Unknown Type: " + Type;
			}
		}
		#endregion
	}
}
