﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.Serialization;
using System.Text;

namespace IntelliSite.CommonUtil
{
	public class DuplicateMessageTypeException : Exception
	{
		public DuplicateMessageTypeException() : base() { }
		public DuplicateMessageTypeException( string message ) : base( message ) { }
		public DuplicateMessageTypeException( string message, Exception inner ) : base( message, inner ) { }

		// A constructor is needed for serialization when an 
		// exception propagates from a remoting server to the client.  
		protected DuplicateMessageTypeException( SerializationInfo info, StreamingContext context ) { }
	}

	public class NoMessageTypeException : Exception
	{
		public NoMessageTypeException() : base() { }
		public NoMessageTypeException( string message ) : base( message ) { }
		public NoMessageTypeException( string message, Exception inner ) : base( message, inner ) { }

		// A constructor is needed for serialization when an 
		// exception propagates from a remoting server to the client.  
		protected NoMessageTypeException( SerializationInfo info, StreamingContext context ) { }
	}

	public class MessageDictionary
	{
		private static Dictionary<short, Type> dictionary = null;
		public LoggerBase Logger { get; set; }

		public List<Assembly> ProductAssemblies { get; set; }
		public List<string> ProcessedAssemblyNames { get; set; }
		public List<string> FailedAssemblyNames { get; set; }
		public Dictionary<short, Type> Dictionary { get { return dictionary; } }

		public MessageDictionary( LoggerBase logger = null )
		{
			Logger = logger;
			Build();
		}

		static IEnumerable<Type> GetTypesWithMessageAttribute( Assembly assembly )
		{
			Type[] assemblyTypes;
			try
			{
				assemblyTypes = assembly.GetTypes();
			}
			catch( System.Reflection.ReflectionTypeLoadException ex )
			{
#if DEBUG
				StringBuilder sb = new StringBuilder( 1024 );
				sb.Append( ex );
				for( int i = 0; i < ex.LoaderExceptions.Length; i++ )
					sb.AppendFormat( "\n  LoaderExceptions[{0}]={1}", i, ex.LoaderExceptions[i] );
				System.Diagnostics.Debug.Print( sb.ToString() );
#endif

				assemblyTypes = ex.Types;
			}

			foreach( Type type in assemblyTypes )
				if( type != null && type.GetCustomAttributes( typeof( MessageAttribute ), true ).Length > 0 )
					yield return type;
		}

		private void Build()
		{
			if( dictionary != null )
				return;

			if( Logger != null )
				Logger.Write( "MessageDictionary:" );

			dictionary = new Dictionary<short, Type>();

			ProductAssemblies = new List<Assembly>( GetLoadedProductAssemblies() );
			ProcessedAssemblyNames = new List<string>();
			FailedAssemblyNames = new List<string>();

			foreach( Assembly assembly in ProductAssemblies )
				BuildFromAssembly( assembly );
		}

		public void BuildFromAssembly( string dllName )
		{
			AssemblyName assemblyName = new AssemblyName( dllName );
			BuildFromAssembly( assemblyName );
		}

		public void BuildFromAssembly( AssemblyName assemblyName )
		{
			string assemblyFullName = assemblyName.FullName;

			if( assemblyName == null || FailedAssemblyNames.Contains( assemblyFullName ) )
				return;

			Assembly[] loadedAssemblies = GetLoadedProductAssemblies().ToArray();
			try
			{
				Assembly assembly = null;
				foreach( Assembly loadedAssembly in loadedAssemblies )
				{
					if( loadedAssembly.GetName() == assemblyName )
					{
						assembly = loadedAssembly;
						break;
					}
				}

				if( assembly == null )
					assembly = ProductInfo.Current.LoadAssemblyFromName( assemblyName );

				if( assembly == null )
					assembly = Assembly.Load( assemblyName );

				if( assembly == null )
					throw new FileLoadException( String.Format( "Unable to load assembly from name: {0}", assemblyName ) );

				BuildFromAssembly( assembly );
			}
			catch( ArgumentNullException ) { FailedAssemblyNames.Add( assemblyFullName ); return; }
			catch( FileNotFoundException ) { FailedAssemblyNames.Add( assemblyFullName ); return; }
			catch( FileLoadException ) { FailedAssemblyNames.Add( assemblyFullName ); return; }
			catch( BadImageFormatException ) { FailedAssemblyNames.Add( assemblyFullName ); return; }
		}

		public void BuildFromAssembly( Assembly assembly )
		{
			if( ProcessedAssemblyNames.Contains( assembly.FullName ) )
				return;

			ProcessedAssemblyNames.Add( assembly.FullName );

			foreach( Type type in GetTypesWithMessageAttribute( assembly ) )
			{
				FieldInfo messageType = null;
				foreach( FieldInfo fieldInfo in type.GetFields() )
					if( fieldInfo.Name == "MessageType" )
					{
						messageType = fieldInfo;
						break;
					}

				if( messageType != null )
				{
					short fieldValue = (short)messageType.GetValue( null );

					if( dictionary.ContainsKey( fieldValue ) )
						throw new DuplicateMessageTypeException();

					// remember that Logger will only be true if SocketLogging is enabled!
					if( Logger != null )
						Logger.Write( "\t" + type.FullName + " (" + fieldValue + ")" );

					dictionary.Add( fieldValue, type );
				}
				else
					throw new NoMessageTypeException();
			}

			foreach( AssemblyName refAssemblyName in assembly.GetReferencedAssemblies() )
				BuildFromAssembly( refAssemblyName );
		}

		public static IEnumerable<Assembly> GetLoadedProductAssemblies()
		{
			// use CommonUtil's public key look for assemblies with the same signing key
			Assembly executingAssembly = Assembly.GetExecutingAssembly();
			byte[] productAssemblyPublicKey = executingAssembly.GetName().GetPublicKey();
			Assembly[] appDomainAssembies = AppDomain.CurrentDomain.GetAssemblies();
			foreach( Assembly assembly in appDomainAssembies )
			{
				byte[] assemblyPublicKey = assembly.GetName().GetPublicKey();
				if( Enumerable.SequenceEqual( productAssemblyPublicKey, assemblyPublicKey ) )
					yield return assembly;
			}
		}

		public object GetMessageOfType( MessageBase message )
		{
			Type type;

			try
			{
				if( !dictionary.TryGetValue( message.Type, out type ) )
					return message;

				// TODO: FIXME!  There must be a better way to do this, as this required boxing and unboxing then you make the parameters for constructor.Invoke.
				Type[] types = new Type[] { typeof( MessageBase ) };
				ConstructorInfo constructor = type.GetConstructor( types );
				Object[] parameters = new Object[] { message };
				object newMessage = constructor.Invoke( parameters );

				return newMessage;
			}
			catch
			{
				return message;
			}
		}
	}
}
