﻿#if TRACE && DEBUG
//#define MessageStreamCheckSize
//#define MessageStreamTrace
#endif

using System;
using System.Collections.Generic;
using System.IO;
using System.Threading;
using IntelliSite.CommonUtil.Extensions;

namespace IntelliSite.CommonUtil
{
	public class MessageStream : DisposableBase
	{
	//	Properties
		public string                   Name              { get; set; }
		public Stream                   Stream            { get; set; }
		public LoggerBase               Logger            { get; set; }
		public byte                     Type              { get; set; }
		public object                   Owner             { get; set; }

		private Queue<MessageBase>      MessageQueue      { get; set; }
		private Thread                  ReadThread        { get; set; }
		private bool                    StopReading       { get; set; }

		public static MessageDictionary MessageDictionary = null;

	//	Events
		public event EventHandler<EventArgs> MessageReceived;
		public event EventHandler<ErrorArgs> ReadMessageError;
		public event EventHandler<EventArgs> ReadMessagesStopped;

	//  Constructors
		public MessageStream( string name, EventHandler<EventArgs> messageReceivedHandler, Stream stream, LoggerBase logger, object owner )
		{
			if( MessageDictionary == null )
				MessageDictionary = new MessageDictionary();

			Name              = name;
			Stream            = stream;
			Logger            = logger;
			Owner             = owner;
			StopReading       = false;
			MessageQueue      = new Queue<MessageBase>();
			MessageReceived  += messageReceivedHandler;

			ReadThread        = new Thread( ReadMessages );
			ReadThread.IsBackground = true;
			ReadThread.Start();
		}

		#region Destructor/Dispose
		~MessageStream()
		{
			Dispose( false );
		}

		protected override void Dispose( bool disposing )
		{
			StopReading = true;
			ReadThread = null;

			if( disposing )
				GC.SuppressFinalize( this );
		}
		#endregion

		public MessageBase GetMessage()
		{
			if( MessageQueue.Count == 0 )
				return null;
			else
			{
				MessageBase message = MessageQueue.Dequeue();
				return message;
			}
		}

		public void ReadMessages()
		{
			using( BinaryReader reader = new BinaryReader( Stream ) )
			{
				MessageBase baseMessage = new MessageBase();

				while( !StopReading )
				{
					try
					{
						if( !reader.BaseStream.CanRead )
						{
							StopReading = true;
							break;
						}


						Int32 totalSize = reader.ReadInt32();
						Int32 chunkSize = reader.ReadInt32();
						Int32 offset = reader.ReadInt32();
						Int16 msgType = reader.ReadInt16();

#if TRACE && MessageStreamTrace
						System.Diagnostics.Trace.TraceInformation(
							"MessageStream: Reading message data." +
							$" [totalSize={totalSize}, chunkSize={chunkSize}, offset={offset}, msgType={msgType}]" );
#endif

#if MessageStreamCheckSize
						const int MIN_MESSAGE_SIZE = 0;
						const int MAX_MESSAGE_SIZE = 100 << 20; // 100 MiB
						if( totalSize < MIN_MESSAGE_SIZE || totalSize > MAX_MESSAGE_SIZE )
							throw new InvalidOperationException( $"Cannot read message stream with bad total size: {totalSize}" );
#endif

						baseMessage.Data = new byte[totalSize];
						baseMessage.Type = msgType;

						int totalDataSize = 0;
						while( totalDataSize < totalSize )
						{
							byte[] chunk = reader.ReadBytes( chunkSize );
							Array.Copy( chunk, 0, baseMessage.Data, offset, chunkSize );
							totalDataSize += chunkSize;

							if( totalDataSize < totalSize )
							{
								totalSize = reader.ReadInt32();
								chunkSize = reader.ReadInt32();
								offset = reader.ReadInt32();
								Int16 chunkType = reader.ReadInt16();
							}
						}

						OutputDebug( "New Message: Type: {0}", baseMessage.Type );
						MessageBase specificMessage = (MessageBase)MessageDictionary.GetMessageOfType( baseMessage );
						OutputDebug( "New Message: {0}", specificMessage );

						MessageQueue.Enqueue( specificMessage );

						EventHandler<EventArgs> handler = MessageReceived;
						if( handler != null )
							handler( this, EventArgs.Empty );
						else
							OutputDebug( "MessageStream[{0}]: ReadMessage message dropped, no MessageReceived handler.", Name );

						Thread.Yield();
					}
					catch( EndOfStreamException ex )
					{
#if TRACE && DEBUG
						System.Diagnostics.Trace.TraceInformation( String.Format( "MessageStream[{0}]: Stop reading. EndOfStreamException received: {1}", Name, ex ) );
#endif
						StopReading = true;
						ReadMessageError.FireEvent( this, new ErrorArgs( this, ex.Message, ex ) );
						break;
					}
					catch( Exception ex )
					{
						string text = String.Format( "MessageStream[{0}]: ReadMessage exception: {1}", Name, ex );
#if TRACE && DEBUG
						System.Diagnostics.Trace.TraceError( text );
#endif
						OutputDebug( text );
						ReadMessageError.FireEvent( this, new ErrorArgs( this, ex.Message, ex ) );
						break;
					}
				}
			}

			ReadMessagesStopped.FireEvent( this, EventArgs.Empty );
		}

		public void SendMessage( MessageBase message )
		{
			lock( this )
			{
				OutputDebug( "{0}: Tx: {1}", Name, message.ToString() );

				try
				{
					int size;
					byte[] data = new byte[message.TransferSize + MessageBase.MessageHeaderSize];
					message.GetNetworkBuffer( out size, data );
					Stream.Write( data, 0, data.Length );
					while( !message.NetworkTransferComplete )
					{
						data = new byte[message.TransferSize + MessageBase.MessageHeaderSize];
						message.GetNextNetworkBuffer( out size, data );
						Stream.Write( data, 0, data.Length );
					}
					message.Transferred = 0;
				}
				catch( Exception e )
				{
					string text = String.Format( "MessageStream[{0}]: SendMessage exception: {1}", Name, e );
#if TRACE && DEBUG
					System.Diagnostics.Trace.TraceError( text );
#endif
					OutputDebug( text );
				}
			}
		}

		public override string ToString()
		{
			return Name;
		}

		public void OutputDebug( string message )
		{
			if( Logger != null )
				Logger.Write( message );
		}

		public void OutputDebug( string format, params object[] list )
		{
			if( Logger != null )
				Logger.Write( String.Format( format, list ) );
		}
	}
}
