﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;

namespace IntelliSite.CommonUtil
{
	public static class PathHelper
	{
		const string _DevicePathPrefix = @"\\.\";
		const string _LongPathPrefix = @"\\?\";

		public static string DevicePathPrefix { get { return _DevicePathPrefix; } }
		public static string LongPathPrefix { get { return _LongPathPrefix; } }

		[DllImport( "kernel32.dll", CharSet = CharSet.Unicode, SetLastError = true )]
		[return: MarshalAs( UnmanagedType.U1 )]
		static extern bool CreateSymbolicLink( string symlinkPath, string targetPath, SymbolicLinkFlags flags );

		enum SymbolicLinkFlags : int
		{
			File = 0,
			Directory = 1
		}

		public static void CreateDirectorySymbolicLink( string symlinkPath, string targetPath )
		{
			if( !CreateSymbolicLink( symlinkPath, targetPath, SymbolicLinkFlags.Directory ) )
			{
				int lastError = Marshal.GetLastWin32Error();
				Marshal.ThrowExceptionForHR( lastError );
			}
		}

		public static void CreateFileSymbolicLink( string symlinkPath, string targetPath )
		{
			if( !CreateSymbolicLink( symlinkPath, targetPath, SymbolicLinkFlags.File ) )
			{
				int lastError = Marshal.GetLastWin32Error();
				Marshal.ThrowExceptionForHR( lastError );
			}
		}

		public static bool TryCreateDirectorySymbolicLink( string symlinkPath, string targetPath )
		{
			return CreateSymbolicLink( symlinkPath, targetPath, SymbolicLinkFlags.Directory );
		}

		public static bool TryCreateFileSymbolicLink( string symlinkPath, string targetPath )
		{
			return CreateSymbolicLink( symlinkPath, targetPath, SymbolicLinkFlags.File );
		}

		public static string GetDevicePrefixedPath( string path )
		{
			return Regex.Replace( path, @"\A([A-Za-z]:(\\|\z))", _DevicePathPrefix + @"$1" );
		}

		public static string GetLongPrefixedPath( string path )
		{
			return Regex.Replace( path, @"\A([A-Za-z]:(\\|\z))", _LongPathPrefix + @"$1" );
		}

		public static bool IsValidDirectoryPath( string path )
		{
			DirectoryInfo info = null;
			try
			{
				info = new DirectoryInfo( path );
			}
			catch( ArgumentException ) { }
			catch( PathTooLongException ) { }
			catch( NotSupportedException ) { }

			return !Object.ReferenceEquals( info, null );
		}

		public static bool IsValidFileName( string fileName )
		{
			FileInfo info = null;
			try
			{
				info = new FileInfo( fileName );
			}
			catch( ArgumentException ) { }
			catch( PathTooLongException ) { }
			catch( NotSupportedException ) { }

			return !Object.ReferenceEquals( info, null );
		}

		public static Version GetVersionFromFile( string filename )
		{
			FileVersionInfo installerVersionInfo = FileVersionInfo.GetVersionInfo( filename );
			if( installerVersionInfo == null )
				throw new IOException( "File doesn't have any version info." );
			if( installerVersionInfo.FileVersion == null )
				throw new IOException( "File's version info doesn't contain a file version." );

			return new Version( installerVersionInfo.FileVersion );
		}

		public static string GetVersionStringFromFile( string filename )
		{
			FileVersionInfo installerVersionInfo = FileVersionInfo.GetVersionInfo( filename );
			if( installerVersionInfo == null )
				throw new IOException( "File doesn't have any version info." );
			return installerVersionInfo.FileVersion;
		}
	}
}
