﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace IntelliSite.CommonUtil
{
	public class PeekableSocketClientStream : PeekableStream
	{
		public string            HostAddress       { get; private set; }
		public int               Port              { get; private set; }
		public LoggerBase        Logger            { get; set; }
		public Socket            Socket            { get; private set; }
		public AutoResetEvent    ConnectEvent      { get; private set; }
		public ErrorArgs         ConnectError      { get; set; }

		protected List<byte[]>   Data              { get; set; }
		protected AutoResetEvent DataReceivedEvent { get; set; }
		public bool              Connected         { get { return Socket != null && Socket.Connected; } }
		private bool             StopReading       { get; set; }
		private byte[]           Buffer            { get; set; }

		public event EventHandler<ErrorArgs> ConnectionLoss;

	//	PeekableStream Properties
		public override bool SupportsMultipleChunks                         { get { return true; } }

		public override int ChunkCount										{ get { lock( Data ) { return Data.Count; } } }
		public override int ChunkSize( int index )	                        { lock( Data ) { return Data[index].Length; } }
		public override byte[] PeekChunk( int index, int offset, int size ) { lock( Data ) { return Data[index]; } }

		public override bool CanRead                                        { get { return true; } }
		public override bool CanSeek                                        { get { return false; } }
		public override bool CanWrite                                       { get { return true; } }
		public override void Flush()                                        { throw new NotImplementedException(); }
		public override long Position                                       { get { throw new NotImplementedException(); } set { throw new NotImplementedException(); } }
		public override void SetLength( long value )                        { throw new NotImplementedException(); }
		public override long Seek( long offset, SeekOrigin origin )			{ throw new NotImplementedException(); }

		public PeekableSocketClientStream( string name = "PeekableSocketClientStream", Encoding encoding = null, LoggerBase logger = null ) : base( name, encoding )
		{
			Init( logger );
		}

		public PeekableSocketClientStream( string name, Socket socket, Encoding encoding = null, LoggerBase logger = null )	: base( name, encoding )
		{
			Init( logger );
			Socket = socket;

			SocketAsyncEventArgs args = new SocketAsyncEventArgs();
			args.SocketError          = SocketError.Success;
			args.RemoteEndPoint       = Socket.RemoteEndPoint;
			OnConnectCompleted( this, args );

			Name = "Client " + HostAddress;
			args.Dispose();
		}

		~PeekableSocketClientStream()
		{
			Dispose( false );
		}

		#region protected override void Dispose( bool disposing )
		private bool disposed = false;
		protected override void Dispose( bool disposing )
		{
			if( !disposed )
			{
				if( disposing )
				{
					StopReading = true;
					lock( Data )
					{
						Data.Clear();
						DataReceivedEvent.Set();
						if( Socket != null )
						{
							Socket.Dispose();
							Socket = null;
						}
					}
				}

				disposed = true;

				base.Dispose( disposing );
			}
		}
		#endregion

		#region public void Init( LoggerBase logger )
		public void Init( LoggerBase logger )
		{
			HostAddress       = "";
			Port              = 0;
			Socket            = null;
			Logger            = logger;
			ConnectionLoss    = null;
			Data              = new List<byte[]>();
			DataReceivedEvent = new AutoResetEvent( false );
			ConnectEvent      = new AutoResetEvent( false );
			StopReading       = false;
			Buffer            = new byte[8192];
		}
		#endregion

		#region protected void ReceiveAsync()
		protected void ReceiveAsync()
		{
			SocketAsyncEventArgs readArgs = new SocketAsyncEventArgs();
			//readArgs.BufferList = new List<ArraySegment<byte>> {
			//    new ArraySegment<byte>( new byte[8192] ),
			//    new ArraySegment<byte>( new byte[8192] )
			//};
			readArgs.SetBuffer( Buffer, 0, Buffer.Length );
			readArgs.UserToken = this;
			readArgs.Completed += OnReadCompleted;
			try
			{
				if( Socket != null && !Socket.ReceiveAsync( readArgs ) )
					OnReadCompleted( this, readArgs );
			}
			catch( Exception )
			{
			}
			readArgs.Dispose();
		}
		#endregion

		#region protected virtual void OnConnectionLoss( ErrorArgs args )
		protected virtual void OnConnectionLoss( ErrorArgs args )
		{
			StopReading = true;
			DataReceivedEvent.Set();
			OutputDebug( "{0}: {1}", Name, args.ToString() );

			Socket = null;

			if( ConnectionLoss != null )
			{
				//	Make a copy to ensure thread safety
				EventHandler<ErrorArgs> temp = ConnectionLoss;
				temp( this, args );
			}
		}
		#endregion
		#region protected virtual void OnConnectCompleted( object sender, SocketAsyncEventArgs args )
		protected virtual void OnConnectCompleted( object sender, SocketAsyncEventArgs args )
		{
			if( args.SocketError == SocketError.Success )
			{
				OutputDebug( "{0}: Connection successful.", Name );

				EndPoint endPoint = args.RemoteEndPoint;
				if( endPoint is DnsEndPoint )
				{
					var dnsEndPoint = (DnsEndPoint)endPoint;
					HostAddress = dnsEndPoint.Host;
					Port = dnsEndPoint.Port;
				}
				else if( endPoint is IPEndPoint )
				{
					var ipEndPoint = (IPEndPoint)endPoint;
					HostAddress = ipEndPoint.Address.ToString();
					Port = ipEndPoint.Port;
				}

				ReceiveAsync();
			}
			else
			{
				OutputDebug( "{0}: Connection failed, {1}", Name, args.SocketError );
				ConnectError = new ErrorArgs( this, "Connect Error, ", args.SocketError );
				Socket = null;
			}

			ConnectEvent.Set();
		}
		#endregion
		#region protected virtual void OnReadCompleted( object sender, SocketAsyncEventArgs args )
		protected virtual void OnReadCompleted( object sender, SocketAsyncEventArgs args )
		{
			if( args.SocketError != SocketError.Success )
			{
				OutputDebug( "{0}: Read error, {1}", Name, args.SocketError );
				OnConnectionLoss( new ErrorArgs( this, "Read error, ", args.SocketError ) );
			}
			else if( args.BytesTransferred == 0 )
			{
				OutputDebug( "{0}: Socket closed by remote side.", Name );
				OnConnectionLoss( new ErrorArgs( this, "Socket closed by remote side." ) );
			}
			else
			{
				byte [] data = new byte[args.BytesTransferred];
				Array.Copy( args.Buffer, data, args.BytesTransferred );

				OutputDebug( "{0}: Rx: {1}", Name, BitConverter.ToString( data ).Replace( "-", " " ) );

				lock( Data )
				{
					Data.Add( data );
				}
				DataReceivedEvent.Set();
				ReceiveAsync();
			}
		}
		#endregion
		#region protected virtual void OnWriteCompleted( object sender, SocketAsyncEventArgs args )
		protected virtual void OnWriteCompleted( object sender, SocketAsyncEventArgs args )
		{
			if( args.SocketError != SocketError.Success )
			{
				OutputDebug( "{0}: Write failed, {1}", Name, args.SocketError );
				OnConnectionLoss( new ErrorArgs( this, "Write failed, ", args.SocketError ) );
			}
		}
		#endregion

		#region public bool Connect( EndPoint endPoint, bool waitForCompletion = true, SocketType socketType = SocketType.Stream, ProtocolType protocolType = ProtocolType.Tcp, TimeSpan? timeout = null )
		public bool Connect( EndPoint endPoint, bool waitForCompletion = true, SocketType socketType = SocketType.Stream, ProtocolType protocolType = ProtocolType.Tcp, TimeSpan? timeout = null )
		{
			ConnectError = null;
			HostAddress = null;
			Port = -1;

			SocketAsyncEventArgs args = new SocketAsyncEventArgs();
			args.RemoteEndPoint       = endPoint;
			args.UserToken            = this;
			args.Completed           += OnConnectCompleted;

			Socket = new Socket( endPoint.AddressFamily, socketType, protocolType );
			try
			{
				if( !Socket.ConnectAsync( args ) )
					OnConnectCompleted( this, args );
				else if( waitForCompletion )
					if( !ConnectEvent.WaitOne( timeout ?? TimeSpan.FromMinutes( 1 ) ) )
						throw new SocketException( (int)SocketError.TimedOut );
			}
			catch( SocketException e )
			{
				OutputDebug( "{0}: Connection failed, {1}", Name, SocketErrorString.FullError( e.ErrorCode ) );
				ConnectError = new ErrorArgs( this, "Connect Error, " + SocketErrorString.FullError( e.ErrorCode ), e );
			}
			catch( Exception e )
			{
				OutputDebug( "{0}: Connection failed, {1}", Name, e );
				ConnectError = new ErrorArgs( this, "Connect Error, ", e );
			}

			args.Dispose();
			if( ConnectError != null )
				Socket = null;
			return ConnectError == null;
		}
		#endregion
		#region public bool Connect( string hostAddress, int port, bool waitForCompletion = true, AddressFamily addressFamily = AddressFamily.InterNetwork, SocketType socketType = SocketType.Stream, ProtocolType protocolType = ProtocolType.Tcp, TimeSpan? timeout = null )
		public bool Connect( string hostAddress, int port, bool waitForCompletion = true, AddressFamily addressFamily = AddressFamily.InterNetwork, SocketType socketType = SocketType.Stream, ProtocolType protocolType = ProtocolType.Tcp, TimeSpan? timeout = null )
		{
			IPEndPoint endPoint = SimpleSocket.GetIPEndPoint( hostAddress, port, Logger, addressFamily == AddressFamily.InterNetworkV6 );
			return Connect( endPoint, waitForCompletion, socketType, protocolType, timeout );
		}
		#endregion
		#region public bool Connect( Uri uri, bool waitForCompletion = true, AddressFamily addressFamily = AddressFamily.InterNetwork, SocketType socketType = SocketType.Stream, ProtocolType protocolType = ProtocolType.Tcp,	TimeSpan? timeout = null )
		public bool Connect( Uri uri, bool waitForCompletion = true, AddressFamily addressFamily = AddressFamily.InterNetwork, SocketType socketType = SocketType.Stream, ProtocolType protocolType = ProtocolType.Tcp,	TimeSpan? timeout = null )
		{
			IPEndPoint endPoint = SimpleSocket.GetIPEndPoint( uri.DnsSafeHost, uri.Port, Logger, addressFamily == AddressFamily.InterNetworkV6 );
			return Connect( endPoint, waitForCompletion, socketType, protocolType, timeout );
		}
		#endregion

		#region public override int Read( byte[] buffer, int offset, int count )
		public override int Read( byte[] buffer, int offset, int count )
		{
			if( !Connected )
				return 0;

			while( !StopReading && ChunkCount == 0 )
				DataReceivedEvent.WaitOne( millisecondsTimeout: 250 );

			int totalRead = 0;
			lock( Data )
			{
				while( Data.Count != 0 )
				{
					byte[] data = Data[0];
				//	If the amount in this chunk is less than the maximum requested, copy it to the buffer and remove it.
					if( ( totalRead + data.Length ) <= count )
					{
						Array.Copy( data, 0, buffer, offset + totalRead, data.Length );
						totalRead += data.Length;
						Data.RemoveAt( 0 );
						if( totalRead == count )
							break;
					}
				//	Otherwise, copy enough to fill the buffer and make a new first chunk with what's left
					else
					{
						int remainderSize = count - totalRead;
						Array.Copy( data, 0, buffer, offset + totalRead, remainderSize );
						totalRead = count;

						byte[] remainder = new byte[data.Length - remainderSize];
						Array.Copy( data, remainderSize, remainder, 0, data.Length - remainderSize );
						Data[0] = remainder;
						break;
					}
				}
			}
			return totalRead;
		}
		#endregion
		#region public override void Write( byte[] buffer, int offset, int count )
		public override void Write( byte[] buffer, int offset, int count )
		{
			if( Socket == null || !Socket.Connected )
			{
				OutputDebug( "Write failed, socket null or disconnected." );
				return;
			}

			SocketAsyncEventArgs args = new SocketAsyncEventArgs();
			args.SetBuffer( buffer, offset, count );
			args.UserToken = this;
			args.Completed += OnWriteCompleted;

			try
			{
				OutputDebug( "{0}: Tx: {1}", Name, BitConverter.ToString( buffer ).Replace( "-", " " ) );
				if( !Socket.SendAsync( args ) )
					OnWriteCompleted( this, args );
			}
			catch( SocketException e )
			{
				OutputDebug( "{0}: Write failed, {1}", Name, SocketErrorString.FullError( e.ErrorCode ) );
				OnConnectionLoss( new ErrorArgs( this, "Write failed, " + SocketErrorString.FullError( e.ErrorCode ), e ) );
			}
			catch( Exception e )
			{
				OutputDebug( "{0}: Write failed, {1}", Name, e );
				OnConnectionLoss( new ErrorArgs( this, "Write failed, ", e ) );
			}

			args.Dispose();
		}
		#endregion

		#region public void OutputDebug( string format, params object[] list )
		public void OutputDebug( string format, params object[] list )
		{
			if( Logger != null )
				Logger.Write( String.Format( format, list ) );
		}
		#endregion	
	}
}
