﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

namespace IntelliSite.CommonUtil
{
	///<summary>
	///	This is an abstract class is used to manage a data stream.  It can handle data coming in from a source in chunks and allow the 
	///</summary>
	public abstract class PeekableStream : Stream
	{
		/// <summary>
		/// Create a PeekableStream with a name and encoding
		/// <param name="name">The name of the stream</param>
		/// <param name="encoding">The encoding of the stream</param>
		/// </summary>
		public PeekableStream( string name = "PeekableStream", Encoding encoding = null )
		{
			Name = name;
			Encoding = encoding ?? DefaultEncoding;
		}

		public string Name { get; set; }

		public abstract bool SupportsMultipleChunks { get; }
		public abstract int ChunkCount { get; }
		public abstract int ChunkSize( int index );
		public abstract byte[] PeekChunk( int index, int offset, int size );

		public virtual Encoding DefaultEncoding { get { return Encoding.UTF8; } }
		public virtual Encoding Encoding { get; set; }

		public byte[] PeekChunk( int index ) { return PeekChunk( index, 0, ChunkSize( index ) ); }

		public void Write( byte[] buffer )
		{
			Write( buffer, 0, buffer.Length );
		}

		public void Write( string contents, Encoding encoding = null )
		{
			byte[] buffer = ( encoding ?? Encoding ).GetBytes( contents );
			Write( buffer, 0, buffer.Length );
		}

		#region Length
		/// <summary>
		/// Gets the total number of available bytes in the DataStream
		/// </summary>
		public override long Length
		{
			get
			{
				int totalBytes = 0;
				for( int i=0; i < ChunkCount; i++ )
					totalBytes += ChunkSize( i );
				return totalBytes;
			}
		}
		#endregion

		#region Peek Method
		///	<summary>public void Peek( int offset, int index, int size, out byte[] data, out in calcOffset )
		/// <para>
		///	This function returns an byte array and offset in that array that contains a certain number of bytes of data.
		///	The byte array and index are then used in by the BitConverter class to return an object of a specific type.
		///
		///	This function will throw an IndexOutOfRangeException if there is not enough data to fill the request,
		///	otherwise the byte array and calculated offset are returned in data and calcOffset.
		///	</para>
		///</summary>
		public ArraySegment<byte> Peek( int offset, int index, int size )
		{
			//	If there is no data, throw an exception
			if( ChunkCount == 0 )
				throw new IndexOutOfRangeException();

			//	Determine the starting byte for the data required, then search for the chunk containing the start
			int start = offset + ( index * size );
			for( int i=0; i < ChunkCount; i++ )
			{
				//	If the start is within this chunk (remember the chunk size in case the implementation of the chunk size is slow like File I/O).
				int chunkSize = ChunkSize( i );
				if( start < chunkSize )
				{
					//	If the end is also within this chunk, return the chunk and the offset.
					if( ( start + size ) <= chunkSize )
					{
						if( SupportsMultipleChunks )
							return new ArraySegment<byte>( PeekChunk( i ), start, size );
						else
							return new ArraySegment<byte>( PeekChunk( 0, start, size ), 0, size );
					}
					//	The requested data spans two or more chunks.
					//	Build a byte array that contains the requested data from multiple chunks.
					else if( SupportsMultipleChunks )
					{
						byte[] data = new byte[size];
						int len = 0;
						while( len < size )
						{
							int thisChunkSize = ChunkSize( i );
							//	Determine how much data I can copy from the current chunk
							int copyLen = size - len;
							if( copyLen < ( thisChunkSize - start ) )
								copyLen = thisChunkSize - start;

							//	Copy the data into the new array
							byte[] chunk = PeekChunk( i );
							Array.Copy( chunk, start, data, len, ChunkSize( i ) - start );
							start = 0;
							len += copyLen;

							//	If I've copied enough, return the copy and an offset of 0
							if( len == size )
								return new ArraySegment<byte>( data, 0, size );

							//	Get the next chunk.  If we're out, throw an exception
							i++;
							if( i >= ChunkCount )
								throw new IndexOutOfRangeException();
						}
					}
					//	I found the start within this chunk, but there's not enough data, throw an exception
					else
						throw new IndexOutOfRangeException();
				}
				//	Reduce start index by the length of this chunk, then go on to the next chunk
				else
					start -= ChunkSize( i );
			}

			throw new IndexOutOfRangeException();
		}
		#endregion

		#region Typed Peek Methods
		public char PeekChar( int offset, int index, Encoding encoding = null )
		{
			ArraySegment<byte> segment = Peek( offset, index, sizeof( char ) );
			char[] chars = ( encoding ?? Encoding ).GetChars( segment.Array, segment.Offset, 1 );
			if( chars.Length <= 0 )
				throw new ArgumentOutOfRangeException( "Not enough characters" );
			return chars[0];
		}

		public Int16 PeekInt16( int offset, int index )
		{
			ArraySegment<byte> segment = Peek( offset, index, sizeof( Int16 ) );
			return BitConverter.ToInt16( segment.Array, segment.Offset );
		}

		public UInt16 PeekUInt16( int offset, int index )
		{
			ArraySegment<byte> segment = Peek( offset, index, sizeof( UInt16 ) );
			return BitConverter.ToUInt16( segment.Array, segment.Offset );
		}

		public Int32 PeekInt32( int offset, int index )
		{
			ArraySegment<byte> segment = Peek( offset, index, sizeof( Int32 ) );
			return BitConverter.ToInt32( segment.Array, segment.Offset );
		}

		public UInt32 PeekUInt32( int offset, int index )
		{
			ArraySegment<byte> segment = Peek( offset, index, sizeof( UInt32 ) );
			return BitConverter.ToUInt32( segment.Array, segment.Offset );
		}

		public Int64 PeekInt64( int offset, int index )
		{
			ArraySegment<byte> segment = Peek( offset, index, sizeof( Int64 ) );
			return BitConverter.ToInt64( segment.Array, segment.Offset );
		}

		public UInt64 PeekUInt64( int offset, int index )
		{
			ArraySegment<byte> segment = Peek( offset, index, sizeof( UInt64 ) );
			return BitConverter.ToUInt64( segment.Array, segment.Offset );
		}

		public float PeekSingle( int offset, int index )
		{
			ArraySegment<byte> segment = Peek( offset, index, sizeof( float ) );
			return BitConverter.ToSingle( segment.Array, segment.Offset );
		}

		public double PeekDouble( int offset, int index )
		{
			ArraySegment<byte> segment = Peek( offset, index, sizeof( double ) );
			return BitConverter.ToDouble( segment.Array, segment.Offset );
		}
		#endregion

		#region Find Methods
		public void Find( int offset, byte find, out int index, out int chunkOffset )
		{
			byte[] findArray = new byte[] { find };
			Find( offset, findArray, out index, out chunkOffset );
		}

		public void Find( int offset, char find, out int index, out int chunkOffset, Encoding encoding = null )
		{
			byte[] findArray = ( encoding ?? Encoding ).GetBytes( new char[] { find } );
			Find( offset, findArray, out index, out chunkOffset );
		}

		public void Find( int offset, string find, out int index, out int chunkOffset, Encoding encoding = null )
		{
			byte[] findArray = ( encoding ?? Encoding ).GetBytes( find );
			Find( offset, findArray, out index, out chunkOffset );
		}

		public void Find( int offset, byte[] find, out int index, out int chunkOffset )
		{
			index = -1;
			chunkOffset = 0;
			if( ChunkCount == 0 )
				return;

			if( SupportsMultipleChunks )
			{
				//	Find the correct starting chunk based on the offset
				int chunkIndex = 0;
				while( chunkIndex < ChunkCount && offset >= ChunkSize( chunkIndex ) )
					offset -= ChunkSize( chunkIndex++ );

				while( chunkIndex < ChunkCount )
				{
					byte[] chunk = PeekChunk( chunkIndex );

					//	Find the first byte in the find array
					int startIndex = Array.IndexOf( chunk, find[0], offset );
					if( startIndex >= 0 )
					{
						bool foundIt = true;

						//	Check to see if the rest of the bytes match too
						int thisChunkOffset = startIndex + 1;
						int thisChunkIndex  = chunkIndex;
						byte[] thisChunk   = chunk;
						int j = 1;
						while( foundIt && j < find.Length )
						{

							if( thisChunkOffset >= thisChunk.Length )
							{
								thisChunkIndex++;
								if( thisChunkIndex >= ChunkCount )
									return;

								thisChunkOffset = 0;
								thisChunk = PeekChunk( thisChunkIndex );
							}

							foundIt = thisChunk[thisChunkOffset++] == find[j++];
						}

						//	If so, return it
						if( foundIt && j == find.Length )
						{
							index = chunkIndex;
							chunkOffset = startIndex;
							return;
						}
					}

					offset = 0;
					chunkIndex++;
				}
			}
			//	If we don't support multiple chunks, peek in small pieces
			else
			{
				int chunkSize = ChunkSize( 0 );

				int peekSize = 256;
				if( peekSize < find.Length )
					peekSize = find.Length * 2;
				else if( ( peekSize + offset ) > chunkSize )
					peekSize = chunkSize - offset;

				byte[] chunk = PeekChunk( 0, offset, peekSize );
				byte[] nextChunk = null;
				int nextSize = Math.Min( peekSize, chunkSize - peekSize - offset );
				if( nextSize > 0 )
					nextChunk = PeekChunk( 0, offset + peekSize + offset, peekSize );

				int thisOffset = offset;
				while( thisOffset < chunkSize )
				{
					int startIndex = Array.IndexOf( chunk, find[0] );
					if( startIndex >= 0 )
					{
						bool foundIt = true;
						int remaining = peekSize - startIndex - 1;
						int j = 1;
						for( ; foundIt && j < find.Length && j <= remaining; j++ )
						{
							if( ( startIndex + j ) < peekSize )
								foundIt = chunk[startIndex + j] == find[j];
							else if( nextChunk != null && ( ( j - remaining ) < nextChunk.Length ) )
								foundIt = nextChunk[j - remaining] == find[j];
							else
								return;
						}

						if( foundIt && ( find.Length == 1 || j == find.Length ) )
						{
							index = 0;
							chunkOffset = thisOffset + startIndex;
							return;
						}
					}

					thisOffset += peekSize;
					chunk = nextChunk;
					if( ( thisOffset + peekSize ) < chunkSize )
						nextChunk = PeekChunk( 0, thisOffset + peekSize, peekSize );
					else if( thisOffset < chunkSize )
						nextChunk = PeekChunk( 0, thisOffset + peekSize, chunkSize - thisOffset );
					else
						nextChunk = null;
				}
			}
		}
		#endregion
	}
}
