#if DEBUG
//#define LogAssemblyResolving
#endif

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using IntelliSite.CommonUtil.Extensions;
using Microsoft.Win32;

namespace IntelliSite.CommonUtil
{
	public class ProductInfo : ICloneable
	{
		public static bool LoadFromBrandedRegistry = true;
		public static string UnbrandedCompanyName = "OSSI";
		public static string UnbrandedProductName = "Intelli-Site";

		public string CompanyName { get; protected set; }
		public string ProductName { get; protected set; }
		public string Brand { get; protected set; }
		public string Copyright { get; protected set; }
		public string RegistryBase { get; protected set; }
		public RegistryHive RegistryHive { get; protected set; }
		public RegistryView RegistryView { get; protected set; }
		public bool IsLiteVersion { get; protected set; }
		public string Version { get; protected set; }

		public string InstallPath { get; protected set; }
		public string DataFolder { get; protected set; }
		public string LogsFolder { get; protected set; }
		public string SetupPackageCache { get; protected set; }

		public string LocalUserDataFolder { get; protected set; }
		public string RoamingUserDataFolder { get; protected set; }
		public string AllProjectsFolder { get; protected set; }
		public string CurrentProjectFolder { get; protected set; }

		public static string DefaultAllProjectsFolderName { get { return "Sites"; } }
		public static string DefaultLogsFolderName { get { return "Logs"; } }

		private static AssemblyName _thisAssemblyName;
		private static byte[] _thisAssemblyPublicKeyToken;

		public ProductInfo()
			: this( Current )
		{
		}

		public ProductInfo( ProductInfo source )
		{
			if( source == null )
				return;

			this.CompanyName           = source.CompanyName;
			this.ProductName           = source.ProductName;
			this.Brand                 = source.Brand;
			this.RegistryBase          = source.RegistryBase;
			this.RegistryHive          = source.RegistryHive;
			this.RegistryView          = source.RegistryView;
			this.IsLiteVersion         = source.IsLiteVersion;
			this.InstallPath           = source.InstallPath;
			this.Version               = source.Version;
			this.DataFolder            = source.DataFolder;
			this.LogsFolder            = source.LogsFolder;
			this.SetupPackageCache     = source.SetupPackageCache;
			this.LocalUserDataFolder   = source.LocalUserDataFolder;
			this.RoamingUserDataFolder = source.RoamingUserDataFolder;
			this.AllProjectsFolder     = source.AllProjectsFolder;
			this.CurrentProjectFolder  = source.CurrentProjectFolder;
		}

		public ProductInfo( string companyName, string productName, RegistryView? view = null, RegistryHive? hive = null, string registryBase = null, string copyright = null, bool loadRegistryValues = true )
		{
			const string brand = null;
			Init( companyName, productName, brand, view, hive, registryBase, copyright, loadRegistryValues );
		}

		public ProductInfo( string companyName, string productName, string brand, RegistryView? view = null, RegistryHive? hive = null, string registryBase = null, string copyright = null, bool loadRegistryValues = true )
		{
			Init( companyName, productName, brand, view, hive, registryBase, copyright, loadRegistryValues );
		}

		object ICloneable.Clone()
		{
			return Clone();
		}

		public ProductInfo Clone()
		{
			return new ProductInfo( this );
		}

		private void Init( string companyName, string productName, string brand, RegistryView? view = null, RegistryHive? hive = null, string registryBase = null, string copyright = null, bool loadRegistryValues = true )
		{
			CompanyName   = companyName;
			ProductName   = productName;
			Brand         = brand ?? productName;
			RegistryBase  = registryBase ?? ( LoadFromBrandedRegistry ? Path.Combine( "Software", companyName, productName ) :  Path.Combine( "Software", UnbrandedCompanyName, UnbrandedProductName ) );
			RegistryHive  = hive ?? RegistryHive.LocalMachine;
			RegistryView = view ?? RegistryView.Default;
			Copyright     = copyright ?? "Copyright © OSSI, LLC. All rights reserved.";
			IsLiteVersion = false;

			if( loadRegistryValues )
			{
				try
				{
					ReadProductKey( action: LoadProductInfo );
				}
				catch( Exception ) { }
			}

			if( InstallPath == null )
			{
				var entryAssembly = Assembly.GetEntryAssembly();
				if( entryAssembly != null )
					InstallPath = Path.GetDirectoryName( entryAssembly.Location );
				else
					InstallPath = AppDomain.CurrentDomain.BaseDirectory;
			}

			if( DataFolder == null )
				DataFolder = Path.Combine( Environment.GetFolderPath( Environment.SpecialFolder.CommonApplicationData ), CompanyName, ProductName );

			if( LogsFolder == null )
				LogsFolder = Path.Combine( DataFolder, DefaultLogsFolderName );

			if( LocalUserDataFolder == null )
				LocalUserDataFolder = Path.Combine( Environment.GetFolderPath( Environment.SpecialFolder.LocalApplicationData ), CompanyName, ProductName );

			if( RoamingUserDataFolder == null )
				RoamingUserDataFolder = Path.Combine( Environment.GetFolderPath( Environment.SpecialFolder.ApplicationData ), CompanyName, ProductName );

			if( AllProjectsFolder == null )
				AllProjectsFolder = Path.Combine( DataFolder, DefaultAllProjectsFolderName );
		}

		public override string ToString()
		{
			StringBuilder sb = new StringBuilder();
			sb.Append( "{ProductInfo:" );
			sb.Append( " " ).Append( CompanyName );
			sb.Append( " " ).Append( ProductName );
			sb.Append( " (Version=" ).Append( Version ).Append( ")" );
			sb.Append( "}" );

			return sb.ToString();
		}

		private ICollection<string> _installedComponents = null;
		public ICollection<string> InstalledComponents
		{
			get
			{
				if( _installedComponents == null )
					_installedComponents = GetSetupInstalledGroupItems( "Components" );

				return _installedComponents;
			}
		}

		private ICollection<string> _installedDatabaseProviders = null;
		public ICollection<string> InstalledDatabaseProviders
		{
			get
			{
				if( _installedDatabaseProviders == null )
					_installedDatabaseProviders = GetSetupInstalledGroupItems( "Database Providers" );

				return _installedDatabaseProviders;
			}
		}

		private ICollection<string> _installedDrivers = null;
		public ICollection<string> InstalledDrivers
		{
			get
			{
				if( _installedDrivers == null )
					_installedDrivers = GetSetupInstalledGroupItems( "Drivers" );

				return _installedDrivers;
			}
		}

		public IEnumerable<Assembly> GetAssemblies( bool currentAppDomain = true, bool loadFromInstallPath = true, bool loadInstalled = true, bool throwLoadExceptions = false )
		{
			HashSet<string> returnedAssemblies = new HashSet<string>( StringComparer.InvariantCultureIgnoreCase );
			Queue<Assembly> loadedAssemblies = new Queue<Assembly>();

			AssemblyLoadEventHandler onAssemblyLoad = ( sender, args ) =>
				{
					lock( loadedAssemblies )
						loadedAssemblies.Enqueue( args.LoadedAssembly );
				};

			Assembly[] initialAssemblies = AppDomain.CurrentDomain.GetAssemblies();
			AppDomain.CurrentDomain.AssemblyLoad += onAssemblyLoad;

			if( currentAppDomain )
			{
				foreach( Assembly loadedAssembly in initialAssemblies )
				{
					string simpleName = loadedAssembly.GetName().Name;
					if( !returnedAssemblies.Contains( simpleName ) )
					{
						returnedAssemblies.Add( simpleName );
						yield return loadedAssembly;
					}
				}
			}

			Func<string, Func<string, Assembly>, Assembly> loadItemAssembly = ( itemName, loadMethod ) =>
			{
				Assembly assembly;
				try
				{
					assembly = loadMethod( itemName );
				}
				catch( Exception )
				{
					assembly = null;
					if( throwLoadExceptions )
					{
						AppDomain.CurrentDomain.AssemblyLoad -= onAssemblyLoad;
						throw;
					}
				}

				if( assembly != null )
				{
					string simpleName = assembly.GetName().Name;
					if( !returnedAssemblies.Contains( simpleName ) )
					{
						returnedAssemblies.Add( simpleName );
						return assembly;
					}
				}

				return null;
			};

			DirectoryInfo installPath = null;
			if( loadFromInstallPath && ( installPath = new DirectoryInfo( InstallPath ) ).Exists )
			{
				foreach( FileInfo installFile in installPath.GetFiles() )
				{
					if( Regex.IsMatch( installFile.Name, @".+\.dll", RegexOptions.CultureInvariant | RegexOptions.IgnoreCase ) )
					{
						Assembly assembly = loadItemAssembly( installFile.FullName, assemblyFile => LoadAssemblyFromFile( assemblyFile ) );
						if( assembly != null )
							yield return assembly;
					}
				}
			}

			if( loadInstalled )
			{
				foreach( string driverName in InstalledDrivers )
				{
					Assembly assembly = loadItemAssembly( driverName, name => LoadDriver( name ) );
					if( assembly != null )
						yield return assembly;
				}

				foreach( string databaseProviderName in InstalledDatabaseProviders )
				{
					Assembly assembly = loadItemAssembly( databaseProviderName, name => LoadDatabaseProvider( name ) );
					if( assembly != null )
						yield return assembly;
				}
			}

			AppDomain.CurrentDomain.AssemblyLoad -= onAssemblyLoad;
			while( loadedAssemblies.Count > 0 )
			{
				Assembly assembly = loadedAssemblies.Dequeue();
				string simpleName;

				if( assembly == null || returnedAssemblies.Contains( ( simpleName = assembly.GetName().Name ) ) )
					continue;

				returnedAssemblies.Add( simpleName );
				yield return assembly;
			}
		}

		public Assembly GetLoadedAssemblyFromSimpleName( string simpleName, StringComparison comparisonType = StringComparison.InvariantCultureIgnoreCase )
		{
			foreach( Assembly assembly in AppDomain.CurrentDomain.GetAssemblies() )
			{
				string assemblySimpleName = assembly.GetName().Name;
				if( assemblySimpleName.Equals( simpleName, comparisonType ) )
					return assembly;
			}

			return null;
		}

		private ICollection<string> GetSetupInstalledGroupItems( string groupName )
		{
			var resultItems = new HashSet<string>();

			ReadProductKey( subkey: Path.Combine( "Setup", groupName ), action: groupKey =>
			{
				if( groupKey != null )
				{
					foreach( string itemName in groupKey.GetSubKeyNames() )
					{
						using( RegistryKey itemKey = groupKey.OpenSubKey( itemName ) )
						{
							if( itemKey != null && (int)itemKey.GetValue( "Installed", 0 ) != 0 )
								resultItems.Add( itemName );

							foreach( string subKeyName in itemKey.GetSubKeyNames() )
								using( RegistryKey subKey = itemKey.OpenSubKey( subKeyName ) )
								{
									if( subKey != null && (int)subKey.GetValue( "Installed", 0 ) != 0 )
										resultItems.Add( itemName + "." + subKeyName );
								}
						}
					}
				}
			} );

			return resultItems;
		}

		public ProductComponentInfo GetComponentInfo( string componentName )
		{
			return ReadComponentSetupKey( componentName, key =>
				key != null ? new ProductComponentInfo( "Components", componentName, key ) : null );
		}

		public ProductComponentInfo GetDatabaseProviderInfo( string providerName )
		{
			return ReadComponentSetupKey( providerName, key =>
				key != null ? new ProductComponentInfo( "Database Providers", providerName, key ) : null );
		}

		public ProductDriverInfo GetDriverInfo( string driverName )
		{
			return ReadComponentSetupKey( driverName, key =>
				key != null ? new ProductDriverInfo( "Drivers", driverName, key ) : null );
		}

		public bool GetIsComponentInstalled( string componentName )
		{
			return InstalledComponents.Contains( componentName );
		}

		public bool GetIsDatabaseProviderInstalled( string providerName )
		{
			return InstalledDatabaseProviders.Contains( providerName );
		}

		public bool GetIsDriverInstalled( string driverName )
		{
			return InstalledDrivers.Contains( driverName );
		}

		public string GetSubkeyPath( string subkey )
		{
			return Path.Combine( RegistryBase, subkey );
		}

		public static bool IsProductAssembly( AssemblyName assemblyName )
		{
			if( assemblyName == null )
				throw new ArgumentNullException( "assemblyName" );

			if( _thisAssemblyPublicKeyToken == null )
			{
				if( _thisAssemblyName == null )
					_thisAssemblyName = Assembly.GetExecutingAssembly().GetName();
				_thisAssemblyPublicKeyToken = _thisAssemblyName.GetPublicKeyToken();
			}

			byte[] publicKeyToken = assemblyName.GetPublicKeyToken();
			return publicKeyToken != null && publicKeyToken.SequenceEqual( _thisAssemblyPublicKeyToken );
		}

		public Assembly LoadAssemblyFromFile( string assemblyFile, bool reloadExisting = false )
		{
			if( assemblyFile == null )
				throw new ArgumentNullException( "assemblyFile" );
			else if( String.IsNullOrEmpty( assemblyFile ) )
				throw new ArgumentOutOfRangeException( "assemblyFile", "File name cannot be empty." );

			AssemblyName assemblyName = AssemblyName.GetAssemblyName( assemblyFile );
			if( assemblyName != null )
			{
				Assembly assembly = null;
				if( !reloadExisting && ( assembly = GetLoadedAssemblyFromSimpleName( assemblyName.Name ) ) != null )
					return assembly;

				return Assembly.LoadFrom( assemblyFile );
			}

			return null;
		}

		public Assembly LoadAssemblyFromName( string assemblyName, ResolveProductAssemblyOptions options = ResolveProductAssemblyOptions.Default )
		{
			if( assemblyName == null )
				throw new ArgumentNullException( "assemblyName" );
			return ResolveProductAssembly( new AssemblyName( assemblyName ), options );
		}

		public Assembly LoadAssemblyFromName( AssemblyName assemblyName, ResolveProductAssemblyOptions options = ResolveProductAssemblyOptions.Default )
		{
			if( assemblyName == null )
				throw new ArgumentNullException( "assemblyName" );
			return ResolveProductAssembly( assemblyName, options );
		}

		public bool AssemblyNameMatches( AssemblyName test, AssemblyName target, bool allowLaterVersion = true )
		{
			if( !AssemblyName.ReferenceMatchesDefinition( target, test ) )
				return false;

			if( target.ProcessorArchitecture == ProcessorArchitecture.None )
			{
				switch( SystemInfo.Current.ProcessorArchitecture )
				{
					case ProcessorArchitecture.Amd64:
						if( test.ProcessorArchitecture != ProcessorArchitecture.Amd64 &&
							test.ProcessorArchitecture != ProcessorArchitecture.MSIL )
							return false;
						break;

					case ProcessorArchitecture.X86:
						if( test.ProcessorArchitecture != ProcessorArchitecture.X86 &&
							test.ProcessorArchitecture != ProcessorArchitecture.MSIL )
							return false;
						break;

					case ProcessorArchitecture.MSIL:
						if( test.ProcessorArchitecture != ProcessorArchitecture.MSIL )
							return false;
						break;

					default:
						return false;
				}
			}
			else if( test.ProcessorArchitecture != target.ProcessorArchitecture )
				return false;

			Version targetVersion = target.Version;
			if( targetVersion != null && ( test.Version < targetVersion || !allowLaterVersion && test.Version != targetVersion ) )
				return false;

			return true;
		}

		private Assembly LoadAssemblyFromItemKey( RegistryKey itemKey )
		{
			if( itemKey == null )
				throw new ArgumentNullException( "itemKey" );

			Assembly assembly = null;

			string assemblyPath = (string)itemKey.GetValue( "AssemblyPath" );
			if( !String.IsNullOrEmpty( assemblyPath ) && File.Exists( assemblyPath ) )
				assembly = LoadAssemblyFromFile( assemblyPath );

			if( assembly == null )
			{
				string assemblyNameString = (string)itemKey.GetValue( "AssemblyName" );
				if( !String.IsNullOrEmpty( assemblyNameString ) )
				{
					// check all default locations except for installed item keys
					ResolveProductAssemblyOptions options = (
						ResolveProductAssemblyOptions.Default &
						~ResolveProductAssemblyOptions.CheckAllInstalledItems );
					assembly = LoadAssemblyFromName( new AssemblyName( assemblyNameString ), options );
				}
			}

			return assembly;
		}

		public Assembly LoadDatabaseProvider( string databaseProviderName )
		{
			Assembly result = null;
			if( InstalledDatabaseProviders.Contains( databaseProviderName ) )
			{
				result = ReadDatabaseProviderSetupKey<Assembly>( databaseProviderName, LoadAssemblyFromItemKey );
				if( result != null )
					return result;
			}

			string path;
			string filename = "ISDatabase." + databaseProviderName + ".dll";
			if( File.Exists( path = Path.Combine( AppDomain.CurrentDomain.BaseDirectory, filename ) ) )
				return Assembly.LoadFile( path );
			else if( File.Exists( path = Path.Combine( InstallPath, filename ) ) )
				return Assembly.LoadFile( path );

			return null;
		}

		public Assembly LoadCommMethod( string commMethodName, bool loadInstalled = true )
		{
			Assembly result = null;
			if( loadInstalled && InstalledDrivers.Contains( commMethodName ) )
			{
				result = ReadDriverSetupKey<Assembly>( commMethodName, LoadAssemblyFromItemKey );
				if( result != null )
					return result;
			}

			string path;
			string filename = "IntelliSite.ISC" + commMethodName + ".dll";
			if( File.Exists( path = Path.Combine( AppDomain.CurrentDomain.BaseDirectory, filename ) ) )
				return Assembly.LoadFile( path );
			else if( File.Exists( path = Path.Combine( InstallPath, filename ) ) )
				return Assembly.LoadFile( path );

			filename = "ISC" + commMethodName + ".dll";
			if( File.Exists( path = Path.Combine( AppDomain.CurrentDomain.BaseDirectory, filename ) ) )
				return Assembly.LoadFile( path );
			else if( File.Exists( path = Path.Combine( InstallPath, filename ) ) )
				return Assembly.LoadFile( path );

			return null;
		}

		public Assembly LoadDriver( string driverName, bool loadInstalled = true )
		{
			Assembly result = null;
			if( loadInstalled && InstalledDrivers.Contains( driverName ) )
			{
				result = ReadDriverSetupKey<Assembly>( driverName, LoadAssemblyFromItemKey );
				if( result != null )
					return result;
			}

			string path;
			string filename = "ISD" + driverName + ".dll";
			if( File.Exists( path = Path.Combine( AppDomain.CurrentDomain.BaseDirectory, filename ) ) )
				return Assembly.LoadFile( path );
			else if( File.Exists( path = Path.Combine( InstallPath, filename ) ) )
				return Assembly.LoadFile( path );

			return null;
		}

		protected Assembly AppDomain_OnAssemblyResolve( object sender, ResolveEventArgs args )
		{
			return ResolveProductAssembly( new AssemblyName( args.Name ) );
		}

		public Assembly ResolveProductAssembly( AssemblyName assemblyName, ResolveProductAssemblyOptions options = ResolveProductAssemblyOptions.Default )
		{
			if( assemblyName == null )
				throw new ArgumentNullException( "assemblyName" );

			bool isProductAssembly = IsProductAssembly( assemblyName );

			bool allowLaterVersion;
			if( isProductAssembly )
				allowLaterVersion = options.HasFlag( ResolveProductAssemblyOptions.AllowLaterVersionForProductAssemblies );
			else
				allowLaterVersion = options.HasFlag( ResolveProductAssemblyOptions.AllowLaterVersionForOtherAssemblies );

			bool loadHighestVersion = isProductAssembly && options.HasFlag( ResolveProductAssemblyOptions.LoadHighestVersionForProductAssemblies );

			Assembly assembly;
			string assemblyPath;
			string assemblyNameString;
			string versionString;

			var candidatesWithLoadMethods = new List<Tuple<AssemblyName, Func<Assembly>>>();

			Action<AssemblyName, Func<Assembly>> addAssemblyCandidate = ( name, loadMethod ) =>
				candidatesWithLoadMethods.Add( new Tuple<AssemblyName, Func<Assembly>>( name, loadMethod ) );

			try
			{
				if( options.HasFlag( ResolveProductAssemblyOptions.CheckAppDomainLoaded ) )
				{
					List<Assembly> assemblies = AppDomain.CurrentDomain.GetAssemblies()
						.Where( asm => AssemblyNameMatches( asm.GetName( copiedName: false ), assemblyName, allowLaterVersion ) )
						.OrderByDescending( asm => asm.GetName( copiedName: false ).Version )
						.ToList();

					int matchCount = assemblies.Count;
					if( matchCount > 0 )
					{
						assembly = assemblies[0];
#if DEBUG && LogAssemblyResolving
						System.Diagnostics.Debug.WriteLine( "Resolving assembly \"{0}\" to loaded assembly: MatchCount={1}, Assembly={2}, CodeBase={3}",
							assemblyName, matchCount, assembly, assembly.CodeBase );
#endif
						return assembly;
					}
				}

				if( options.HasFlag( ResolveProductAssemblyOptions.CheckInstallPath ) && !String.IsNullOrEmpty( InstallPath ) )
				{
					assemblyPath = Path.Combine( InstallPath, assemblyName.Name + ".dll" );

					if( File.Exists( assemblyPath ) )
					{
						AssemblyName checkAssemblyName = AssemblyName.GetAssemblyName( assemblyPath );
						if( AssemblyNameMatches( checkAssemblyName, assemblyName, allowLaterVersion ) )
						{
							string installAssemblyPath = assemblyPath;
							addAssemblyCandidate( checkAssemblyName, () => Assembly.LoadFrom( installAssemblyPath ) );
						}
					}
				}

				Action<RegistryKey> loadItemInfoFromRegistry = key =>
				{
					if( key != null && key.GetValue<int>( "Installed", 0 ) != 0 )
					{
						assemblyNameString = key.GetValue<string>( "AssemblyName" );
						assemblyPath = key.GetValue<string>( "AssemblyPath" );
						versionString = key.GetValue<string>( "Version" );
					}
				};

				List<Action> loadItemInfoActions = new List<Action>();

				if( options.HasFlag( ResolveProductAssemblyOptions.CheckInstalledComponents ) )
					loadItemInfoActions.Add( () => ReadComponentSetupKey( assemblyName.Name, loadItemInfoFromRegistry ) );

				if( options.HasFlag( ResolveProductAssemblyOptions.CheckInstalledDrivers ) )
					loadItemInfoActions.Add( () => ReadDriverSetupKey( assemblyName.Name, loadItemInfoFromRegistry ) );

				if( options.HasFlag( ResolveProductAssemblyOptions.CheckInstalledDatabaseProviders ) )
					loadItemInfoActions.Add( () => ReadDatabaseProviderSetupKey( assemblyName.Name, loadItemInfoFromRegistry ) );

				foreach( Action loadItemInfoAction in loadItemInfoActions )
				{
					assemblyPath = null;
					assemblyNameString = null;
					versionString = null;

					loadItemInfoAction();

					try
					{
						if( !String.IsNullOrWhiteSpace( assemblyPath ) && File.Exists( assemblyPath ) )
						{
							string itemPath = assemblyPath;
							AssemblyName itemAssemblyName = AssemblyName.GetAssemblyName( assemblyPath );
							if( AssemblyNameMatches( itemAssemblyName, assemblyName, allowLaterVersion ) )
								addAssemblyCandidate( itemAssemblyName, () => Assembly.LoadFrom( itemPath ) );
						}
						else if( !String.IsNullOrWhiteSpace( assemblyNameString ) )
						{
							string itemPath = Fusion.FusionWrapper.GetAssemblyPath( assemblyNameString );
							AssemblyName itemAssemblyName = AssemblyName.GetAssemblyName( itemPath );
							if( AssemblyNameMatches( itemAssemblyName, assemblyName, allowLaterVersion ) )
								addAssemblyCandidate( itemAssemblyName, () => Assembly.Load( itemAssemblyName ) );
						}
					}
#if DEBUG
					catch( Exception ex )
					{
						System.Diagnostics.Debug.WriteLine( "Exception while resolving installed item assembly \"{0}\": {1}", assemblyName, ex );
					}
#else
					catch { }
#endif
				}

				if( options.HasFlag( ResolveProductAssemblyOptions.CheckGlobalAssemblyCache ) )
				{
					AssemblyName enumFromAssemblyName = (AssemblyName)assemblyName.Clone();
					if( allowLaterVersion )
						enumFromAssemblyName.Version = null;

					foreach( AssemblyName gacAssemblyName in Fusion.FusionWrapper.EnumCachedAssemblyNames( enumFromAssemblyName.FullName, parseDisplayName: true, globalAssemblyCache: true ) )
						if( AssemblyNameMatches( gacAssemblyName, assemblyName, allowLaterVersion ) )
						{
							AssemblyName itemAssemblyName = gacAssemblyName;
							addAssemblyCandidate( itemAssemblyName, () => Assembly.Load( itemAssemblyName ) );
						}
				}

				if( candidatesWithLoadMethods.Count > 0 )
				{
					foreach( Tuple<AssemblyName, Func<Assembly>> candidate in candidatesWithLoadMethods.OrderByDescending( item => item.Item1.Version ) )
					{
						AssemblyName candidateAssemblyName = candidate.Item1;
						Func<Assembly> candidateLoadMethod = candidate.Item2;

						try
						{
#if DEBUG && LogAssemblyResolving
							System.Diagnostics.Debug.WriteLine( "Resolving assembly \"{0}\" => \"{1}\"...", assemblyName, candidateAssemblyName );
#endif

							assembly = candidateLoadMethod();

#if DEBUG && LogAssemblyResolving
							System.Diagnostics.Debug.WriteLine( "Resolve assembly \"{0}\" to: Assembly={1}, CodeBase={2}",
								assemblyName, assembly, assembly != null ? assembly.CodeBase : null );
#endif

							if( assembly != null )
								return assembly;
						}
#if DEBUG
						catch( Exception ex )
						{
							System.Diagnostics.Debug.WriteLine( "Error loading candidate assembly \"{0}\": {1}", candidateAssemblyName, ex );
						}
#else
						catch { }
#endif
					}
				}
			}
#if DEBUG
			catch( Exception ex )
			{
				System.Diagnostics.Debug.WriteLine( "Exception while resolving assembly \"{0}\": {1}", assemblyName, ex );
				throw;
			}
#else
			catch
			{
				throw;
			}
#endif

			return null;
		}

		public void ReadComponentSetupKey( string componentName, Action<RegistryKey> action )
		{
			ReadProductKey( Path.Combine( @"Setup\Components", componentName ), action );
		}

		public TResult ReadComponentSetupKey<TResult>( string componentName, Func<RegistryKey, TResult> func )
		{
			return ReadProductKey( Path.Combine( @"Setup\Components", componentName ), func );
		}

		public void ReadDatabaseProviderSetupKey( string providerName, Action<RegistryKey> action )
		{
			ReadProductKey( Path.Combine( @"Setup\Database Providers", providerName ), action );
		}

		public TResult ReadDatabaseProviderSetupKey<TResult>( string providerName, Func<RegistryKey, TResult> func )
		{
			return ReadProductKey( Path.Combine( @"Setup\Database Providers", providerName ), func );
		}

		public void ReadDriverSetupKey( string driverName, Action<RegistryKey> action )
		{
			string name = driverName.Replace( '.', '\\' );
			ReadProductKey( Path.Combine( @"Setup\Drivers", name ), action );
		}

		public TResult ReadDriverSetupKey<TResult>( string driverName, Func<RegistryKey, TResult> func )
		{
			string name = driverName.Replace( '.', '\\' );
			return ReadProductKey( Path.Combine( @"Setup\Drivers", name ), func );
		}

		public void ReadProductKey( Action<RegistryKey> action )
		{
			string keyPath = RegistryBase;
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( RegistryHive, RegistryView ) )
			using( RegistryKey key = baseKey.OpenSubKey( keyPath ) )
			{
				action( key );
			}
		}

		public void ReadProductKey( RegistryHive hive, Action<RegistryKey> action )
		{
			string keyPath = RegistryBase;
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( hive, RegistryView ) )
			using( RegistryKey key = baseKey.OpenSubKey( keyPath ) )
			{
				action( key );
			}
		}

		public void ReadProductKey( string subkey, Action<RegistryKey> action )
		{
			string keyPath = Path.Combine( RegistryBase, subkey );
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( RegistryHive, RegistryView ) )
			using( RegistryKey key = baseKey.OpenSubKey( keyPath ) )
			{
				action( key );
			}
		}

		public void ReadProductKey( RegistryHive hive, string subkey, Action<RegistryKey> action )
		{
			string keyPath = Path.Combine( RegistryBase, subkey );
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( hive, RegistryView ) )
			using( RegistryKey key = baseKey.OpenSubKey( keyPath ) )
			{
				action( key );
			}
		}

		public TResult ReadProductKey<TResult>( Func<RegistryKey, TResult> func )
		{
			string keyPath = RegistryBase;
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( RegistryHive, RegistryView ) )
			using( RegistryKey key = baseKey.OpenSubKey( keyPath ) )
			{
				return func( key );
			}
		}

		public TResult ReadProductKey<TResult>( string subkey, Func<RegistryKey, TResult> func )
		{
			string keyPath = Path.Combine( RegistryBase, subkey );
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( RegistryHive, RegistryView ) )
			using( RegistryKey key = baseKey.OpenSubKey( keyPath ) )
			{
				return func( key );
			}
		}

		public void WriteProductKey( Action<RegistryKey> action )
		{
			string keyPath = RegistryBase;
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( RegistryHive, RegistryView ) )
			using( RegistryKey key = baseKey.CreateSubKey( keyPath ) )
			{
				action( key );
			}
		}

		public void WriteProductKey( RegistryHive hive, Action<RegistryKey> action )
		{
			string keyPath = RegistryBase;
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( hive, RegistryView ) )
			using( RegistryKey key = baseKey.CreateSubKey( keyPath ) )
			{
				action( key );
			}
		}

		public void WriteProductKey( string subkey, Action<RegistryKey> action )
		{
			string keyPath = Path.Combine( RegistryBase, subkey );
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( RegistryHive, RegistryView ) )
			using( RegistryKey key = baseKey.CreateSubKey( keyPath ) )
			{
				action( key );
			}
		}

		public void WriteProductKey( RegistryHive hive, string subkey, Action<RegistryKey> action )
		{
			string keyPath = Path.Combine( RegistryBase, subkey );
			using( RegistryKey baseKey = RegistryKey.OpenBaseKey( hive, RegistryView ) )
			using( RegistryKey key = baseKey.CreateSubKey( keyPath ) )
			{
				action( key );
			}
		}

		bool assemblyResolverAttached = false;

		public void EnableProductAssemblyResolving()
		{
			if( !assemblyResolverAttached )
			{
				AppDomain.CurrentDomain.AssemblyResolve += AppDomain_OnAssemblyResolve;
				assemblyResolverAttached = true;
			}
		}

		public void DisableProductAssemblyResolving()
		{
			if( assemblyResolverAttached )
			{
				AppDomain.CurrentDomain.AssemblyResolve -= AppDomain_OnAssemblyResolve;
				assemblyResolverAttached = false;
			}
		}

		protected void LoadProductInfo( RegistryKey key )
		{
			if( key != null )
			{
				CompanyName       = (string)key.GetValue( "Company Name", CompanyName );
				ProductName       = (string)key.GetValue( "Product Name", ProductName );
				Brand             = (string)key.GetValue( "Brand", Brand );
				RegistryBase      = LoadFromBrandedRegistry ? Path.Combine( "Software", CompanyName, ProductName ) : Path.Combine( "Software", UnbrandedCompanyName, UnbrandedProductName );
				InstallPath       = (string)key.GetValue( "Install Path", InstallPath );
				DataFolder        = (string)key.GetValue( "Data Folder", DataFolder );
				AllProjectsFolder = (string)key.GetValue( "Projects Folder", AllProjectsFolder );
				LogsFolder        = (string)key.GetValue( "Logs Folder", LogsFolder );
				IsLiteVersion     = (Int32)key.GetValue( "LiteVersion", (Int32)0 ) != 0;
				Version           = (string)key.GetValue( "Version", Version );

				using( RegistryKey setupKey = key.OpenSubKey( "Setup" ) )
				{
					if( setupKey != null )
					{
						SetupPackageCache = (string)setupKey.GetValue( "Package Cache", SetupPackageCache );
					}
				}
			}
		}

		private static ProductInfo _current;
		public static ProductInfo Current
		{
			get
			{
				if( _current == null )
				{
					_current = new ProductInfo( "OSSI", "Intelli-Site" );
				}

				return _current;
			}

			set { _current = value; }
		}

		private static ProductInfo _old3x;
		public static ProductInfo Old3x
		{
			get
			{
				if( _old3x == null )
				{
					_old3x = new ProductInfo( "Intelli-Site, Inc", "Intelli-Site", RegistryView.Registry32 );
				}

				return _old3x;
			}
		}
	}
}
