﻿using System;
using System.IO;
using System.Collections;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Timers;
using Microsoft.Win32;

namespace IntelliSite.CommonUtil
{
	/// <summary>
	/// Summary description for RequestHandler.
	/// </summary>

	public class ReconnectTimer : System.Timers.Timer
	{
		public RequestHandler RequestHandler { get; private set; }

		public ReconnectTimer( RequestHandler requestHandler, ElapsedEventHandler handler ) : base()
		{
			RequestHandler = requestHandler;
			Elapsed       += new ElapsedEventHandler( handler );
			AutoReset      = true;
			Interval       = 3000;
			Enabled        = true;
		}
	}

	public class HeartbeatTimer : System.Timers.Timer
	{
		public RequestHandler RequestHandler { get; private set; }

		public HeartbeatTimer( RequestHandler requestHandler, ElapsedEventHandler handler ) : base()
		{
			RequestHandler = requestHandler;
			Elapsed       += new ElapsedEventHandler( handler );
			AutoReset      = true;
			Interval       = 30000;
			Enabled        = true;
		}
	}

	public class ConnectionInfo
	{
		public MessageSocket Socket { get; private set; }
		public Int32 PrimaryID { get; private set; }
		public Int32 SecondaryID { get; private set; }

		public ConnectionInfo( MessageSocket socket, Int32 primaryID, Int32 secondaryID )
		{
			Socket      = socket;
			PrimaryID   = primaryID;
			SecondaryID = secondaryID;
		}
	}

	public class RequestHandler : IDisposable
	{
		public string ServiceName { get; set; }	
		public ProductInfo ProductInfo { get; private set; }
		public MessageSocket Socket { get; private set; }
		public LoggerBase Logger { get; private set; }
		public bool SocketLoggingEnabled { get; private set; }
		public string Version { get; private set; }

		private string HostAddress { get; set; }
		private UInt16 Port { get; set; }
		private ReconnectTimer ReconnectTimer { get; set; }
		private HeartbeatTimer HeartbeatTimer { get; set; }

		public RequestHandler( ProductInfo productInfo, string serviceName, UInt16 port, string version=default(string) )
		{
			Init( productInfo, serviceName, null, port, null, false, version );
		}

		public RequestHandler( ProductInfo productInfo, string serviceName, string hostAddress, UInt16 port, LoggerBase logger, bool isSocketLoggingEnabled, string version=default(string) )
		{
			Init( productInfo, serviceName, hostAddress, port, logger, isSocketLoggingEnabled, version );
		}

		#region Destructor/Dispose
		~RequestHandler()
		{
			Dispose( false );
		}

		public void Dispose()
		{
			Dispose( true );
		}

		private bool disposed = false;
		protected void Dispose( bool disposing )
		{
			if( !disposed )
			{
				if( disposing )
				{
					if( Socket != null )
						Socket.Dispose();
				}
			}
		}
		#endregion		

		public void Init( ProductInfo productInfo, string serviceName, string hostAddress, UInt16 port, LoggerBase logger, bool isSocketLoggingEnabled, string version )
		{
			ProductInfo          = productInfo;
			ServiceName          = serviceName;
			HostAddress          = hostAddress;
			Port                 = port;
			Socket               = null;
			SocketLoggingEnabled = false;
			Version              = version;

			if( logger == null )
			{
				string registryKeyName = ProductInfo.RegistryBase + "\\" + ServiceName;

				try
				{
					RegistryKey key = Registry.LocalMachine.OpenSubKey( registryKeyName, RegistryKeyPermissionCheck.ReadSubTree );
					if( key != null )
					{
						if( (Int32)key.GetValue( "MessageLogging", (Int32)0 ) != 0 )
						{
							Logger = new LoggerBase( Path.Combine( ProductInfo.LogsFolder, serviceName + "Msg.txt" ) );
							SocketLoggingEnabled = (Int32)key.GetValue( "SocketLogging", (Int32)0 ) != 0;
						}
					}
					else
					{
						key = Registry.LocalMachine.CreateSubKey( registryKeyName );
						key.SetValue( "MessageLogging", 0, RegistryValueKind.DWord );
						key.SetValue( "SocketLogging", 0, RegistryValueKind.DWord );
					}

					if( key != null )
						key.Close();
				}
				catch( Exception )
				{
				}
			}
			else
			{
				Logger               = logger;
				SocketLoggingEnabled = isSocketLoggingEnabled;
			}

			if( Logger != null )
				Logger.Write( serviceName + " started" );

			ReconnectTimer = new ReconnectTimer( this, OnReconnectCheckEvent );
			ReconnectTimer.Start();

			HeartbeatTimer = new HeartbeatTimer( this, OnHeartbeat );
			HeartbeatTimer.Start();
		}

		public void Connect()
		{
			lock( this )
				if( Socket == null || ( !Socket.Listening && !Connected ) )
				{
					if( Socket != null )
						Socket.Dispose();

					string socketName = HostAddress != null ? ( HostAddress + " Socket" ) : "Listening Socket";
					LoggerBase socketLogger = SocketLoggingEnabled ? Logger : null;
					Socket = new MessageSocket( socketName, socketLogger );
					Socket.SocketConnectionLoss = new SocketConnectionLoss( SocketConnectionLoss );
			
					if( Socket.Open( HostAddress, Port, SocketType.Stream, ProtocolType.Tcp ) )
					{
						if( Logger != null )
							Logger.Write( socketName + " opened." );

						if( HostAddress != null )
						{
							Socket.Send( new ConnectMessage( Dns.GetHostName(), 0, Version ) );
							OnConnect();
						}

						ThreadPool.RegisterWaitForSingleObject( Socket.FromRemoteEvent, new WaitOrTimerCallback( ReceiveCallback ), this, 1000, false );
					}
					else
						Socket = null;
				}
		}
		
		private static void OnReconnectCheckEvent(object source, ElapsedEventArgs e) 
		{
			ReconnectTimer timer = (ReconnectTimer)source;
			timer.RequestHandler.Connect();
		}
		
		private static void OnHeartbeat(object source, ElapsedEventArgs e) 
		{
			HeartbeatTimer timer = (HeartbeatTimer)source;
			MessageBase message = new MessageBase();
			message.Type = HeartbeatMessage.MessageType;
			if( timer.RequestHandler.Socket != null && timer.RequestHandler.Socket.Listening )
				timer.RequestHandler.SendAllClientMessage( message );
			else
				timer.RequestHandler.SendMessage( message );
		}

		public bool Connected
		{
			get
			{ 
				if( HostAddress == null )
					return Socket != null && Socket.AcceptedSockets.Count > 0;
				else
					return Socket != null && Socket.SocketState.Socket.Connected;
			}
		}

		public virtual void OnAccept()
		{
		}

		public virtual void OnConnect()
		{
		}

		public void SendMessage( MessageBase message )
		{
			if( Logger != null )
				Logger.Write( "Tx " + message.ToString() );
				
			Socket.Send( Socket.SocketState, message );
		}

		public void SendClientMessage( SocketState socketState, MessageBase message )
		{
			if( Logger != null )
				Logger.Write( "Tx (" + (string)socketState.Parent.Name + "): " + message.ToString() );
				
			Socket.Send( socketState, message );
		}

		public void SendAllClientMessage( MessageBase message )
		{
			SendAllClientMessage( message, 0 );
		}

		public void SendAllClientMessage( MessageBase message, byte type )
		{
			if( Logger != null )
				Logger.Write( "Tx (All " + type + "): " + message.ToString() );
				
			Socket.SendAll( message, type );
		}

		public static void ReceiveCallback(object state, bool timedOut) 
		{
			RequestHandler handler = (RequestHandler)state;
			try
			{
				MessageFrom		messageFrom	= null;

				lock( handler.Socket.FromRemoteQueue )
				{
					if( handler.Socket.FromRemoteQueue.Count > 0 )
						messageFrom = handler.Socket.FromRemoteQueue.Dequeue();
				}

				if( messageFrom != null && messageFrom.Message != null )
				{
					switch( messageFrom.Message.Type )
					{
						case ConnectMessage.MessageType:
							{
								messageFrom.SocketState.Parent.Name = ( (ConnectMessage)messageFrom.Message ).ClientName;
								messageFrom.SocketState.Type = 0;
								break;
							}
					}

					if( handler.Logger != null && handler.SocketLoggingEnabled )
						handler.Logger.Write( "ReceiveCallback: " + messageFrom.Message.ToString() );

					handler.ProcessMessage( messageFrom.SocketState, messageFrom.Message );
				}
			}
			catch( Exception e )
			{
				if( handler.Logger != null && handler.SocketLoggingEnabled )
					handler.Logger.Write( "ReceiveCallback Exception: " + e.ToString() );
			}
		}

		public virtual void ProcessMessage( SocketState socketState, MessageBase message )
		{
			if( Logger != null )
				Logger.Write( "Rx (" + (string)socketState.Parent.Name + "): " + message.ToString() );
					

			switch( message.Type )
			{
				#region Connect
				case ConnectMessage.MessageType:
				{
					socketState.Parent.Name = ((ConnectMessage)message).ClientName;
					SendClientMessage( socketState, new ConnectStatusMessage( ConnectionStatus.Success ) );
					break;
				}
				#endregion
				#region Disconnect
				case DisconnectMessage.MessageType:
				{
					break;
				}
				#endregion
			}
		}

		private void SocketConnectionLoss( SocketState socketState, Exception e )
		{
			MessageConnectionLoss( e );
		}

		public virtual void MessageConnectionLoss( Exception e )
		{
		}
	}
}
