﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.InteropServices;
using System.ServiceProcess;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace IntelliSite.CommonUtil
{
	public static class ServiceHelper
	{
		[DllImport( "advapi32.dll", CharSet = CharSet.Unicode, SetLastError = true )]
		[return: MarshalAs( UnmanagedType.Bool )]
		private static extern bool ChangeServiceConfig(
			IntPtr hService,
			ServiceType nServiceType,
			ServiceStartupMode nStartupMode,
			ServiceErrorControl nErrorControl,
			[Optional] string lpBinaryPathName,
			[Optional] string lpLoadOrderGroup,
			[Optional] IntPtr lpdwTagId,
			[In, Optional] char[] lpDependencies,
			[Optional] string lpServiceStartName,
			[Optional] string lpPassword,
			[Optional] string lpDisplayName );

		[DllImport( "advapi32.dll", CharSet = CharSet.Unicode, SetLastError = true )]
		[return: MarshalAs( UnmanagedType.Bool )]
		private static extern bool ChangeServiceConfig2(
			IntPtr hService,
			ServiceConfigInfoLevel dwInfoLevel,
			[Optional] IntPtr lpInfo );

		[DllImport( "advapi32.dll", CharSet = CharSet.Unicode, SetLastError = true )]
		private static extern IntPtr OpenSCManager(
			string machineName,
			string databaseName,
			ServerControlManagerAccess accessFlags );

		[DllImport( "advapi32.dll", CharSet = CharSet.Unicode, SetLastError = true )]
		private static extern IntPtr OpenService(
			IntPtr hSCManager,
			string lpServiceName,
			ServiceAccess desiredAccessFlags );

		[DllImport( "advapi32.dll", SetLastError = true )]
		[return: MarshalAs( UnmanagedType.Bool )]
		private static extern bool CloseServiceHandle( IntPtr hSCObject );

		public const int SERVICE_NO_CHANGE = -1;

		/// <summary>
		/// The ACCESS_MASK data type is a double word value that defines standard, specific, and generic rights. These rights are used in access control entries (ACEs) and are the primary means of specifying the requested or granted access to an object.
		/// </summary>
		[Flags]
		public enum AccessMask
		{
			DELETE = 0x00010000,
			READ_CONTROL = 0x00020000,
			WRITE_DAC = 0x00040000,
			WRITE_OWNER = 0x00080000,
			SYNCHRONIZE = 0x00100000,

			STANDARD_RIGHTS_REQUIRED = 0x000F0000,

			STANDARD_RIGHTS_READ = 0x00020000,
			STANDARD_RIGHTS_WRITE = 0x00020000,
			STANDARD_RIGHTS_EXECUTE = 0x00020000,

			STANDARD_RIGHTS_ALL = 0x001F0000,

			SPECIFIC_RIGHTS_ALL = 0x0000FFFF,

			ACCESS_SYSTEM_SECURITY = 0x01000000,

			MAXIMUM_ALLOWED = 0x02000000,

			GENERIC_READ = int.MinValue,
			GENERIC_WRITE = 0x40000000,
			GENERIC_EXECUTE = 0x20000000,
			GENERIC_ALL = 0x10000000,

			DESKTOP_READOBJECTS = 0x00000001,
			DESKTOP_CREATEWINDOW = 0x00000002,
			DESKTOP_CREATEMENU = 0x00000004,
			DESKTOP_HOOKCONTROL = 0x00000008,
			DESKTOP_JOURNALRECORD = 0x00000010,
			DESKTOP_JOURNALPLAYBACK = 0x00000020,
			DESKTOP_ENUMERATE = 0x00000040,
			DESKTOP_WRITEOBJECTS = 0x00000080,
			DESKTOP_SWITCHDESKTOP = 0x00000100,

			WINSTA_ENUMDESKTOPS = 0x00000001,
			WINSTA_READATTRIBUTES = 0x00000002,
			WINSTA_ACCESSCLIPBOARD = 0x00000004,
			WINSTA_CREATEDESKTOP = 0x00000008,
			WINSTA_WRITEATTRIBUTES = 0x00000010,
			WINSTA_ACCESSGLOBALATOMS = 0x00000020,
			WINSTA_EXITWINDOWS = 0x00000040,
			WINSTA_ENUMERATE = 0x00000100,
			WINSTA_READSCREEN = 0x00000200,

			WINSTA_ALL_ACCESS = 0x0000037F
		}

		[Flags]
		public enum ServerControlManagerAccess
		{
			/// <summary>
			/// Required to connect to the service control manager.
			/// </summary>
			SC_MANAGER_CONNECT = 0x00001,

			/// <summary>
			/// Required to call the CreateService function to create a service
			/// object and add it to the database.
			/// </summary>
			SC_MANAGER_CREATE_SERVICE = 0x00002,

			/// <summary>
			/// Required to call the EnumServicesStatusEx function to list the
			/// services that are in the database.
			/// </summary>
			SC_MANAGER_ENUMERATE_SERVICE = 0x00004,

			/// <summary>
			/// Required to call the LockServiceDatabase function to acquire a
			/// lock on the database.
			/// </summary>
			SC_MANAGER_LOCK = 0x00008,

			/// <summary>
			/// Required to call the QueryServiceLockStatus function to retrieve
			/// the lock status information for the database.
			/// </summary>
			SC_MANAGER_QUERY_LOCK_STATUS = 0x00010,

			/// <summary>
			/// Required to call the NotifyBootConfigStatus function.
			/// </summary>
			SC_MANAGER_MODIFY_BOOT_CONFIG = 0x00020,

			STANDARD_RIGHTS_REQUIRED = AccessMask.STANDARD_RIGHTS_REQUIRED,
			STANDARD_RIGHTS_READ = AccessMask.STANDARD_RIGHTS_READ,
			STANDARD_RIGHTS_WRITE = AccessMask.STANDARD_RIGHTS_WRITE,
			STANDARD_RIGHTS_EXECUTE = AccessMask.STANDARD_RIGHTS_EXECUTE,

			/// <summary>
			/// Includes STANDARD_RIGHTS_REQUIRED, in addition to all access
			/// rights in this table.
			/// </summary>
			SC_MANAGER_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED |
				SC_MANAGER_CONNECT |
				SC_MANAGER_CREATE_SERVICE |
				SC_MANAGER_ENUMERATE_SERVICE |
				SC_MANAGER_LOCK |
				SC_MANAGER_QUERY_LOCK_STATUS |
				SC_MANAGER_MODIFY_BOOT_CONFIG,

			GENERIC_READ = STANDARD_RIGHTS_READ |
				SC_MANAGER_ENUMERATE_SERVICE |
				SC_MANAGER_QUERY_LOCK_STATUS,

			GENERIC_WRITE = STANDARD_RIGHTS_WRITE |
				SC_MANAGER_CREATE_SERVICE |
				SC_MANAGER_MODIFY_BOOT_CONFIG,

			GENERIC_EXECUTE = STANDARD_RIGHTS_EXECUTE |
				SC_MANAGER_CONNECT | SC_MANAGER_LOCK,

			GENERIC_ALL = SC_MANAGER_ALL_ACCESS,
		}

		/// <summary>
		///  Specific access rights for a service.
		/// </summary>
		[Flags]
		public enum ServiceAccess
		{
			STANDARD_RIGHTS_REQUIRED = 0xF0000,
			SERVICE_QUERY_CONFIG = 0x00001,
			SERVICE_CHANGE_CONFIG = 0x00002,
			SERVICE_QUERY_STATUS = 0x00004,
			SERVICE_ENUMERATE_DEPENDENTS = 0x00008,
			SERVICE_START = 0x00010,
			SERVICE_STOP = 0x00020,
			SERVICE_PAUSE_CONTINUE = 0x00040,
			SERVICE_INTERROGATE = 0x00080,
			SERVICE_USER_DEFINED_CONTROL = 0x00100,
			SERVICE_ALL_ACCESS = ( STANDARD_RIGHTS_REQUIRED |
							  SERVICE_QUERY_CONFIG |
							  SERVICE_CHANGE_CONFIG |
							  SERVICE_QUERY_STATUS |
							  SERVICE_ENUMERATE_DEPENDENTS |
							  SERVICE_START |
							  SERVICE_STOP |
							  SERVICE_PAUSE_CONTINUE |
							  SERVICE_INTERROGATE |
							  SERVICE_USER_DEFINED_CONTROL )
		}

		public enum ServiceConfigInfoLevel
		{
			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_DESCRIPTION structure.
			/// </summary>
			Description = 1,

			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_FAILURE_ACTIONS structure.
			/// 
			/// If the service controller handles the SC_ACTION_REBOOT action, the caller must have the SE_SHUTDOWN_NAME privilege. For more information, see Running with Special Privileges.
			/// </summary>
			FailureActions = 2,

			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_DELAYED_AUTO_START_INFO structure.
			/// 
			/// This value is not supported until Windows Vista.
			/// </summary>
			DelayedAutomaticStart = 3,

			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_FAILURE_ACTIONS_FLAG structure.
			/// 
			/// This value is not supported until Windows Vista.
			/// </summary>
			FailureActionsFlag = 4,

			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_SID_INFO structure.
			/// </summary>
			ServiceSid = 5,

			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_REQUIRED_PRIVILEGES_INFO structure.
			/// </summary>
			RequiredPrivileges = 6,

			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_PRESHUTDOWN_INFO structure.
			/// 
			/// This value is not supported until Windows Vista.
			/// </summary>
			PreShutdown = 7,

			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_TRIGGER_INFO structure.
			/// 
			/// This value is not supported until Windows 7.
			/// </summary>
			Trigger = 8,

			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_TRIGGER_INFO structure.
			/// 
			/// This value is not supported until Windows 7.
			/// </summary>
			PreferredNode = 9,

			/// <summary>
			/// The lpInfo parameter is a pointer to a SERVICE_TRIGGER_INFO structure.
			/// 
			/// This value is not supported until Windows 8.1.
			/// </summary>
			LaunchProtected = 12,

		}

		/// <summary>
		/// The service start options.
		/// </summary>
		public enum ServiceStartupMode
		{
			/// <summary>
			/// A device driver started by the system loader. This value is valid only for driver services.
			/// </summary>
			Bootup = 0x0000,

			/// <summary>
			/// A device driver started by the IoInitSystem function. This value is valid only for driver services.
			/// </summary>
			SystemStartup = 0x0001,

			/// <summary>
			/// A service started automatically by the service control manager during system startup.
			/// </summary>
			Automatic = 0x0002,

			/// <summary>
			/// A service started by the service control manager when a process calls the StartService function.
			/// </summary>
			Manual = 0x0003,

			/// <summary>
			/// A service that cannot be started. Attempts to start the service result in the error code ERROR_SERVICE_DISABLED.
			/// </summary>
			Disabled = 0x0004,
		}

		/// <summary>
		/// The severity of the error, and action taken, if this service fails to start.
		/// </summary>
		public enum ServiceErrorControl
		{
			/// <summary>
			/// The startup program logs the error in the event log, if possible. If the last-known-good configuration is being started, the startup operation fails. Otherwise, the system is restarted with the last-known good configuration.
			/// </summary>
			Ignore = 0x0000,

			/// <summary>
			/// The startup program ignores the error and continues the startup operation.
			/// </summary>
			Normal = 0x0001,

			/// <summary>
			/// The startup program logs the error in the event log but continues the startup operation.
			/// </summary>
			Severe = 0x0002,

			/// <summary>
			/// The startup program logs the error in the event log. If the last-known-good configuration is being started, the startup operation continues. Otherwise, the system is restarted with the last-known-good configuration.
			/// </summary>
			Critical = 0x0003,
		}

		public static void ChangeServiceParameters(
			ServiceController service,
			ServiceType? type = null,
			ServiceStartupMode? mode = null,
			ServiceErrorControl? errorControl = null,
			string binaryPathName = null,
			string loadOrderGroup = null,
			string[] dependencies = null,
			string serviceStartName = null,
			string password = null,
			string displayName = null )
		{
			if( service == null )
				throw new ArgumentNullException( "service" );

			var serviceName = service.ServiceName;
			var serviceType = type ?? (ServiceType)SERVICE_NO_CHANGE;
			var serviceMode = mode ?? (ServiceStartupMode)SERVICE_NO_CHANGE;
			var serviceErrorControl = errorControl ?? (ServiceErrorControl)SERVICE_NO_CHANGE;

			char[] lpDependencies = null;
			if( dependencies != null )
			{
				if( dependencies.Any( val => val.Contains( '\0' ) ) )
					throw new ArgumentOutOfRangeException( "dependencies", "Argument must not contain any values with embedded nulls." );

				lpDependencies = ( String.Join( "\0", value: dependencies ) + "\0\0" ).ToCharArray();
			}

			if( displayName != null )
			{
				if( displayName.Length == 0 )
					throw new ArgumentOutOfRangeException( "displayName", "Argument must not be an empty string." );

				if( displayName.Length > 255 )
					throw new ArgumentOutOfRangeException( "displayName", "Argument must not be longer than 256 characters." );
			}

			var scManagerHandle = IntPtr.Zero;
			var serviceHandle = IntPtr.Zero;
			var scManagerAccessFlags = ServerControlManagerAccess.SC_MANAGER_ALL_ACCESS;
			var accessFlags = ServiceAccess.SERVICE_QUERY_CONFIG | ServiceAccess.SERVICE_CHANGE_CONFIG;

			try
			{
				scManagerHandle = OpenSCManager( null, null, scManagerAccessFlags );
				if( scManagerHandle == IntPtr.Zero )
				{
					int nError = Marshal.GetLastWin32Error();
					var win32Exception = new Win32Exception( nError );
					throw new ExternalException( "Open service manager error: " + win32Exception.Message, win32Exception );
				}

				serviceHandle = OpenService( scManagerHandle, serviceName, accessFlags );
				if( serviceHandle == IntPtr.Zero )
				{
					int nError = Marshal.GetLastWin32Error();
					var win32Exception = new Win32Exception( nError );
					throw new ExternalException( "Open service error: " + win32Exception.Message, win32Exception );
				}

				bool succeeded = ChangeServiceConfig(
					serviceHandle,
					serviceType,
					serviceMode,
					serviceErrorControl,
					binaryPathName,
					loadOrderGroup,
					IntPtr.Zero,
					lpDependencies,
					serviceStartName,
					password,
					displayName );

				if( !succeeded )
				{
					int nError = Marshal.GetLastWin32Error();
					var win32Exception = new Win32Exception( nError );
					throw new ExternalException( "Could not change service display name: " + win32Exception.Message, win32Exception );
				}
			}
			finally
			{
				CloseServiceHandle( serviceHandle );
				CloseServiceHandle( scManagerHandle );
			}
		}

		public static void ChangeDisplayName( ServiceController service, string displayName )
		{
			ChangeServiceParameters( service, displayName: displayName );
		}

		public static void ChangeStartMode( ServiceController service, ServiceStartMode serviceStartMode )
		{
			ChangeServiceParameters( service, mode: (ServiceStartupMode)serviceStartMode );
		}

		public static ServiceController[] GetServices()
		{
			return ServiceController.GetServices();
		}

		public static ServiceController[] GetServices( string machineName )
		{
			return ServiceController.GetServices( machineName );
		}

		public static bool IsValidDisplayName( string displayName )
		{
			if( String.IsNullOrEmpty( displayName ) || 
				displayName.Length > 255 )
				return false;
			return true;
		}

		public static bool IsValidNewServiceName( string serviceName, out string message, bool checkValidFileName = true )
		{
			message = null;
			if( !ServiceHelper.IsValidServiceName( serviceName ) )
				message = "Invalid service name. Service names cannot contain \"\\\" nor \"/\".";
			else if( ServiceHelper.InstalledServiceNames.Contains( serviceName ) )
				message = "A service with this name already exists.";
			else if( !PathHelper.IsValidFileName( serviceName ) )
				message = "Service name cannot contain: { \"" + String.Join( "\", \"", System.IO.Path.GetInvalidFileNameChars() ) + "\" }";

			return String.IsNullOrEmpty( message );
		}

		public static bool IsValidServiceName( string serviceName )
		{
			if( String.IsNullOrEmpty( serviceName ) || 
				serviceName.Length > 255 || 
				serviceName.Contains( '/' ) || 
				serviceName.Contains( '\\' ) )
				return false;
			return true;
		}

		private static ICollection<string> _installedServiceNames;
		public static ICollection<string> InstalledServiceNames
		{
			get
			{
				if( _installedServiceNames == null )
					RefreshInstalledServiceNames();
				return _installedServiceNames;
			}
		}

		public static void RefreshInstalledServiceNames()
		{
			IEnumerable<string> serviceNames = GetServices().Select( service => service.ServiceName );
			_installedServiceNames = new HashSet<string>( serviceNames, StringComparer.InvariantCultureIgnoreCase );
		}

		public static void StartService( string name )
		{
			StartService( name, TimeSpan.MaxValue );
		}

		public static void StartService( string name, TimeSpan timeout )
		{
			ServiceController controller = null;

			try
			{
				controller = new ServiceController( name );
				StartService( controller, timeout );
			}
			finally
			{
				if( controller != null )
					controller.Dispose();
			}
		}

		public static void StartService( ServiceController controller, TimeSpan timeout )
		{
			switch( controller.Status )
			{
				case ServiceControllerStatus.PausePending:
					controller.WaitForStatus( ServiceControllerStatus.Paused, timeout );
					goto case ServiceControllerStatus.Paused;
				case ServiceControllerStatus.StopPending:
					controller.WaitForStatus( ServiceControllerStatus.Stopped, timeout );
					goto case ServiceControllerStatus.Stopped;
				case ServiceControllerStatus.Stopped:
					controller.Start();
					goto case ServiceControllerStatus.StartPending;
				case ServiceControllerStatus.Paused:
					controller.Continue();
					goto case ServiceControllerStatus.ContinuePending;
				case ServiceControllerStatus.ContinuePending:
				case ServiceControllerStatus.StartPending:
					controller.WaitForStatus( ServiceControllerStatus.Running, timeout );
					goto case ServiceControllerStatus.Running;
				case ServiceControllerStatus.Running:
				default:
					break;
			}
		}

		public static ServiceControllerStatus ServiceStatus( string name )
		{
			try
			{
				ServiceController controller = new ServiceController( name );
				return controller.Status;
			}
			catch( Exception )
			{
				return ServiceControllerStatus.Stopped;
			}
		}

		public static void StopService( string name )
		{
			StopService( name, TimeSpan.MaxValue );
		}

		public static void StopService( string name, TimeSpan timeout )
		{
			ServiceController controller = null;

			try
			{
				controller = new ServiceController( name );
				StopService( controller, timeout );
			}
			finally
			{
				if( controller != null )
					controller.Dispose();
			}
		}

		public static void StopService( ServiceController controller, TimeSpan timeout )
		{
			switch( controller.Status )
			{
				case ServiceControllerStatus.PausePending:
					controller.WaitForStatus( ServiceControllerStatus.Paused, timeout );
					goto case ServiceControllerStatus.Running;
				case ServiceControllerStatus.ContinuePending:
				case ServiceControllerStatus.StartPending:
					controller.WaitForStatus( ServiceControllerStatus.Running, timeout );
					goto case ServiceControllerStatus.Running;
				case ServiceControllerStatus.Running:
				case ServiceControllerStatus.Paused:
					controller.Stop();
					goto case ServiceControllerStatus.StopPending;
				case ServiceControllerStatus.StopPending:
					controller.WaitForStatus( ServiceControllerStatus.Stopped, timeout );
					goto case ServiceControllerStatus.Stopped;
				case ServiceControllerStatus.Stopped:
				default:
					break;
			}
		}

		public static bool TryStartService( string name, TimeSpan timeout )
		{
			try
			{
				StartService( name, timeout );
				return true;
			}
			catch
			{
				return false;
			}
		}

		public static bool TryStopService( string name, TimeSpan timeout )
		{
			try
			{
				StartService( name, timeout );
				return true;
			}
			catch
			{
				return false;
			}
		}
	}
}
