﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Collections;

namespace IntelliSite.CommonUtil
{
	/// <summary>
	/// Summary description for SimpleSocket.
	/// </summary>

	public class SimpleSocketStateClass : IDisposable
	{
		public SimpleSocket	Parent	{ get; set; }
		public Socket		Socket { get; set; }
		public object		Data { get; set; }
		public byte			Type { get; set; }
		public bool			IsBuffer1 { get; set; }
		public byte []		FromRemoteBuffer1 { get; set; }
		public byte []		FromRemoteBuffer2 { get; set; }
		public bool			IsSendActive { get; set; }
		
		public SimpleSocketStateClass( SimpleSocket parent, Socket socket )
		{
			Parent            = parent;
			Socket            = socket;
			Data              = null;
			Type              = 0;
			IsBuffer1         = true;
			FromRemoteBuffer1 = new byte[16384];
			FromRemoteBuffer2 = new byte[16384];
			IsSendActive      = false;

			IsDisposed        = false;
		}

		#region Destructor/Dispose
		~SimpleSocketStateClass()
		{
			Dispose( false );
		}

		public void Dispose()
		{
			Dispose( true );
		}

		public bool IsDisposed { get; set; }
		protected void Dispose( bool disposing )
		{
			if( !IsDisposed )
			{
				if( disposing )
				{
					if( Socket != null )
					{
						Socket.Dispose();
						Socket = null;
					}
				}
			}
		}
		#endregion		
	}
	
	public class SimpleSocket : IDisposable
	{
		public SimpleSocketStateClass SocketState { get; set; }
		public AutoResetEvent         FromRemoteEvent { get; set; }
		public byte []                FromRemoteData { get; set; }
		public byte []                TxLockBuffer { get; set; }
		public byte []                LockBuffer { get; set; }
		public string                 Name { get; set; }
		public Exception              LastError { get; set; }
		
		public SimpleSocket( string name )
		{
			Name            = name;
			FromRemoteEvent = new AutoResetEvent( false );
			FromRemoteData  = null;
			TxLockBuffer    = new byte[1];
			LockBuffer      = new byte[1];
			LastError       = null;
		}

		#region Destructor/Dispose
		~SimpleSocket()
		{
			Dispose( false );
		}

		public void Dispose()
		{
			Dispose( true );
		}

		private bool disposed = false;
		protected void Dispose( bool disposing )
		{
			if( !disposed )
			{
				if( disposing )
				{
					Close();
					FromRemoteEvent.Dispose();
				}
			}
		}
		#endregion		

		public static IPEndPoint GetIPEndPoint( string address, int port, LoggerBase logger=null, bool ipV6=false )
		{
			IPEndPoint endPoint = null;
			try
			{
				IPAddress ipAddress = null;
				if( !IPAddress.TryParse( address, out ipAddress ) )
				{
					IPHostEntry ipHostInfo = Dns.GetHostEntry( address );
					foreach( IPAddress test in ipHostInfo.AddressList )
						if( ( !ipV6 && test.AddressFamily == AddressFamily.InterNetwork ) ||
							( ipV6 && test.AddressFamily == AddressFamily.InterNetworkV6 ) )
						{
							ipAddress = test;
							break;
						}
				}

				endPoint = new IPEndPoint( ipAddress, port );
			}
			catch( Exception e )
			{
				if( logger != null )
					logger.Write( "SimpleSocket.GetIPEndPoint exception: ", e.ToString() );
			}

			if( logger != null )
			{
				if( endPoint == null )
					logger.Write( "SimpleSocket.GetIPEndPoint could not resolve {0} endpoint for {1}, port: {2}: ", ipV6 ? "IPv6" : "IPv4", address, port );
				else
					logger.Write( "SimpleSocket.GetIPEndPoint resolved endpoint for {0}, port: {1} to: {2}", address, port, endPoint.ToString() );
			}

			return endPoint;
		}
		
		public bool Open( string address, int port, int timeout=3000 )
		{
			try
			{
				Socket socket = new Socket( AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp );
				SocketState = new SimpleSocketStateClass( this, socket );

				IPEndPoint endPoint = GetIPEndPoint( address, port );
				var result = socket.BeginConnect( endPoint, null, null );

				bool success = result.AsyncWaitHandle.WaitOne( timeout, true );
				if( success )
				{
					socket.EndConnect( result );
					socket.BeginReceive( SocketState.FromRemoteBuffer1, 0, SocketState.FromRemoteBuffer1.Length, SocketFlags.None, new AsyncCallback( ReceiveCallback ), SocketState );
					return true;
				}
				else
				{
					socket.Close();
					throw new SocketException( 10060 ); // Connection timed out.
				}
			}
			catch( Exception e ) 
			{
				LastError = e;
				Console.WriteLine( Name + " Open Exception: " + e.Message );
				return false;
			}
		}

		public void Close()
		{
			if( SocketState != null )
			{
				SocketState.Dispose();
				SocketState = null;
			}
		}
		
		public bool Connected { get { return SocketState.Socket != null && SocketState.Socket.Connected; } }

		public bool Send( byte [] message )
		{
			lock( TxLockBuffer )
			{
				try
				{
					SocketState.Socket.Send( message );
				}
				catch (Exception e)
				{
					LastError = e;
					Console.WriteLine( Name + " Send error: " + e.Message );
					return false;
				}
			}
			return true;
		}

		public byte [] Receive( int timeout )
		{
			if( FromRemoteEvent.WaitOne( timeout, false ) )
			{
				lock( LockBuffer )
				{
					byte [] retVal = FromRemoteData;
					FromRemoteData = null;
					return retVal;
				}
			}
			return null;
		}

		public bool SendString( string message )
		{
			byte [] data = new byte[message.Length];
			System.Buffer.BlockCopy( message.ToCharArray(), 0, data, 0, message.Length );
			return Send( data );
		}

		public string ReceiveString( int timeout )
		{
			byte [] data = Receive( timeout );
			if( data != null )
				return System.Text.Encoding.ASCII.GetString( data );

			return null;
		}

		private static void ReceiveCallback( IAsyncResult ar ) 
		{
			SimpleSocketStateClass rxSocketState = (SimpleSocketStateClass) ar.AsyncState;
			if( rxSocketState == null || rxSocketState.IsDisposed || rxSocketState.Socket == null )
				return;

			lock( rxSocketState )
			{
				try 
				{
					int bytes = rxSocketState.Socket.EndReceive( ar );
					if( bytes == 0 )
					{
						if( rxSocketState.Socket.Available == 0 && rxSocketState.Socket.Poll( 1, SelectMode.SelectRead ) )
						{
							Console.WriteLine( rxSocketState.Parent.Name + " Socket closed." );
							rxSocketState.Socket = null;
							return;
						}
					}
					byte [] rxBuffer = rxSocketState.IsBuffer1 ? rxSocketState.FromRemoteBuffer1 : rxSocketState.FromRemoteBuffer2;

					rxSocketState.IsBuffer1 = !rxSocketState.IsBuffer1;
					byte [] nextRXBuffer = rxSocketState.IsBuffer1 ? rxSocketState.FromRemoteBuffer1 : rxSocketState.FromRemoteBuffer2;
					rxSocketState.Socket.BeginReceive( nextRXBuffer, 0,  rxSocketState.FromRemoteBuffer1.Length, SocketFlags.None, new AsyncCallback( ReceiveCallback ), rxSocketState );

					lock( rxSocketState.Parent.LockBuffer )
					{
						byte [] newData = null;
						if( rxSocketState.Parent.FromRemoteData != null )
						{
							newData = new byte[rxSocketState.Parent.FromRemoteData.Length+bytes];
							Array.Copy( rxSocketState.Parent.FromRemoteData, newData, rxSocketState.Parent.FromRemoteData.Length );
							Array.Copy( rxBuffer, 0, newData, rxSocketState.Parent.FromRemoteData.Length, bytes );
						}
						else
						{
							newData = new byte[bytes];
							Array.Copy( rxBuffer, newData, bytes );
						}
						rxSocketState.Parent.FromRemoteData = newData;
					}
					rxSocketState.Parent.FromRemoteEvent.Set();
				} 
				catch (Exception e) 
				{
					rxSocketState.Parent.LastError = e;
					Console.WriteLine( rxSocketState.Parent.Name + " Receive Exception: " + e.Message );
				}
			}
		}
	}
}
