﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;

namespace IntelliSite.CommonUtil
{
	public class ConnectionArgs : EventArgs
	{
		public PeekableSocketClientStream Stream { get; set; }

		public ConnectionArgs( PeekableSocketClientStream stream )
		{
			Stream = stream;
		}
	}

	public class SocketServer : IDisposable
	{
		public string                           Name            { get; set; }
		public List<PeekableSocketClientStream> Clients         { get; private set; }
		public Socket                           ListeningSocket { get; private set; }
		public int                              Port            { get; private set; }
		public Encoding                         Encoding        { get; private set; }
		public LoggerBase                       Logger          { get; private set; }
		public ErrorArgs                        ErrorArgs       { get; private set; }

		public event EventHandler<ConnectionArgs> ConnectionAccepted;
		public event EventHandler<ErrorArgs> ConnectionLoss;

		public SocketServer( string name = "SocketServer", Encoding encoding = null, LoggerBase logger = null )
		{
			Name             = name;
			Clients          = new List<PeekableSocketClientStream>();
			ListeningSocket  = null;
			Encoding         = encoding;
			Logger           = logger;
			Port             = 0;
			ErrorArgs        = null;
		}

		#region Destructor/Dispose
		~SocketServer()
		{
			Dispose( false );
		}

		public void Dispose()
		{
			Dispose( true );
		}

		private bool disposed = false;
		protected void Dispose( bool disposing )
		{
			if( !disposed )
			{
				if( disposing )
				{
					Close();
				}
			}
		}
		#endregion

		public void Close()
		{
			foreach( PeekableSocketClientStream stream in Clients )
			{
				stream.Close();
				stream.Dispose();
			}

			lock( Clients )
			{
				Clients.Clear();
			}

			if( ListeningSocket != null )
			{
				if( ListeningSocket.Connected )
					ListeningSocket.Close();

				ListeningSocket.Dispose();
				ListeningSocket = null;
			}
		}

		public void OnAccept( object sender, SocketAsyncEventArgs args )
		{
			if( args.SocketError == SocketError.Success )
			{
				PeekableSocketClientStream clientStream = new PeekableSocketClientStream( "Client", args.AcceptSocket, Encoding, Logger );
				clientStream.ConnectionLoss += OnClientConnectionLoss;

				lock( Clients )
				{
					Clients.Add( clientStream );
				}

				ConnectionAccepted( this, new ConnectionArgs( clientStream ) );
				AcceptAsync();
			}
			else
				OutputDebug( "{0}: OnAccept error, {1}", Name, args.SocketError );
		}

		private void AcceptAsync()
		{
			SocketAsyncEventArgs args = new SocketAsyncEventArgs();
			args.UserToken = this;
			args.Completed += OnAccept;

			if( !ListeningSocket.AcceptAsync( args ) )
				OnAccept( this, args );
		}

		public bool Listening { get { return ListeningSocket != null && ListeningSocket.IsBound; } }

		public bool Listen( int           port,
							bool          waitForCompletion = true,
							AddressFamily addressFamily     = AddressFamily.InterNetwork,
							SocketType    socketType        = SocketType.Stream,
							ProtocolType  protocolType      = ProtocolType.Tcp )
		{
			Close();

			Port = port;

			try
			{
				ListeningSocket = new Socket( addressFamily, socketType, protocolType );
				ListeningSocket.Bind( new IPEndPoint( IPAddress.Any, port ) );
				ListeningSocket.Listen( 50 );

				AcceptAsync();
			}
			catch( SocketException e )
			{
				OutputDebug( "{0}: Listen failed, {1}", Name, SocketErrorString.FullError( e.ErrorCode ) );
				ErrorArgs = new ErrorArgs( this, "Listen error, " + SocketErrorString.FullError( e.ErrorCode ), e );
				return false;
			}
			catch( Exception e )
			{
				OutputDebug( "{0}: Listen failed, {1}", Name, e.ToString() );
				ErrorArgs = new ErrorArgs( this, "Listen error, ", e );
				return false;
			}
			return true;
		}

		public void OnClientConnectionLoss( object sender, ErrorArgs args )
		{
			lock( Clients )
			{
				Clients.Remove( sender as PeekableSocketClientStream );
			}

			if( ConnectionLoss != null )
				ConnectionLoss( sender, args );
		}

		#region public void OutputDebug( string format, params object[] list )
		public void OutputDebug( string format, params object[] list )
		{
			if( Logger != null )
				Logger.Write( String.Format( format, list ) );
		}
		#endregion
	}
}
