﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;

namespace IntelliSite.CommonUtil
{
	using System.Collections;
	using IntelliSite.CommonUtil.SystemInfoApi;

	public class SystemInfo
	{
		protected SystemInfo() { }

		internal SystemInfo( SYSTEM_INFO info )
		{
			ProcessorArchitecture = GetProcessorArchitecture( info );
			PageSize = info.dwPageSize;
			MinimumApplicationAddress = info.lpMinimumApplicationAddress;
			MaximumApplicationAddress = info.lpMaximumApplicationAddress;
			ActiveProcessorMask = info.dwActiveProcessorMask;
			NumberOfProcessors = info.dwNumberOfProcessors;
			AllocationGranularity = info.dwAllocationGranularity;
			ProcessorLevel = info.wProcessorLevel;
			ProcessorRevision = info.wProcessorRevision;
		}

		public UIntPtr ActiveProcessorMask { get; protected set; }
		public uint AllocationGranularity { get; protected set; }
		public IntPtr MinimumApplicationAddress { get; protected set; }
		public IntPtr MaximumApplicationAddress { get; protected set; }
		public uint NumberOfProcessors { get; protected set; }
		public uint PageSize { get; protected set; }
		public ProcessorArchitecture ProcessorArchitecture { get; protected set; }
		public ushort ProcessorLevel { get; protected set; }
		public ushort ProcessorRevision { get; protected set; }

		static SystemInfo _currentSystemInfo;
		static SystemInfo _currentNativeSystemInfo;

		public static SystemInfo Current
		{ 
			get
			{
				if( _currentSystemInfo == null )
				{
					SYSTEM_INFO tmpInfo;
					Native.GetSystemInfo( out tmpInfo );
					_currentSystemInfo = new SystemInfo( tmpInfo );
				}

				return _currentSystemInfo;
			}
		}

		public static SystemInfo CurrentNative
		{
			get
			{
				if( _currentNativeSystemInfo == null )
				{
					SYSTEM_INFO tmpInfo;
					Native.GetNativeSystemInfo( out tmpInfo );
					_currentNativeSystemInfo = new SystemInfo( tmpInfo );
				}

				return _currentNativeSystemInfo;
			}
		}

		internal static ProcessorArchitecture GetProcessorArchitecture( SYSTEM_INFO info )
		{
			return ToProcessorArchitecture( (PROCESSOR_ARCHITECTURE)info.wProcessorArchitecture );
		}

		internal static ProcessorArchitecture ToProcessorArchitecture( PROCESSOR_ARCHITECTURE architecture )
		{
			switch( architecture )
			{
				case PROCESSOR_ARCHITECTURE.INTEL:
				case PROCESSOR_ARCHITECTURE.IA32_ON_WIN64:
					return ProcessorArchitecture.X86;
				case PROCESSOR_ARCHITECTURE.IA64:
					return ProcessorArchitecture.IA64;
				case PROCESSOR_ARCHITECTURE.MSIL:
					return ProcessorArchitecture.MSIL;
				case PROCESSOR_ARCHITECTURE.AMD64:
					return ProcessorArchitecture.Amd64;
				default:
					return ProcessorArchitecture.None;
			}
		}
	}
}

namespace IntelliSite.CommonUtil.SystemInfoApi
{
	internal static class Native
	{
		[DllImport( "kernel32.dll", SetLastError = true )]
		internal static extern void GetSystemInfo( out SYSTEM_INFO lpSystemInfo );

		[DllImport( "kernel32.dll", SetLastError = true )]
		internal static extern void GetNativeSystemInfo( out SYSTEM_INFO lpSystemInfo );
	}

	internal enum PROCESSOR_ARCHITECTURE : ushort
	{
		INTEL = 0, // Intel x86
		MIPS = 1,
		ALPHA = 2,
		PPC = 3, // PowerPC
		SHX = 4,
		ARM = 5,
		IA64 = 6, // Intel Itanium
		ALPHA64 = 7,
		MSIL = 8, // .NET Framework CIL/MSIL
		AMD64 = 9, // x86-64
		IA32_ON_WIN64 = 10, // x86 on 64-bit Windows
		NEUTRAL = 11,
		UNKNOWN = 0xffff
	}

	[StructLayout( LayoutKind.Sequential )]
	internal struct SYSTEM_INFO
	{
		public ushort wProcessorArchitecture;
		public ushort wReserved;
		public uint dwPageSize;
		public IntPtr lpMinimumApplicationAddress;
		public IntPtr lpMaximumApplicationAddress;
		public UIntPtr dwActiveProcessorMask;
		public uint dwNumberOfProcessors;
		public uint dwProcessorType;
		public uint dwAllocationGranularity;
		public ushort wProcessorLevel;
		public ushort wProcessorRevision;
	}
}