﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using IntelliSite.WinForms.VideoView;

namespace StandaloneVideoView
{
	public partial class AvailableCamerasForm : Form
	{
		public MainForm MainForm { get; set; }

		public bool Adding { get; set; }
		public bool Editing { get; set; }

		public BindingList<EditProperty> EditProperties { get; set; }

		public AvailableCamerasForm( MainForm mainForm )
		{
			InitializeComponent();

			MainForm = mainForm;

			EditProperties = new BindingList<EditProperty>();
			propertiesGrid.DataSource = EditProperties;

			LoadTree();
			EnableControls();
		}

		#region public void LoadTree()
		public void LoadTree()
		{
			tree.Nodes.Clear();
			foreach( AvailableCamera data in MainForm.Configuration.AvailableCameras )
				tree.Nodes.Add( CreateTreeNode( data ) );

		}
		#endregion
		#region public TreeNode CreateTreeNode( AvailableCamera data )
		public TreeNode CreateTreeNode( AvailableCamera data )
		{
			TreeNode treeNode = new TreeNode();
			treeNode.Text = data.Name;
			treeNode.Tag = data;

			foreach( AvailableCamera child in data.Children )
				treeNode.Nodes.Add( CreateTreeNode( child ) );

			return treeNode;
		}
		#endregion

		#region public void EnableControls()
		public void EnableControls()
		{
			bool isPreset = tree.SelectedNode != null && ((AvailableCamera)tree.SelectedNode.Tag).IsCamera;

			tree.Enabled = !Editing;
			name.Enabled = Editing;
			nameLabel.Enabled = Editing;
			isCamera.Enabled = Editing && !isPreset;
			playbackEnabled.Enabled = Editing && isCamera.Checked;
			ptzEnabled.Enabled = Editing && isCamera.Checked;
			propertiesLabel.Enabled = Editing;
			propertiesGrid.Enabled = Editing;

			addDataButton.Enabled = Editing || tree.SelectedNode == null || !((AvailableCamera)tree.SelectedNode.Tag).IsPreset;
			addDataButton.Text = Editing ? "Save" : "Add";
			editDataButton.Text = Editing ? "Cancel" : "Edit";
			deleteDataButton.Enabled = !Editing;
		}
		#endregion

		#region private void OnAfterSelect( object sender, TreeViewEventArgs e )
		private void OnAfterSelect( object sender, TreeViewEventArgs e )
		{
			if( tree.SelectedNode != null )
			{
				AvailableCamera data = tree.SelectedNode.Tag as AvailableCamera;

				name.Text               = data.Name;
				isCamera.Checked        = data.IsCamera;
				playbackEnabled.Checked = data.PlaybackEnabled;
				ptzEnabled.Checked      = data.PTZEnabled;

				EditProperties.Clear();
				foreach( KeyValuePair<string,string> kvp in data.PropertyDictionary )
					EditProperties.Add( new EditProperty( kvp.Key, kvp.Value ) );

				EnableControls();
			}
		}
		#endregion
		#region private void OnIsCameraChanged( object sender, EventArgs e )
		private void OnIsCameraChanged( object sender, EventArgs e )
		{
			EnableControls();
		}
		#endregion
		#region private void OnAddData( object sender, EventArgs e )
		private void OnAddData( object sender, EventArgs e )
		{
			if( !Editing )
			{
				Adding = true;
				name.Text = String.Empty;
				isCamera.Checked = false;
				playbackEnabled.Checked = false;
				ptzEnabled.Checked = false;
				EditProperties.Clear();
			}
			else
			{
				if( Adding  )
				{
					AvailableCamera newData = new AvailableCamera( name.Text, isCamera.Checked, playbackEnabled.Checked, ptzEnabled.Checked );
					foreach( EditProperty editProperty in EditProperties )
						newData.PropertyDictionary[editProperty.Name] = editProperty.Value;

					if( tree.SelectedNode == null || MessageBox.Show( "Do you want to add this as a child of " + tree.SelectedNode.Text + "?", "Confirm Add Location", MessageBoxButtons.YesNo ) == DialogResult.No )
					{
						MainForm.Configuration.AvailableCameras.Add( newData );
						tree.Nodes.Add( CreateTreeNode( newData ) );
					}
					else
					{
						newData.IsPreset = tree.SelectedNode != null && ((AvailableCamera)tree.SelectedNode.Tag).IsCamera;
						((AvailableCamera)tree.SelectedNode.Tag).Children.Add( newData );
						tree.SelectedNode.Nodes.Add( CreateTreeNode( newData ) );
					}
						
					Adding = false;
				}
				else
				{
					AvailableCamera data = tree.SelectedNode.Tag as AvailableCamera;
					data.Name = name.Text;
					data.IsCamera = isCamera.Checked;
					data.PlaybackEnabled = playbackEnabled.Checked;
					data.PTZEnabled = ptzEnabled.Checked;

					data.Properties.Clear();
					foreach( EditProperty editProperty in EditProperties )
						data.PropertyDictionary[editProperty.Name] = editProperty.Value;

					tree.SelectedNode.Text = name.Text;
				}
			}

			Editing = !Editing;
			EnableControls();
		}
		#endregion
		#region private void OnEditData( object sender, EventArgs e )
		private void OnEditData( object sender, EventArgs e )
		{
			if( Editing )
				OnAfterSelect( this, new TreeViewEventArgs( tree.SelectedNode ) );

			Editing = !Editing;
			EnableControls();
		}
		#endregion
		#region private void OnDeleteData( object sender, EventArgs e )
		private void OnDeleteData( object sender, EventArgs e )
		{
			if( tree.SelectedNode != null && MessageBox.Show( "Are you sure you want to delete " + ((AvailableCamera)tree.SelectedNode.Tag ).Name + "?", "Confirm Delete", MessageBoxButtons.YesNo ) == DialogResult.Yes )
			{
				TreeNode selected = tree.SelectedNode;
				if( selected.Parent != null )
				{
					( (AvailableCamera)selected.Parent.Tag ).Children.Remove( selected.Tag );
					selected.Parent.Nodes.Remove( selected );
				}
				else
				{
					MainForm.Configuration.AvailableCameras.Remove( selected.Tag );
					tree.Nodes.Remove( selected );
				}
			}
		}
		#endregion

		#region private void OnNew( object sender, EventArgs e )
		private void OnNew( object sender, EventArgs e )
		{
			if( MessageBox.Show( "Are you sure you want to create a new configuration?  All unsaved changes will be lost!", "Confirm New Configuration", MessageBoxButtons.OKCancel ) == DialogResult.OK )
			{
				MainForm.Configuration = new Configuration();
				LoadTree();
			}
		}
		#endregion
		#region private void OnLoad( object sender, EventArgs e )
		private void OnLoad( object sender, EventArgs e )
		{
			OpenFileDialog openFileDialog = new OpenFileDialog();
			openFileDialog.InitialDirectory = Environment.GetFolderPath( Environment.SpecialFolder.MyDocuments );
			openFileDialog.CheckFileExists = true;
			openFileDialog.Filter = "JSON Files (*.json)|*.json";
			openFileDialog.DefaultExt = "json";
			if( openFileDialog.ShowDialog() == DialogResult.OK && MainForm.LoadConfiguration( openFileDialog.FileName ) )
			{
				MainForm.LastConfigFileName = openFileDialog.FileName;
				LoadTree();
				//ProductInfo.Current.ReadProductKey( RegistryHive.CurrentUser, "DesktopClient", key =>
				//{
				//	if( key != null )
				//		key.SetValue( "Last Simulation", LastSimulationFileName );
				//} );
			}
		}
		#endregion
		#region private void OnSave( object sender, EventArgs e )
		private void OnSave( object sender, EventArgs e )
		{
			MainForm.Configuration.VideoLayout = MainForm.videoView.VideoLayout;
			MainForm.Configuration.CameraNames.Clear();
			foreach( VideoControl videoControl in MainForm.videoView.Controls )
				MainForm.Configuration.CameraNames.Add( videoControl.CameraName );

			SaveFileDialog saveFileDialog = new SaveFileDialog();
			saveFileDialog.InitialDirectory = Environment.GetFolderPath( Environment.SpecialFolder.MyDocuments );
			saveFileDialog.Filter = "JSON Files (*.json)|*.json";
			saveFileDialog.DefaultExt = "json";
			if( saveFileDialog.ShowDialog() == DialogResult.OK )
			{
				if( !saveFileDialog.CheckFileExists || MessageBox.Show( this, "Overwrite existing file?", "Confirm Save Overwrite", MessageBoxButtons.YesNo ) == DialogResult.Yes )
				{
					MainForm.Configuration.Name = saveFileDialog.FileName;
					MainForm.SaveConfiguration();
				}
			}
		}
        #endregion

        private string GetProperty(AvailableCamera data, string propertyName, bool required = true)
        {
            string retval = null;
            EditProperty editProperty = data.Properties.Find(x => x.Name == propertyName);

            if (editProperty != null)
            {
                retval = editProperty.Value;
            }
            else
            {
               if (required) MessageBox.Show("Missing Property: " + propertyName+ ". You must supply the " + propertyName, "Missing Property", MessageBoxButtons.OK);
            }

            return retval;
        }

		private void OnDiscover( object sender, EventArgs e )
		{
            if (tree.SelectedNode == null) return;
            
            AvailableCamera data = tree.SelectedNode.Tag as AvailableCamera;
            if( data.IsCamera )
            {
                MessageBox.Show("Discover only works on NVRs, not cameras.", "Wrong Tree Node Selected", MessageBoxButtons.OK);
                return;
            }
            string driverName = GetProperty(data, "Driver");
            if (driverName == null) return;

            string Address = GetProperty(data, "Address");
            if (Address == null) return;

            string Port = GetProperty(data,"Port");
            if (Port == null) return;

            string Username = GetProperty(data, "Username");
            if (Username == null) return;

            string Password = GetProperty(data, "Password");
            if (Password == null)  return;
            
            string AuthenticationMethod = GetProperty(data, "AuthenticationMethod", false);  // used by some NVRs
            string Protocol = GetProperty(data, "Protocol", false);                          // used by some NVRs

            System.Reflection.Assembly driverDll;
            System.Reflection.Assembly dataDll;

            switch ( driverName )
            {
                case "DigitalWatchdog" :

                    if (Protocol == null) Protocol = "rtsp";
                    string driverDllPath = Path.Combine(".\\Drivers", driverName, driverName +".Driver.dll");
                    if( !File.Exists(driverDllPath) )
                    {
                        MessageBox.Show("Did not find " + driverDllPath, "Missing DLL", MessageBoxButtons.OK);
                        return;
                    }
                    string dataDllPath = Path.Combine(".\\Drivers", driverName, driverName + ".Data.dll");
                    if (!File.Exists(dataDllPath))
                    {
                        MessageBox.Show("Did not find " + dataDllPath, "Missing DLL", MessageBoxButtons.OK);
                        return;
                    }

                    try
                    {
                        driverDll = Assembly.LoadFrom(driverDllPath);
                        dataDll = Assembly.LoadFrom(dataDllPath);
                    }
                    catch (Exception exception)
                    {
                        MessageBox.Show(exception.Data.ToString(), "Error", MessageBoxButtons.OK);
                        return;
                    }
                    //
                    // Create the panel (RTU) object and fill in the required properties to connect to the NVR.
                    //
                    string temp = "IntelliSite." + driverName + ".Data.IS" + driverName + "RTU";
                    Type paneltype = dataDll.GetType(temp);
                    object panel = Activator.CreateInstance(paneltype);

                    PropertyInfo panelPropertyInfo = paneltype.GetProperty("IPAddress");
                    panelPropertyInfo.SetValue(panel, Address);

                    panelPropertyInfo = paneltype.GetProperty("Port");
                    UInt16 port;
                    if (UInt16.TryParse(Port, out port)) panelPropertyInfo.SetValue(panel, port);
                    else
                    {
                        MessageBox.Show("Invalid Port value: " + Port, "Error", MessageBoxButtons.OK);
                        return;
                    }

                    panelPropertyInfo = paneltype.GetProperty("Username");
                    panelPropertyInfo.SetValue(panel, Username);

                    panelPropertyInfo = paneltype.GetProperty("Password");
                    panelPropertyInfo.SetValue(panel, Password);

                    panelPropertyInfo = paneltype.GetProperty("Protocol");
                    var protocolType = panelPropertyInfo.PropertyType;
                    object protocol = Enum.Parse(protocolType, Protocol);
                    panelPropertyInfo.SetValue(panel, Enum.Parse(panelPropertyInfo.PropertyType, Protocol));


                    break;
                default:
                    MessageBox.Show("Discover is not yet implemented for " + driverName, "Wrong Driver Type", MessageBoxButtons.OK);
                    return;
            }
		}
	}
}
