﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.Serialization.Json;
using System.Windows.Forms;
using Microsoft.Win32;
using IntelliSite.CommonUtil;
using IntelliSite.WinForms.VideoView;

namespace StandaloneVideoView
{
	public partial class MainForm : Form
	{
		public Configuration Configuration { get; set; }
		public string LastConfigFileName { get; set; }
		public LoggerBase Logger { get; set; }

		public MainForm()
		{
			ProductInfo.LoadFromBrandedRegistry = false;

			Logger = new LoggerBase( OutputDebug );

			InitializeComponent();

			Configuration = new Configuration();
			videoView.AvailableCameras = Configuration.AvailableCameras;
			videoView.SetAvailableCameraType( typeof( AvailableCamera ) );
            videoView.SetPresetType(typeof(AvailableCamera));
			videoView.Exception += OnException;

            LastConfigFileName = LoadLastConfiguration();
			if( !String.IsNullOrEmpty( LastConfigFileName ) )
				LoadConfiguration( LastConfigFileName );

            // Load the digitalwatchdog.driver.dll
            string currentDirectory = Directory.GetCurrentDirectory();
            string assemblyFile = Path.Combine(currentDirectory, @"Drivers\DigitalWatchdog\DigitalWatchdog.Driver.dll");
            ProductInfo.Current.LoadAssemblyFromFile(assemblyFile);

		}

		private void OnException( object sender, ErrorArgs e )
		{
			MessageBox.Show( String.Format( "{0}: {1}", sender.ToString(), e.ToString() ) );
		}

		#region private void OnSizeChanged( object sender, EventArgs e )
		private void OnSizeChanged( object sender, EventArgs e )
		{
			videoView.ResizeControls();
		}
		#endregion

		#region public object FindCameraByName( AvailableCamera parent, string name )
		public object FindCameraByName( AvailableCamera parent, string name )
		{
			if( parent == null )
			{
				foreach( AvailableCamera camera in Configuration.AvailableCameras )
				{
					object find = FindCameraByName( camera, name );
					if( find != null )
						return find;
				}
			}
			else
			{
				if( parent.IsCamera && parent.Name == name )
					return parent;

				foreach( AvailableCamera child in parent.Children )
				{
					object find = FindCameraByName( child, name );
					if( find != null )
						return find;
				}
			}

			return null;
		}
		#endregion

		#region internal bool LoadConfiguration( string fileName )
		internal bool LoadConfiguration( string fileName )
		{
			if( File.Exists( fileName ) )
			{
				try
				{
					DataContractJsonSerializer serializer = new DataContractJsonSerializer( typeof( Configuration ) );
					using( FileStream fileStream = new FileStream( fileName, FileMode.Open ) )
					{
						Configuration configuration = (Configuration)serializer.ReadObject( fileStream );
						Configuration = configuration;
						LastConfigFileName = fileName;
						SaveLastConfiguration();

						videoView.AvailableCameras = Configuration.AvailableCameras;
						videoView.SetLayout( configuration.VideoLayout );
						if( configuration.CameraNames == null )
							configuration.CameraNames = new List<string>();

						for( int i=0; i<configuration.CameraNames.Count && i < videoView.Controls.Count; i++ )
						{
							object camera = FindCameraByName( null, configuration.CameraNames[i] );
							((VideoControl)videoView.Controls[i]).CameraInfo = camera ?? ((VideoControl)videoView.Controls[i]).NoCamera;
						}
					}
				}
				catch( Exception e )
				{
					MessageBox.Show( "Load failed for: " + Environment.NewLine + fileName + Environment.NewLine + Environment.NewLine + e.ToString() );
					return false;
				}
			}
			else
			{
				MessageBox.Show( "File: " + fileName + " is does not exist." );
				return false;
			}		
			
			return true;
		}
		#endregion
		#region internal void SaveConfiguration()
		internal void SaveConfiguration()
		{
			DataContractJsonSerializer serializer = new DataContractJsonSerializer( typeof( Configuration ) );
			using( FileStream fileStream = new FileStream( Configuration.Name, FileMode.Create ) )
			{
				serializer.WriteObject( fileStream, Configuration );
			}

			LastConfigFileName = Configuration.Name;
			SaveLastConfiguration();
		}
		#endregion

		#region private void SaveLastConfiguration()
		private void SaveLastConfiguration()
		{
			using( RegistryKey hkcu = Registry.CurrentUser )
			using( RegistryKey software = hkcu.OpenSubKey( "Software" ) )
			{
				RegistryKey ossi = software.OpenSubKey( "OSSI", true );
				if( ossi == null )
					ossi = software.CreateSubKey( "OSSI", RegistryKeyPermissionCheck.ReadWriteSubTree );

				RegistryKey key = ossi.OpenSubKey( "StandaloneVideoView", true );
				if( key == null )
					key = ossi.CreateSubKey( "StandaloneVideoView", RegistryKeyPermissionCheck.ReadWriteSubTree );

				key.SetValue( "Last Configuration", LastConfigFileName );

				key.Dispose();
				ossi.Dispose();
			}
		}
		#endregion
		#region private string LoadLastConfiguration()
		private string LoadLastConfiguration()
		{
			string retVal = String.Empty;
			using( RegistryKey hkcu = Registry.CurrentUser )
			using( RegistryKey software = hkcu.OpenSubKey( "Software" ) )
			{
				RegistryKey ossi = software.OpenSubKey( "OSSI" );
				if( ossi != null )
				{
					RegistryKey key = ossi.OpenSubKey( "StandaloneVideoView" );
					if( key != null )
					{
						retVal = (string)key.GetValue( "Last Configuration" );
						key.Dispose();
					}
					ossi.Dispose();
				}
			}
			return retVal;
		}
		#endregion

		#region protected override void WndProc(ref Message m)
		protected override void WndProc(ref Message m)
		{
			const int WM_NCRBUTTONDOWN = 0xA4;

			if (m.Msg == WM_NCRBUTTONDOWN)
			{
				AvailableCamerasForm form = new AvailableCamerasForm( this );
				form.ShowDialog();
			}
			else
			{
				base.WndProc(ref m);
			}
		}
		#endregion

		#region private void OutputDebug( object sender, LogMessageArgs args )
		public delegate void AddDebugTextDelegate( string output );
		private void OutputDebug( object sender, LogMessageArgs args )
		{
			Console.WriteLine( args.Message );
		}
		#endregion
	}
}
