﻿using System;
using System.IO;
using Newtonsoft.Json;

namespace IntelliSite.DigitalWatchdog.Driver
{

    public class PtzReply
    {


        #region Properties

        public decimal x { get; set; }
        public decimal y { get; set; }
        public decimal z { get; set; }

        #endregion


    }

    public class CommandResult
    {


        #region Properties

        public string error { get; set; }
        public string errorString { get; set; }
        public PtzReply reply { get; set; }

        #endregion


    }

    public class PtzParameters
    {


        #region Constants

        public const string API_PTZ_COMMAND = "api/ptz";
        public const decimal PAN_STEP = .1M;
        public const decimal SPEED_STEP = .01M;
        public const decimal TILT_STEP = .1M;
        public const decimal ZOOM_STEP = .01M;

        public const string API_PTZ_COMMAND_PARAMETER_GET_PRESETS = "GetPresetsPtzCommand";
        public const string API_PTZ_COMMAND_PARAMETER_GET_DEVICE_POSITION = "GetDevicePositionPtzCommand";

        #endregion


        #region Private member variables

        private decimal _x;
        private decimal _y;
        private decimal _z;
        private decimal _speed;

        #endregion


        #region Constructors

        public PtzParameters()
        {
            this.X = 0.0M;
            this.Y = 0.0M;
            this.Z = 0.0M;
            this.Speed = 1;
        }

        public PtzParameters(string response)
        {
            this.OriginalMessage = response;
            // Format: {"error": "0","errorString": "","reply": {"x": 0,"y": 0,"z": 0}}
            // Format: {"error": "0","errorString": "","reply": null}
            using (StringReader reader = new StringReader(response))
            {
                try
                {
                    CommandResult result = JsonConvert.DeserializeObject<CommandResult>(response);
                    if (result != null)
                    {
                        this.Error = result.error;
                        this.ErrorString = result.errorString;
                        if (result.reply != null)
                        {
                            X = result.reply.x;
                            Y = result.reply.y;
                            Z = result.reply.z;
                        }
                        Speed = 1;
                    }
                }
                catch (Exception ex)
                {
                    throw new ArgumentException("JSON parsing error.  See inner exception or check the format of the response: \"" + response + "\"", ex);
                }
            }
        }

        #endregion


        #region Properties

        public string Error { get; set; }

        public string ErrorString { get; set; }

        public bool HasError { get { return !String.IsNullOrWhiteSpace(this.ErrorString); } }

        public string OriginalMessage { get; set; }

        public decimal Speed
        {
            get
            {
                return this._speed;
            }
            set
            {
                this._speed = value;
                if (this._speed < 0.0M)
                {
                    this._speed = 0.0M;
                }
                else if (this._speed > 1.0M)
                {
                    this._speed = 1.0M;
                }

            }
        }

        public decimal X
        {
            get { return this._x; }
            private set
            {
                this._x = value;
                if (this._x < -1.0M)
                {
                    this._x = 1.0M;
                }
                else if (this._x > 1.0M)
                {
                    this._x = 1.0M;
                }
            }
        }

        public decimal Y
        {
            get { return this._y; }
            private set
            {
                this._y = value;
                if (this._y < -1.0M)
                {
                    this._y = 1.0M;
                }
                else if (this._y > 1.0M)
                {
                    this._y = 1.0M;
                }
            }
        }

        public decimal Z
        {
            get { return this._z; }
            private set
            {
                this._z = value;
                if (this._z < 0.0M)
                {
                    this._z = 0.0M;
                }
                else if (this._z > 1.0M)
                {
                    this._z = 1.0M;
                }
            }
        }

        #endregion


        #region Overridden base methods

        public override string ToString()
        {
            return String.Format("Pan: {0}, Tilt: {1}, Zoom: {2}, Speed: {3}", this.X, this.Y, this.Z, this.Speed);
        }

        #endregion


        #region Public methods

        public void IncrementPan()
        {
            this.X += PAN_STEP;
        }

        public void DecrementPan()
        {
            this.X -= PAN_STEP;
        }

        public void IncrementSpeed()
        {
            this.Speed += SPEED_STEP;
        }

        public void DecrementSpeed()
        {
            this.Speed -= SPEED_STEP;
        }

        public void IncrementTilt()
        {
            this.Y += TILT_STEP;
        }

        public void DecrementTilt()
        {
            this.Y -= TILT_STEP;
        }

        public void IncrementZoom()
        {
            this.Z += ZOOM_STEP;
        }

        public void DecrementZoom()
        {
            this.Z -= ZOOM_STEP;
        }

        #endregion


    }

}
