﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using IntelliSite.CommonUtil;

namespace IntelliSite.VideoCommon
{
	#region public enum PTZCommands
	public enum PTZCommands
	{
		PTZUp,
		PTZDown,
		PTZLeft,
		PTZRight,
		PTZUpLeft,
		PTZUpRight,
		PTZDownLeft,
		PTZDownRight,
		PTZHome,
		PTZSpeedUp,
		PTZSpeedDown,
		ZoomIn,
		ZoomOut,
		ZoomSpeedUp,
		ZoomSpeedDown,
		FocusNear,
		FocusFar,
		FocusSpeedUp,
		FocusSpeedDown,
		IrisOpen,
		IrisClose,
		IrisSpeedUp,
		IrisSpeedDown
	}
	#endregion

	#region Video EventArgs classes
	public class TimeChangeEventArgs : EventArgs
	{
		public DateTime OldTime { get; set; }
		public DateTime NewTime { get; set; }
	}
	#endregion

	//
	// The base clase for all Cameras in Intelli-Site
	//
	public class ISCameraBase : IDisposable
	{
		public Type CameraType { get; set; }
		public PropertyInfo PropertiesProperty { get; set; }
		public PropertyInfo NameProperty { get; set; }
		public PropertyInfo ValueProperty { get; set; }

		public object CameraInfo { get; set; } 
		public virtual string CameraName { get; set; }
		public virtual string UserName { get; set; }
		public virtual string Password { get; set;  }
		public virtual bool IsConnected { get; set; }
		public virtual bool IsPTZConnected { get; set; }
		public virtual bool IsRecording { get; set; }
		public virtual bool InPlaybackMode { get; set; }
		public virtual object Parent { get; set; }
		public virtual DateTime PlaybackStartTime { get; set; }
		public virtual DateTime PlaybackEndTime { get; set; }
		public virtual DateTime PlaybackCurrentTime { get; set; }
		public virtual System.Windows.Forms.Control VideoViewerWinForms { get; set; }
		public event EventHandler<VideoStatusArgs> VideoStatus = delegate {};
		public event EventHandler<VideoStatusArgs> VideoSetupInfo = delegate {};
		public event EventHandler<EventArgs> Streaming = delegate {};
		public event EventHandler<EventArgs> PTZAvailable = delegate {};
		public LoggerBase Logger { get; set; }
		public event EventHandler<AutoDetectProgressArgs> AutoDetectProgress = delegate { };
		public event EventHandler<EventArgs> AutoDetectionComplete = delegate { };
		public List<Dictionary<string,string>> AutoDetectedCameras;

        private string _message;
        public string Message
        {
            get
            {
                return this._message;
            }
            set
            {
                this._message = value;
                this.OnVideoStatus(this, new VideoStatusArgs(value));
            }
        }

		#region public static Type GetTypeEx( string fullTypeName )
		public static Type GetTypeEx( string fullTypeName )
		{
			return Type.GetType( fullTypeName ) ??
				   AppDomain.CurrentDomain.GetAssemblies()
							.Select( a => a.GetType( fullTypeName ) )
							.FirstOrDefault( t => t != null );
		}
		#endregion
		#region public static Type GetTypeFromName( string fullTypeName )
		public static Type GetTypeFromName( string fullTypeName )
		{
			Type type = GetTypeEx( fullTypeName );
			if( type == null )
			{
				string assemblyName = fullTypeName;
				if( assemblyName.StartsWith( "IntelliSite." ) )
					assemblyName = assemblyName.Substring( 12 );
				int index = assemblyName.LastIndexOf( "." );
				assemblyName = assemblyName.Substring( 0, index );

				ProductInfo.Current.LoadAssemblyFromName( assemblyName );
				type = GetTypeEx( fullTypeName );
			}
			return type;
		}
		#endregion
		#region public static ISCameraBase CreateDriver( string driverName, LoggerBase logger )
		public static ISCameraBase CreateDriver( string driverName, LoggerBase logger )
		{
			string cameraClassName = driverName;
			if( !cameraClassName.StartsWith( "IntelliSite" ) )
				cameraClassName = "IntelliSite." + driverName + ".Driver." + driverName + "Camera";
			Type driverType = GetTypeFromName( cameraClassName );
			if( driverType != null )
			{
				try
				{
					object driverObj = Activator.CreateInstance( driverType, new object[] { logger } );
					ISCameraBase driver = driverObj as ISCameraBase;
					driver.Init();
					return driver;
				}
				catch( Exception e )
				{
					Console.WriteLine( "Could not create type {0}, {1}", cameraClassName, e.ToString() );
				}
			}
			return null;
		}
		#endregion
		#region public string CameraProperty( string name )
		public string CameraProperty( string name )
		{
			if( CameraInfo != null )
			{
			//	Check to see if the CameraInfo has a specific property of the requested name.  If so, return it.
				PropertyInfo propertyInfo = CameraInfo.GetType().GetProperty( name.Replace( " ", null ) );
				if( propertyInfo != null )
					return propertyInfo.GetValue( CameraInfo, null ).ToString();

			//	Check the list of GenProto "Properties" for the requested name.
				IEnumerable properties = PropertiesProperty.GetValue( CameraInfo, null ) as IEnumerable;

				foreach( object property in properties )
				{
					if( NameProperty == null )
					{
						NameProperty = property.GetType().GetProperty( "Name" );
						ValueProperty = property.GetType().GetProperty( "Value" );
					}

					string propertyName = String.Empty;
					if( NameProperty != null )
						propertyName = (string)NameProperty.GetValue( property, null );

					if( propertyName == name )
					{
						string propertyValue = String.Empty;
						if( ValueProperty != null )
							propertyValue = (string)ValueProperty.GetValue( property, null );
						return propertyValue;
					}
				}

			}
			return String.Empty;
		}
		#endregion

		#region Supported Features
		public virtual bool IsPTZSupported { get; }
		public virtual bool IsRecordingSupported { get; }
		public virtual bool IsPlaybackSupported { get; }
		public virtual bool IsAudioReceiveSupported { get; }
		public virtual bool IsAudioTransmitSupported { get; }
		public virtual bool IsAlarmsSupported { get; }
		#endregion

		#region Event Handlers
		public event EventHandler<TimeChangeEventArgs> PlaybackCurrentTimeChanged = delegate { };
		public event EventHandler ConnectionLost = delegate { };
		public event EventHandler<ErrorArgs> Exception = delegate { };
		#endregion

		public ISCameraBase( LoggerBase logger=null )
		{
			Logger = logger;
			CameraName = null;
			UserName = null;
			Password = null;
			InPlaybackMode = false;
			disposed = false;
		}

		#region Destructor/Dispose
		~ISCameraBase()
		{
			Dispose( false );
		}

		public void Dispose()
		{
			Dispose( true );
		}

		private bool disposed = false;
		protected void Dispose( bool disposing )
		{
			if( !disposed )
			{
				if( disposing )
				{
				}
			}
		}
		#endregion		

		public virtual void Init() { }
		#region public virtual void ChangeCamera( object cameraInfo )
		public virtual void ChangeCamera( object cameraInfo )
		{
			if( CameraType != cameraInfo.GetType() )
			{
				PropertiesProperty = cameraInfo.GetType().GetProperty( "Properties" );
				NameProperty       = null;
				ValueProperty      = null;
			}

			CameraInfo = cameraInfo;
			CameraType = cameraInfo.GetType();

			CameraName = null;
			PropertyInfo nameProperty = cameraInfo.GetType().GetProperty( "Name" );
			if( nameProperty != null )
				CameraName = (string)nameProperty.GetValue( cameraInfo, null );

			UserName = CameraProperty( "Username" );
			Password = CameraProperty( "Password" );
		}
		#endregion
		public virtual void ChangeSize(System.Drawing.Size size) { }
		public virtual void ChangeLocation (System.Drawing.Point point) { }
		#region public virtual bool Connect( object parent )
		public virtual bool Connect( object parent )
		{
			Parent = parent;
			InPlaybackMode = false;
			IsConnected = true;
			return IsConnected;
		}
		#endregion
		public virtual void Reconnect() {}
		public virtual void Disconnect() { IsConnected = false; }
		public virtual void Snapshot() {  }
		public virtual void Record() { }

		#region Live Streaming Commands
		public virtual bool Start() { return IsConnected; }
		public virtual bool Stop() { return true; }
		public virtual void Live() { if (InPlaybackMode) { if (!IsConnected) Connect( Parent ); Start(); InPlaybackMode = false; } }
		protected void OnVideoStatus( object sender, VideoStatusArgs args ) { VideoStatus( sender, args ); }
		protected void OnVideoSetupInfo( object sender, VideoStatusArgs args ) { VideoSetupInfo( sender, args ); }
		protected void OnStreaming( object sender, EventArgs args ) { Streaming( sender, args ); }
		#endregion

		#region PTZ Commands
		public virtual bool PTZConnect() { if( IsPTZSupported ) IsPTZConnected = true; return IsPTZConnected; }
		public virtual void PTZDisconnect() { if( IsPTZSupported && IsPTZConnected ) IsPTZConnected = false; }
		public virtual void PTZ(PTZCommands ptzCommand, bool start) { }
		protected void OnPTZAvailable( object sender, EventArgs args ) { PTZAvailable( sender, args ); }
		public virtual void GoToPreset( string presetInfo ) { }
		#endregion

		public virtual void AutoDetect( object autoDetectParameters ) {}
		protected void OnAutoDetectProgress( object sender, AutoDetectProgressArgs args ) { AutoDetectProgress( sender, args ); }
		protected void OnAutoDetectionComplete( object sender, EventArgs args ) { AutoDetectionComplete( sender, args ); }

		#region Playback Commands
		public virtual void Play() { }
		public virtual void Pause() { }
		public virtual void Forward() { }
		public virtual void Reverse() { }
		public virtual void FastFoward() { }
		public virtual void FastReverse() { }

		public virtual void SaveClip() {  }
		#endregion

		#region Alarms
		public virtual List<VideoAlarm> GetAlarms(DateTime startTime, DateTime endTime, bool systemAlarms, bool bookmarks, bool events, bool alarms) { return null; }
		#endregion

		#region internal EventHandler helpers

		public void OnPlaybackCurrentTimeChanged(object sender, TimeChangeEventArgs args) {

			// Only set the playback start/end times if the new date is outside the current playback start/end time range.
			if ((DateTime.Compare(this.PlaybackStartTime, args.NewTime) > 0) || (DateTime.Compare(args.NewTime, this.PlaybackEndTime) > 0))
			{
				this.PlaybackStartTime = args.NewTime.AddHours(-2);
				this.PlaybackEndTime = args.NewTime.AddHours(2);
			}

			PlaybackCurrentTimeChanged(sender, args);
		}

		public void OnConnectionLost(object sender, EventArgs args) { ConnectionLost(sender, args); }
		public void OnException(object sender, ErrorArgs args) { Exception(sender, args); }
		#endregion
	}


}
