﻿using System;
using System.Collections;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;
using IntelliSite.VideoCommon;

namespace IntelliSite.WinForms.VideoView
{
	public partial class ToolsForm : Form
	{
		public VideoControl Control { get; set; }
		private Image RecordImage { get; set; }
		private Image StopRecordImage { get; set; }
		public bool MouseInForm { get; set; }

		public ToolsForm( VideoControl control )
		{
			Control = control;

			InitializeComponent();

			SetStyle( ControlStyles.SupportsTransparentBackColor, true );
			BackColor = Color.Transparent;
			TransparencyKey = Color.Transparent;

			Assembly assembly = Assembly.GetAssembly( typeof( IntelliSite.WinForms.VideoView.VideoControl ) );
			RecordImage       = Image.FromStream( assembly.GetManifestResourceStream( "IntelliSite.WinForms.VideoView.Images.record_24.png" ) );
			StopRecordImage   = Image.FromStream( assembly.GetManifestResourceStream( "IntelliSite.WinForms.VideoView.Images.record_disabled_24.png" ) );

			OnCameraChanged();
			RecordingChanged();
			EnableControls();
		}

		protected override void OnPaintBackground( PaintEventArgs e ) {}

		public float SecondsPerPixel() { return ( 240.0f * 60.0f ) / (float)Width; }

		#region public void EnableControls()
		public void EnableControls()
		{
			liveButton.Visible        = !Control.View.DisableConfigChange && Control.PlaybackEnabled && Control.PlaybackMode;
			fastReverseButton.Visible = !Control.View.DisableConfigChange && Control.PlaybackEnabled;
			reverseButton.Visible     = !Control.View.DisableConfigChange && Control.PlaybackEnabled;
			playButton.Visible        = !Control.View.DisableConfigChange && Control.PlaybackEnabled;
			forwardButton.Visible     = !Control.View.DisableConfigChange && Control.PlaybackEnabled;
			fastForwardButton.Visible = !Control.View.DisableConfigChange && Control.PlaybackEnabled;

			recordingButton.Visible   = !Control.View.DisableConfigChange && Control.PlaybackEnabled;
			back30Button.Visible      = !Control.View.DisableConfigChange && Control.PlaybackEnabled;
			calendarButton.Visible    = !Control.View.DisableConfigChange && Control.PlaybackEnabled;
			alarmsButton.Visible      = !Control.View.DisableConfigChange && Control.PlaybackEnabled;

			saveClipButton.Visible    = !Control.View.DisableConfigChange && Control.PlaybackEnabled;
			snapshotButton.Visible    = Control.Connected;
			ptzButton.Visible         = Control.PTZEnabled;

			settingsButton.Visible    = !Control.View.DisableConfigChange;
			cameraButton.Visible      = !Control.View.DisableCameraChange;

			Invalidate();
		}
		#endregion
		#region public void SetWidth( int width )
		public void SetWidth( int width )
		{
			Size = new Size( width, Height );
			flowLayoutPanel.Size = new Size( width, flowLayoutPanel.Height );
			timelinePanel.Size = new Size( width, timelinePanel.Height );
			controlsPanel.Size = new Size( width, controlsPanel.Height );

			playbackControlsPanel.Location = new Point( ( controlsPanel.Width - playbackControlsPanel.Width ) / 2, playbackControlsPanel.Top );
			additionalPlaybackControlsPanel.Location = new Point( playbackControlsPanel.Right + 24, additionalPlaybackControlsPanel.Top );
		}
		#endregion

		public void ConnectedChanged()		{ EnableControls(); }
		public void RecordingChanged()		{ recordingButton.Image = Control.Recording ? StopRecordImage : RecordImage; }
		public void PlaybackTimeChanged()	{ timelinePanel.Invalidate(); }

		#region private void OnPaintTimeLine( object sender, PaintEventArgs e )
		private void OnPaintTimeLine( object sender, PaintEventArgs e )
		{
			if( Control.View.DisableConfigChange || !Control.PlaybackEnabled )
				return;

			if( Control.PlaybackEnabled )
			{
				DateTime startTime = Control.RulerStartTime();

				using( SolidBrush brush = new SolidBrush( Color.DimGray ) )
					e.Graphics.FillRectangle( brush, new Rectangle( 0, 0, Width, 20 ) );

				//	Draw the ruler
				using( SolidBrush brush = new SolidBrush( Color.CornflowerBlue ) )
					e.Graphics.FillRectangle( brush, new Rectangle( 0, 20, Width, 20 ) );

				//	Draw the tick marks and times
				using( Pen pen = new Pen( Color.LightGray ) )
				{
					DateTime lastTime = startTime;
					float secondsPerPixel = SecondsPerPixel();
					for( int i = 0; i <= Width; i++ )
					{
						DateTime tickTime = startTime.AddSeconds( (int)( secondsPerPixel * (float)i ) );
						
						if( ( lastTime.Minute % 5 ) != 0 && ( tickTime.Minute % 5 ) == 0 )
						{
							if( tickTime.Minute == 0 )
							{
								string hour = tickTime.Hour.ToString() + ":00";
								using( SolidBrush brush = new SolidBrush( Color.LightGray ) )
									e.Graphics.DrawString( hour, DefaultFont, brush, new Point( i + 4, 3 ) );
								e.Graphics.DrawLine( pen, new Point( i + 2, 0 ), new Point( i + 2, timelinePanel.Height- 1 ) );
							}
							else if( ( tickTime.Minute % 15 ) == 0 )
								e.Graphics.DrawLine( pen, new Point( i + 2, 25 ), new Point( i + 2, timelinePanel.Height - 1 ) );
							else
								e.Graphics.DrawLine( pen, new Point( i + 2, 35 ), new Point( i + 2, timelinePanel.Height - 1 ) );
						}

                        // This line is drawn after the others so that it will appear on top of any grey lines.
                        if (Control.PlaybackMode && lastTime < Control.PlaybackTime && tickTime >= Control.PlaybackTime)
                        {
                            using (Pen yellow = new Pen(Color.Yellow))
                                e.Graphics.DrawLine(yellow, new Point(i + 2, 0), new Point(i + 2, timelinePanel.Height - 1));
                        }

                        if ( Control.VideoAlarms != null )
							foreach( VideoAlarm alarm in Control.VideoAlarms )
								if( lastTime < alarm.StartTime && tickTime >= alarm.StartTime )
								{
									Color color = Color.White;
									switch( alarm.AlarmType )
									{
										case VideoAlarmType.SystemAlarm:
											color = Color.Red;
											break;
										case VideoAlarmType.Bookmark:
											color = Color.Green;
											break;
										case VideoAlarmType.Event:
											color = Color.Blue;
											break;
										case VideoAlarmType.Alarm:
											color = Color.Salmon;
											break;
									}
									int width = Math.Max( (int)( ( alarm.EndTime - alarm.StartTime ).TotalSeconds / secondsPerPixel ), 7 );
									alarm.Rect = new Rectangle( new Point( i + 2, 30 ), new Size( width, 10 ) );
									using( SolidBrush brush = new SolidBrush( color ) )
										e.Graphics.FillRectangle( brush, alarm.Rect );
								}

						lastTime = tickTime;
					}
				}
			}
		}
		#endregion

		#region private void OnSettings( object sender, EventArgs e )
		private void OnSettings( object sender, EventArgs e )
		{
			Control.View.ShowSettingsContextMenu( PointToScreen( settingsButton.Location ) );
		}
		#endregion
		#region private void OnCameraComboDropDown( object sender, EventArgs e )
		private void OnCameraComboDropDown( object sender, EventArgs e )
		{
			OnCameraCombo();
		}
		#endregion

		#region internal void OnCameraChanged()
		internal void OnCameraChanged()
		{
			cameraButton.Text = Control.CameraName;
			liveButton.Visible = false;
		}
		#endregion
		#region public ToolStripMenuItem CreateCameraMenuItem( object camera )
		public ToolStripMenuItem CreateCameraMenuItem( object camera )
		{
			string name = "<unknown>";
			if( Control.View.IsAvailableCameraType( camera ) || Control.View.IsPreset( camera ) )
			{
				if( Control.View.NameProperty != null )
					name = (string)Control.View.NameProperty.GetValue( camera, null );
			}
			else
				name = camera.ToString();

			ToolStripMenuItem newCamera = new ToolStripMenuItem( name );
			newCamera.Tag = camera;
			newCamera.Click += OnCameraClick;
			return newCamera;
		}
		#endregion
		#region public void AddMenuItems( ToolStripMenuItem menuItem, object camera )
		public void AddMenuItems( ToolStripMenuItem menuItem, object camera )
		{
			if( Control.View.IsAvailableCameraType( menuItem.Tag ) && ( Control.View.IsCameraProperty == null || (bool)Control.View.IsCameraProperty.GetValue( menuItem.Tag ) ) )
			{
				IEnumerable children = Control.View.ChildrenProperty.GetValue( camera, null ) as IEnumerable;
				if( children != null )
					foreach( object child in children )
					{
						if( Control.View.IsPreset( child ) )
						{
							ToolStripMenuItem presetMenuItem = CreateCameraMenuItem( child );
							menuItem.DropDownItems.Add( presetMenuItem );
						}
						else
							AddMenuItems( menuItem, child );
					}
			}
			else if( ( Control.View.IsAvailableCameraType( camera ) && Control.View.ChildrenProperty != null ) )
			{
				IEnumerable children = Control.View.ChildrenProperty.GetValue( camera, null ) as IEnumerable;
				if( children != null )
					foreach( object child in children )
					{
						ToolStripMenuItem newCamera = CreateCameraMenuItem( child );
						AddMenuItems( newCamera, child );
						menuItem.DropDownItems.Add( newCamera );
					}
			}
			//else if( ( Control.View.IsPresetType( camera ) && Control.View.PresetChildrenProperty != null ) )
			//{
			//	IEnumerable children = Control.View.PresetChildrenProperty.GetValue( camera, null ) as IEnumerable;
			//	if( children != null )
			//		foreach( object child in children )
			//		{
			//			ToolStripMenuItem newCamera = CreateCameraMenuItem( child );
			//			AddMenuItems( newCamera, child );
			//			menuItem.DropDownItems.Add( newCamera );
			//		}
			//}
		}
		#endregion

		private ContextMenuStrip cameraMenu;
		#region public void OnCameraCombo()
		public void OnCameraCombo()
		{
			cameraMenu = new ContextMenuStrip();
			if( Control.CameraInfo != Control.NoCamera )
			{
				if( Control.Connected )
				{
					ToolStripMenuItem disconnect = CreateCameraMenuItem( "Disconnect" );
					ToolStripMenuItem disconnectAll = CreateCameraMenuItem( "All" );
					disconnectAll.Click += OnDisconnectAll;
					disconnect.DropDownItems.Add( disconnectAll );
					ToolStripMenuItem disconnectThisCamera = CreateCameraMenuItem( "This Camera" );
					disconnectThisCamera.Click += OnDisconnectThisCamera;
					disconnect.DropDownItems.Add( disconnectThisCamera );
					cameraMenu.Items.Add( disconnect );
				}
				else
				{
					ToolStripMenuItem connect = CreateCameraMenuItem( "Connect" );
					connect.Click += OnConnect;
					cameraMenu.Items.Add( connect );
				}
			}
			else
				cameraMenu.Items.Add( CreateCameraMenuItem( Control.NoCamera ) );

			foreach( object camera in Control.View.AvailableCameras )
			{
				ToolStripMenuItem newCamera = CreateCameraMenuItem( camera );
				AddMenuItems( newCamera, camera );
				cameraMenu.Items.Add( newCamera );
			}

			Rectangle cameraButtonRect = cameraButton.ClientRectangle;
			cameraMenu.Show( new Point( Location.X + cameraButtonRect.Location.X, Location.Y + cameraButtonRect.Location.Y ) );
		}
		#endregion

		#region private void OnCameraClick( object sender, EventArgs e )
		private void OnCameraClick( object sender, EventArgs e )
		{
			this.cameraMenu.Close();
			ToolStripMenuItem menuItem = sender as ToolStripMenuItem;
			object camera = menuItem.Tag;
			if( Control.View.IsAvailableCameraType( camera ) && Control.CameraInfo != camera )
			{
				if( Control.View.IsCameraProperty != null && (bool)Control.View.IsCameraProperty.GetValue( camera, null ) )
				{
					Control.CameraInfo = menuItem.Tag;
					Invalidate();
				}
			}
			else if( Control.View.IsPreset( camera ) )
			{
				Control.GoToPreset( Control.View.PresetAddress( camera ) );
			}
		}
		#endregion

		private void OnConnect( object sender, EventArgs e )				{ Control.Connect(); }
		private void OnDisconnectThisCamera( object sender, EventArgs e )	{ Control.Disconnect(true); }
		#region private void OnDisconnectAll( object sender, EventArgs e )
		private void OnDisconnectAll( object sender, EventArgs e )
		{
			Control.View.DisconnectAll(true);
		}
		#endregion
		private void OnLive( object sender, EventArgs e )					{ Control.Live();			UpdatePlayback( false );	}
		private void OnFastReverse( object sender, EventArgs e )			{ Control.FastReverse();	UpdatePlayback( true );		}
		private void OnReverse( object sender, EventArgs e )				{ Control.Reverse();		UpdatePlayback( true );		}
		#region private void OnPlay( object sender, EventArgs e ) 
		private void OnPlay( object sender, EventArgs e ) 
		{
			if( !Control.PlaybackMode )
				Control.PlaybackTime = DateTime.Now;
			Control.Play();
			UpdatePlayback( true );
		}
		#endregion
		private void OnForward( object sender, EventArgs e )				{ Control.Forward();		UpdatePlayback( true );		}
		private void OnFastForward( object sender, EventArgs e )			{ Control.FastForward();	UpdatePlayback( true );		}
		private void OnRecord( object sender, EventArgs e )					{ Control.Record(); }
		private void OnSaveClip( object sender, EventArgs e )				{ Control.SaveClip(); }
		private void OnSnapshot( object sender, EventArgs e )				{ Control.Snapshot(); }

		#region private void OnBack30( object sender, EventArgs e )
		private void OnBack30( object sender, EventArgs e )
		{
			if( !Control.PlaybackMode )
				Control.PlaybackTime = DateTime.Now;
			else
				Control.PlaybackTime -= new TimeSpan( 0, 0, 30 );

			Control.OnPlaybackTimeChanged();
			UpdatePlayback( true );
		}
		#endregion
		#region private void OnCalendar( object sender, EventArgs e )
		private void OnCalendar( object sender, EventArgs e )
		{
			CalendarForm calendar = new CalendarForm();
			calendar.dateTimePicker.Value = Control.PlaybackTime;
			calendar.Location = new Point( Control.ClientRectangle.Left + Control.Size.Width - calendar.Width - 2, Control.ClientRectangle.Top + Control.Size.Height - calendar.Height - 2 ); 
			if( calendar.ShowDialog( this ) == DialogResult.OK )
			{
				Control.PlaybackTime = calendar.dateTimePicker.Value;
				Control.OnPlaybackTimeChanged();
				UpdatePlayback( true );
			}
		}
		#endregion
		#region private void OnAlarms( object sender, EventArgs e )
		private void OnAlarms( object sender, EventArgs e )
		{
			AlarmsForm alarmsForm = new AlarmsForm( Control, Control.VideoAlarms );
			alarmsForm.Show();
			UpdatePlayback( Control.PlaybackMode );
		}
		#endregion
		#region private void OnPTZ( object sender, EventArgs e )
		private void OnPTZ( object sender, EventArgs e )
		{
			if( Control.PTZForm == null )
			{
				Form parentForm = Control.FindForm();
				Screen screen = Screen.FromControl( parentForm );
				Rectangle area = screen.WorkingArea;

				Control.PTZForm = new PTZForm( Control );
				int left = Control.ClientRectangle.Right;
				int top = Control.ClientRectangle.Top;

				if( left + Control.PTZForm.Width > screen.WorkingArea.Width )
					left = Control.ClientRectangle.Left - Control.PTZForm.Width;
				if( top + Control.PTZForm.Height > screen.WorkingArea.Width )
					top = Control.ClientRectangle.Top - Control.PTZForm.Height;

				if( left < 0 || top < 0 )
				{
					left = Control.ClientRectangle.Left;
					top  = Control.ClientRectangle.Top;
				}

				Control.PTZForm.Location = new Point( left, top ); 
				Control.PTZForm.Show();
			}
			else
				Control.PTZForm.Activate();
		}
		#endregion
		#region private void OnTimelineMouseDown( object sender, MouseEventArgs e )
		private void OnTimelineMouseDown( object sender, MouseEventArgs e )
		{
			if( !Control.View.DisableConfigChange )
			{
				Control.PlaybackTime = Control.RulerStartTime().AddSeconds( (int)( (float)e.X * SecondsPerPixel() ) );
				Control.OnPlaybackTimeChanged();
				UpdatePlayback( true );
			}
		}
		#endregion

		#region private void UpdatePlayback( bool newPlaybackMode )
		private void UpdatePlayback( bool newPlaybackMode )
		{
			Control.PlaybackMode = newPlaybackMode;
			liveButton.Visible = Control.PlaybackMode;
			timelinePanel.Invalidate();
		}
		#endregion

		#region private void OnPaintCameraButton( object sender, PaintEventArgs e )
		private void OnPaintCameraButton( object sender, PaintEventArgs e )
		{
			base.OnPaint(e);

			Graphics g = e.Graphics;
			Rectangle bounds = cameraButton.ClientRectangle;
			int BorderSize = SystemInformation.Border3DSize.Width * 2;
			const int PushButtonWidth = 14;

			ButtonRenderer.DrawButton( g, bounds, System.Windows.Forms.VisualStyles.PushButtonState.Normal );

		//	Calculate the current dropdown rectangle.
			Rectangle dropDownRectangle = new Rectangle( bounds.Right - PushButtonWidth - 1, BorderSize, PushButtonWidth, bounds.Height - BorderSize * 2 );

			int internalBorder = BorderSize;
			Rectangle focusRect = new Rectangle( internalBorder, internalBorder, bounds.Width - dropDownRectangle.Width - internalBorder, bounds.Height - ( internalBorder * 2 ) );  

		//	Draw an arrow in the correct location
			Point middle = new Point( Convert.ToInt32( dropDownRectangle.Left + dropDownRectangle.Width / 2 ), Convert.ToInt32( dropDownRectangle.Top + dropDownRectangle.Height / 2 ) );
		//	If the width is odd – favor pushing it over one pixel right.
			middle.X += ( dropDownRectangle.Width % 2 );

			Point[] arrow = new Point[] { new Point( middle.X - 2, middle.Y - 1 ), new Point( middle.X + 3, middle.Y - 1 ), new Point( middle.X, middle.Y + 2 ) };
			g.FillPolygon( SystemBrushes.ControlText, arrow);

		//	Figure out how to draw the text
			TextFormatFlags formatFlags = TextFormatFlags.HorizontalCenter | TextFormatFlags.VerticalCenter | TextFormatFlags.HidePrefix;
			if( !string.IsNullOrEmpty( cameraButton.Text ) )
				TextRenderer.DrawText(g, cameraButton.Text, Font, focusRect, SystemColors.ControlText, formatFlags);
		}
		#endregion
		#region private void OnCameraButtonClicked( object sender, EventArgs e )
		private void OnCameraButtonClicked( object sender, EventArgs e )
		{
			OnCameraCombo();
		}
		#endregion
	}
}
