﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;
using IntelliSite.CommonUtil;
using IntelliSite.VideoCommon;

namespace IntelliSite.WinForms.VideoView
{
	public class VideoControl : Control
	{
		private object _cameraInfo;

		public VideoView View { get; set; }
		public object CameraInfo { get { return _cameraInfo; } set { ChangeCameras(value); } }
		public object NoCamera { get; set; }
		public bool Hover { get; set; }
		public bool LastActive { get; set; }
		public Point MouseLocation { get; set; }
		public DateTime PlaybackStartTime { get; set; }
		public DateTime PlaybackTime { get; set; }
		public bool PlaybackMode { get; set; }
		public string VideoStatusText { get; set; }
		public List<VideoAlarm> VideoAlarms { get; set; }

		internal ToolsForm Tools { get; set; }
		internal PTZForm PTZForm { get; set; }
		private ISCameraBase CameraDriver { get; set; }

		public event EventHandler<GetAlarmsArgs> ProcessGetAlarms;
		public event EventHandler<ErrorArgs> Exception;

		public VideoControl(VideoView view) : base()
		{
			View = view;
			NoCamera = "None";
			CameraInfo = NoCamera;
			CameraDriver = null;
			VideoStatusText = String.Empty;

			PlaybackStartTime = DateTime.Now.AddHours(-4);
			PlaybackTime = DateTime.Now.AddHours(-2);

			Paint += OnPaint;
			MouseDown += OnMouseDown;
			MouseMove += OnMouseMove;
			MouseHover += OnMouseHover;
			MouseLeave += OnMouseLeave;
			KeyDown += OnKeyDown;
            KeyUp += OnKeyUp;
			SizeChanged += OnSizeChanged;
		}

		public void Init()
		{
        }

        #region public string CameraProperty( string name )
        public string CameraProperty(string name)
		{
			if (CameraInfo != null && CameraInfo != NoCamera && View.PropertiesProperty != null)
			{
				IEnumerable properties = View.PropertiesProperty.GetValue(CameraInfo, null) as IEnumerable;

				PropertyInfo nameProperty = null;
				PropertyInfo valueProperty = null;
				foreach (object property in properties)
				{
					if (nameProperty == null)
					{
						nameProperty = property.GetType().GetProperty("Name");
						valueProperty = property.GetType().GetProperty("Value");
					}

					string propertyName = String.Empty;
					if (nameProperty != null)
						propertyName = (string)nameProperty.GetValue(property, null);

					if (propertyName == name)
					{
						string propertyValue = String.Empty;
						if (valueProperty != null)
							propertyValue = (string)valueProperty.GetValue(property, null);
						return propertyValue;
					}
				}

			}
			return String.Empty;
		}
		#endregion

		#region public bool Connected { get; }
		public bool Connected
		{
			get
			{
				if (CameraInfo != null && CameraInfo != NoCamera && View.IsAvailableCameraType(CameraInfo) && CameraDriver != null)
					return CameraDriver.IsConnected;
				return CameraInfo != NoCamera && CameraInfo != null;
			}
		}
		#endregion
		#region public bool Recording { get; }
		public bool Recording
		{
			get
			{
				if (View.IsAvailableCameraType(CameraInfo) && CameraDriver != null)
					return CameraDriver.IsRecording;
				return false;
			}
		}
		#endregion
		#region public bool PlaybackEnabled { get; }
		public bool PlaybackEnabled
		{
			get
			{
				return Connected && View.IsAvailableCameraType(CameraInfo) && View.PlaybackEnabledProperty != null && (bool)View.PlaybackEnabledProperty.GetValue(CameraInfo, null);
			}
		}
		#endregion
		#region public bool PTZEnabled { get; }
		public bool PTZEnabled
		{
			get
			{
				return Connected && !View.DisablePTZ && View.IsAvailableCameraType(CameraInfo) && View.PTZEnabledProperty != null && (bool)View.PTZEnabledProperty.GetValue(CameraInfo, null);
			}
		}
		#endregion
		#region public string CameraName { get; }
		public string CameraName
		{
			get
			{
				if (CameraInfo != NoCamera && View.IsAvailableCameraType(CameraInfo) && View.NameProperty != null)
					return (string)View.NameProperty.GetValue(CameraInfo, null);
				return String.Empty;
			}
		}
		#endregion
		public string CameraDriverName { get { return CameraProperty("Driver"); } }

		#region public DateTime RulerStartTime()
		public DateTime RulerStartTime()
		{
			return PlaybackMode ? PlaybackStartTime : DateTime.Now.AddHours(-2);
		}
		#endregion

		#region private void ChangeCameras( object newCameraInfo )
		private void ChangeCameras(object newCameraInfo)
		{
			try
			{

                Cursor.Current = Cursors.WaitCursor;

                if( PlaybackEnabled )
					Live();

				if( Connected )
					Disconnect();

				_cameraInfo = newCameraInfo;

				if (CameraInfo != null && CameraInfo != NoCamera)
				{

					string driver = CameraDriverName;
					if (!String.IsNullOrEmpty(driver))
					{
						if (CameraDriver != null && !CameraDriver.GetType().ToString().Contains(driver))
							CameraDriver.Dispose();

                        if (CameraDriver == null)
                        {
                            CameraDriver = ISCameraBase.CreateDriver(driver, View.Logger);
							if( CameraDriver != null )
							{
								CameraDriver.VideoStatus += VideoStatus;
								CameraDriver.Exception += OnException;
							}
                        }
					}

					if (CameraDriver != null)
					{
						CameraDriver.ChangeCamera(CameraInfo);
						Connect();
					}
				}
			}
			catch (Exception e)
			{
                this.CameraDriver.Message = e.Message + "\n\nPlease check your connection parameters.";
				Console.WriteLine("CameraChanged exception, " + e.ToString());
			}
            finally
            {
                Cursor.Current = Cursors.Default;
            }
		}
		#endregion

		#region public bool Connect()
		public bool Connect()
		{
            if (View.IsAvailableCameraType(CameraInfo) && CameraDriver != null)
			{
				bool connected = CameraDriver.Connect( this );
                if (connected)
                {
                    CameraDriver.VideoViewerWinForms.Location = new Point(0, 0);
                    CameraDriver.VideoViewerWinForms.Size = Size;
                    CameraDriver.VideoViewerWinForms.MouseDown += OnMouseDown;
                    CameraDriver.VideoViewerWinForms.MouseHover += OnMouseHover;
                    CameraDriver.VideoViewerWinForms.MouseLeave += OnMouseLeave;
                    CameraDriver.VideoStatus += VideoStatus;
                    CameraDriver.VideoSetupInfo += VideoSetupInfo;
                    CameraDriver.Streaming += Streaming;
					CameraDriver.ConnectionLost += ConnectionLost;
                    BeginInvoke(new AddControlDelegate(AddControl), new object[] { CameraDriver.VideoViewerWinForms });
                }

                if (Tools != null)
                    Tools.ConnectedChanged();
            }
            return true;
		}
		#endregion

		private void ConnectionLost( object sender, EventArgs e ) { BeginInvoke( new ReconnectDelegate( Reconnect ), null ); }

		private delegate void ReconnectDelegate();
		private void Reconnect() { CameraDriver.Reconnect(); }

		#region private void VideoSetupInfo( object sender, VideoStatusArgs e )
		private void VideoSetupInfo( object sender, VideoStatusArgs e )
		{
			if( View.IsAvailableCameraType(CameraInfo) && CameraDriver != null )
			{
				PropertyInfo setupProperty = CameraInfo.GetType().GetProperty( "Setup" );
				if( setupProperty != null )
				{
					string setupInfo = (string)setupProperty.GetValue( CameraInfo );
					if( String.IsNullOrEmpty( setupInfo ) )
					{
						setupProperty.SetValue( CameraInfo, e.Text );
						View.OnSetupChanged( CameraInfo );
					}
				}
			}
		}
		#endregion
		#region private void VideoStatus( object sender, VideoStatusArgs e )
		private void VideoStatus(object sender, VideoStatusArgs e)
		{
			VideoStatusText = e.Text;
			Invalidate();
		}
		#endregion
		#region private void OnException( object sender, ErrorArgs e )
		private void OnException( object sender, ErrorArgs e )
		{
			Exception( sender, e );
		}
		#endregion

		private delegate void AddControlDelegate(Control control);
		private void AddControl(Control control) { Controls.Add(control); }

		private delegate void RemoveControlDelegate(Control control);
		private void RemoveControl(Control control) { Controls.Remove(control); }

		#region public void Streaming(object sender, EventArgs e)
		public void Streaming(object sender, EventArgs e)
		{
			if( !Controls.Contains( CameraDriver.VideoViewerWinForms ) )
				BeginInvoke(new AddControlDelegate(AddControl), new object[] { CameraDriver.VideoViewerWinForms });
		}
        #endregion

        public void Disconnect()
        {
            this.Disconnect(false);
        }

        public void Disconnect(bool dereferenceCamera)
        {

            Cursor.Current = Cursors.WaitCursor;

			if (PTZForm != null)
			{
				PTZForm.Close();
				PTZForm = null;
			}

			if (Tools != null)
			{
				Tools.Close();
				Tools = null;
                Hover = false;
			}

			if (View.IsAvailableCameraType(CameraInfo) && CameraDriver != null)
			{
                if (CameraDriver.VideoViewerWinForms != null)
                {
                    CameraDriver.VideoViewerWinForms.MouseDown -= OnMouseDown;
                    CameraDriver.VideoViewerWinForms.MouseHover -= OnMouseHover;
                    CameraDriver.VideoViewerWinForms.MouseLeave -= OnMouseLeave;
                    Controls.Remove(CameraDriver.VideoViewerWinForms);
                }

                CameraDriver.Disconnect();
				if (Tools != null)
					Tools.ConnectedChanged();

                if (CameraDriver.VideoViewerWinForms != null)
                {
                    BeginInvoke(new RemoveControlDelegate(RemoveControl), new object[] { CameraDriver.VideoViewerWinForms });
                }

                if (dereferenceCamera)
                {
                    CameraInfo                   = NoCamera;
					CameraDriver.VideoStatus    -= VideoStatus;
                    CameraDriver.VideoSetupInfo -= VideoSetupInfo;
					CameraDriver.Exception      -= OnException;
                    CameraDriver.Streaming      -= Streaming;
					CameraDriver.ConnectionLost -= ConnectionLost;
                    this.CameraDriver.Dispose();
                    this.CameraDriver = null;
                }

            }

            Cursor.Current = Cursors.Default;

        }

		public void Live()
		{
			CameraDriver.Live();
			PlaybackMode = false;
			Invalidate();
		}

		public void PlaybackTimeChanged()
		{
			TimeChangeEventArgs args = new TimeChangeEventArgs();
			args.OldTime = CameraDriver.PlaybackCurrentTime;
			args.NewTime = PlaybackTime;
			CameraDriver.PlaybackCurrentTime = PlaybackTime;

			CameraDriver.OnPlaybackCurrentTimeChanged( this, args );
		}

		public void Play()        { CameraDriver.Play(); }
		public void Forward()     { CameraDriver.Forward(); }
		public void Reverse()     { CameraDriver.Reverse(); }
		public void FastForward() { CameraDriver.FastFoward(); }
		public void FastReverse() { CameraDriver.FastReverse();  }

		public void Record()
		{
			CameraDriver.Record();
			if( Tools != null )
				Tools.RecordingChanged();
			Invalidate();
		}

		public void Snapshot() { CameraDriver.Snapshot(); }
		public void SaveClip() { CameraDriver.SaveClip(); }

		public void PTZConnect()
		{
			if (!CameraDriver.IsPTZConnected)
			{
				CameraDriver.PTZAvailable += PTZAvailable;
				CameraDriver.PTZConnect();
			}
		}

		public void PTZDisconnect()
		{
			if (CameraDriver.IsPTZConnected)
			{
				CameraDriver.PTZAvailable -= PTZAvailable;
				CameraDriver.PTZDisconnect();
			}

			PTZForm = null;
		}

		public bool IsPTZConnected { get { return CameraDriver != null && CameraDriver.IsPTZConnected; } }

		public void PTZ(PTZCommands command, bool start)
		{
			if (CameraDriver.IsPTZConnected)
				CameraDriver.PTZ(command, start);
		}

		public void GoToPreset( string presetInfo )
		{
			if( CameraDriver != null )
				CameraDriver.GoToPreset( presetInfo );
		}

		private void PTZAvailable(object sender, EventArgs e)
		{
			if (PTZForm != null)
				BeginInvoke(new PTZForm.EnableControlsDelegate(PTZForm.EnableControls), new object[] { true });
		}

		public List<VideoAlarm> GetAlarms(DateTime startTime, DateTime endTime, bool systemAlarms, bool bookmarks, bool events, bool alarms)
		{
			if( ProcessGetAlarms != null )
				ProcessGetAlarms( this, new GetAlarmsArgs( startTime, endTime, systemAlarms, bookmarks, events, alarms ) );
			else
				VideoAlarms = CameraDriver.GetAlarms( startTime, endTime, systemAlarms, bookmarks, events, alarms );
			return VideoAlarms;
		}

		#region private void OnSizeChanged( object sender, EventArgs e )
		private void OnSizeChanged(object sender, EventArgs e)
		{
			if (Tools != null)
				Tools.SetWidth(ClientRectangle.Width - 4);

			if (CameraDriver != null && CameraDriver.VideoViewerWinForms != null)
			{
				CameraDriver.VideoViewerWinForms.Location = new Point(0, 0);
				CameraDriver.VideoViewerWinForms.Size = Size;
			}
		}
		#endregion
		#region private void OnPaint( object sender, PaintEventArgs e )
		private void OnPaint(object sender, PaintEventArgs e)
		{
			e.Graphics.FillRectangle(new SolidBrush(Color.Black), ClientRectangle);
			if (LastActive)
				ControlPaint.DrawBorder(e.Graphics, ClientRectangle, Color.LightGray, 3, ButtonBorderStyle.Dashed, Color.LightGray, 3, ButtonBorderStyle.Dashed, Color.LightGray, 3, ButtonBorderStyle.Dashed, Color.LightGray, 3, ButtonBorderStyle.Dashed);
			else
				ControlPaint.DrawBorder(e.Graphics, ClientRectangle, Color.LightGray, ButtonBorderStyle.Solid);

			if (!String.IsNullOrEmpty(VideoStatusText))
				using (SolidBrush white = new SolidBrush(Color.White))
					e.Graphics.DrawString(VideoStatusText, SystemFonts.DialogFont, white, new Point(5, 5));
		}
		#endregion
		#region private void OnMouseDown( object sender, MouseEventArgs e )
		private void OnMouseDown(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
				if (!LastActive)
				{
					LastActive = true;
					Invalidate();
					foreach (VideoControl control in View.Controls)
						if (control != this && control.LastActive)
						{
							control.LastActive = false;
							control.Invalidate();
						}
				}
		}
		#endregion
		#region private void OnMouseHover( object sender, EventArgs e )
		public void OnMouseHover(object sender, EventArgs e)
		{
			foreach (VideoControl control in View.Controls)
				if (control != this)
					control.OnMouseLeave(sender, e);

			if (!Hover && !View.DisableToolbar)
			{
				Point location = PointToScreen(Point.Empty);

				Tools = new ToolsForm(this);
				Tools.Location = new Point(location.X + 2, location.Y + Height - Tools.Height - 2);
				Tools.SetWidth(ClientRectangle.Width - 4);
				Tools.Show();

				Hover = true;
			}
		}
		#endregion
		#region public void OnMouseLeave( object sender, EventArgs e )
		public void OnMouseLeave(object sender, EventArgs e)
		{
			Rectangle controlRect = new Rectangle(PointToScreen(Point.Empty), Size);
			if (Hover && !View.DisableToolbar && !controlRect.Contains(MousePosition))
			{
                if (Tools != null)
                {
                    Tools.Close();
                    Tools = null;
                    Hover = false;
                }
            }
		}
		#endregion
		#region private void OnMouseMove( object sender, MouseEventArgs e )
		private void OnMouseMove(object sender, MouseEventArgs e)
		{
			OnMouseHover(sender, new EventArgs());
		}
		#endregion
		#region protected override bool IsInputKey( Keys keyData )
		protected override bool IsInputKey( Keys keyData )
		{
			switch( keyData )
			{
				case Keys.Escape:
				case Keys.Right:
				case Keys.Left:
				case Keys.Up:
				case Keys.Down:
				case Keys.Shift | Keys.Right:
				case Keys.Shift | Keys.Left:
				case Keys.Shift | Keys.Up:
				case Keys.Shift | Keys.Down:
				return true;
			}

			return base.IsInputKey( keyData );
		}
		#endregion
		#region private void OnKeyDown( object sender, KeyEventArgs e )
		private void OnKeyDown( object sender, KeyEventArgs e )
		{	
			if( e.KeyCode == Keys.Escape && FindForm().FormBorderStyle == FormBorderStyle.None )
				View.OnFullScreen( sender, new EventArgs() );
		}
        #endregion
        #region private void OnKeyUp( object sender, KeyEventArgs e )
        public void OnKeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Escape && FindForm().FormBorderStyle == FormBorderStyle.None)
            {
                View.OnFullScreen(sender, new EventArgs());
            }
        }
        #endregion
        #region internal void OnPlaybackTimeChanged()
        internal void OnPlaybackTimeChanged()
		{
			if (!PlaybackMode && PlaybackTime > RulerStartTime())
				PlaybackStartTime = DateTime.Now;

			if (PlaybackTime <= PlaybackStartTime || PlaybackTime >= PlaybackStartTime.AddHours(4) )
				PlaybackStartTime = PlaybackTime.AddHours(-2);

			PlaybackMode = true;

			if (Tools != null)
				Tools.PlaybackTimeChanged();

			PlaybackTimeChanged();

			Invalidate();
		}
		#endregion
	}
}
