﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.Drawing;
using System.Reflection;
using System.Threading.Tasks;
using System.Windows.Forms;
using IntelliSite.CommonUtil;

namespace IntelliSite.WinForms.VideoView
{
	public class VideoView : Control, IDisposable
	{
		#region public enum VideoLayouts
		public enum VideoLayouts
		{
			One,
			OneByTwo,
			TwoByTwo,
			OnePlusSeven,
			ThreeByThree,
			TwoPlusEight,
			OnePlusTwelve,
			FourByFour,
			FiveByFive,
			SixBySix
		}
		#endregion

		public VideoLayouts VideoLayout { get; private set; }
		public List<object> AvailableCameras { get; set; }

		private Type AvailableCameraType { get; set; }
		internal PropertyInfo NameProperty { get; set; }
		internal PropertyInfo ChildrenProperty { get; set; }
		internal PropertyInfo IsCameraProperty { get; set; }
		internal PropertyInfo PlaybackEnabledProperty { get; set; }
		internal PropertyInfo PTZEnabledProperty { get; set; }
		internal PropertyInfo PropertiesProperty { get; set; }

		private Type PresetType { get; set; }
		internal PropertyInfo PresetChildrenProperty { get; set; }
		internal PropertyInfo IsPresetProperty { get; set; }
		internal PropertyInfo PresetAddressProperty { get; set; }

		public bool DisableFullScreen { get; set; }
		public bool DisableToolbar { get; set; }
		public bool DisablePTZ { get; set; }
		public bool DisableCameraChange { get; set; }
		public bool DisableConfigChange { get; set; }
		public bool DisablePlayback { get; set; }

		public bool ToolTipsDisabled { get; set; }

		public event EventHandler<GetAlarmsArgs> ProcessGetAlarms = delegate {};
		public event EventHandler<SetupChangedArgs> SetupChanged = delegate {};
		public event EventHandler<ErrorArgs> Exception = delegate {};

		public LoggerBase Logger { get; set; }

		private Form FullScreenForm { get; set; }
		private Control ParentControl { get; set; }

		public VideoView(LoggerBase logger) : base()
		{
			InitializeComponent();

			AvailableCameras = new List<object>();
			Logger = logger;

			KeyDown += OnKeyDown;
			ParentChanged += OnParentChanged;
			SizeChanged += OnSizeChanged;

			FullScreenForm = null;
			ParentControl = null;
			
			SetLayout(VideoLayouts.TwoByTwo);
		}

		private void OnSizeChanged( object sender, EventArgs e ) { ResizeControls(); }

		#region public void SetAvailableCameraType( Type type )
		public void SetAvailableCameraType(Type type)
		{
			AvailableCameraType = type;
			NameProperty = AvailableCameraType.GetProperty("Name");
			ChildrenProperty = AvailableCameraType.GetProperty("Children");
			IsCameraProperty = AvailableCameraType.GetProperty("IsCamera");
			IsPresetProperty = AvailableCameraType.GetProperty("IsPreset");
			PresetAddressProperty = AvailableCameraType.GetProperty("Address");
			PlaybackEnabledProperty = AvailableCameraType.GetProperty("PlaybackEnabled");
			PTZEnabledProperty = AvailableCameraType.GetProperty("PTZEnabled");
			PropertiesProperty = AvailableCameraType.GetProperty("Properties");
		}
		#endregion
		#region public void SetPresetType( Type type )
		public void SetPresetType(Type type)
		{
			PresetType = type;
			PresetChildrenProperty = AvailableCameraType.GetProperty("Children");
			IsPresetProperty = AvailableCameraType.GetProperty("IsPreset");
			PresetAddressProperty = AvailableCameraType.GetProperty("Address");
		}
		#endregion
		public bool IsAvailableCameraType(object camera) { return camera != null && (camera.GetType() == AvailableCameraType || camera.GetType().IsSubclassOf(AvailableCameraType)); }
		public bool IsPresetType(object preset) { return preset != null && (preset.GetType() == PresetType || preset.GetType().IsSubclassOf(PresetType)); }
		public bool IsPreset(object preset) { return IsPresetType( preset ) && IsPresetProperty != null && (bool)IsPresetProperty.GetValue( preset ); }
		public string PresetAddress(object preset) { if( IsPresetType( preset ) && PresetAddressProperty != null ) return (string)PresetAddressProperty.GetValue( preset ); return String.Empty; }

		#region public int ControlCount( VideoLayouts layout )
		public int ControlCount(VideoLayouts layout)
		{
			switch (layout)
			{
				case VideoLayouts.One: return 1;
				case VideoLayouts.OneByTwo: return 2;
				case VideoLayouts.TwoByTwo: return 4;
				case VideoLayouts.OnePlusSeven: return 8;
				case VideoLayouts.ThreeByThree: return 9;
				case VideoLayouts.TwoPlusEight: return 10;
				case VideoLayouts.OnePlusTwelve: return 13;
				case VideoLayouts.FourByFour: return 16;
				case VideoLayouts.FiveByFive: return 25;
				case VideoLayouts.SixBySix: return 36;
			}

			return 0;
		}
		#endregion

		#region public void SetLayout( VideoLayouts layout )
		public void SetLayout(VideoLayouts layout)
		{
			VideoLayout = layout;
			VideoLayoutChanged();
		}
		#endregion
		#region public void VideoLayoutChanged()
		public void VideoLayoutChanged()
		{

			try
			{
				int count = ControlCount(VideoLayout);
				while (Controls.Count > count)
				{
					VideoControl videoControl = Controls[Controls.Count - 1] as VideoControl;
					CloseVideoControl(videoControl);
					Controls.Remove(videoControl);
				}

				while (Controls.Count < count)
					Controls.Add(CreateVideoControl());

				ResizeControls();

				foreach( VideoControl videoControl in Controls )
					if( videoControl.Tools != null )
						videoControl.Tools.Close();
			}
			catch(Exception ex)
			{
				Logger.Write(ex.ToString());
			}
		}
		#endregion

		#region public void ResizeControls()
		public void ResizeControls()
		{
			switch (VideoLayout)
			{
				#region case VideoLayouts.One:
				case VideoLayouts.One:
					Controls[0].Location = new Point(0, 0);
					Controls[0].Size = Size;
					break;
				#endregion
				#region case VideoLayouts.OneByTwo:
				case VideoLayouts.OneByTwo:
					{
						bool wide = Size.Width > Size.Height;
						Size size = wide ? new Size(Size.Width / 2, Size.Height) : new Size(Size.Width, Size.Height / 2);
						foreach (Control control in Controls)
							control.Size = size;

						for (int i = 0; i < Controls.Count; i++)
							if (wide)
								Controls[i].Location = new Point((i % 2) * size.Width, (i / 2) * size.Height);
							else
								Controls[i].Location = new Point((i / 2) * size.Width, (i % 2) * size.Height);
						break;
					}
				#endregion
				#region case VideoLayouts.TwoByTwo:
				case VideoLayouts.TwoByTwo:
					{
						Size size = new Size(Size.Width / 2, Size.Height / 2);
						foreach (Control control in Controls)
							control.Size = size;

						for (int i = 0; i < Controls.Count; i++)
							Controls[i].Location = new Point((i % 2) * size.Width, (i / 2) * size.Height);
						break;
					}
				#endregion
				#region case VideoLayouts.OnePlusSeven:
				case VideoLayouts.OnePlusSeven:
					{
						Size small = new Size(Size.Width / 4, Size.Height / 4);
						Controls[0].Size = new Size(small.Width * 3, small.Height * 3);
						for (int i = 1; i < 8; i++)
							Controls[i].Size = small;

						Controls[0].Location = new Point(0, 0);
						Controls[1].Location = new Point(0, small.Height * 3);
						Controls[2].Location = new Point(small.Width, small.Height * 3);
						Controls[3].Location = new Point(small.Width * 2, small.Height * 3);
						Controls[4].Location = new Point(small.Width * 3, 0);
						Controls[5].Location = new Point(small.Width * 3, small.Height);
						Controls[6].Location = new Point(small.Width * 3, small.Height * 2);
						Controls[7].Location = new Point(small.Width * 3, small.Height * 3);
						break;
					}
				#endregion
				#region case VideoLayouts.ThreeByThree:
				case VideoLayouts.ThreeByThree:
					{
						Size size = new Size(Size.Width / 3, Size.Height / 3);
						foreach (Control control in Controls)
							control.Size = size;

						for (int i = 0; i < Controls.Count; i++)
							Controls[i].Location = new Point((i % 3) * size.Width, (i / 3) * size.Height);
						break;
					}
				#endregion
				#region case VideoLayouts.TwoPlusEight:
				case VideoLayouts.TwoPlusEight:
					{
						Size small = new Size(Size.Width / 4, Size.Height / 4);
						Controls[0].Size = new Size(small.Width * 2, small.Height * 2);
						Controls[1].Size = new Size(small.Width * 2, small.Height * 2);
						for (int i = 2; i < Controls.Count; i++)
							Controls[i].Size = small;

						Controls[0].Location = new Point(small.Width, 0);
						Controls[1].Location = new Point(small.Width, small.Height * 2);
						Controls[2].Location = new Point(0, 0);
						Controls[3].Location = new Point(0, small.Height);
						Controls[4].Location = new Point(0, small.Height * 2);
						Controls[5].Location = new Point(0, small.Height * 3);
						Controls[6].Location = new Point(small.Width * 3, 0);
						Controls[7].Location = new Point(small.Width * 3, small.Height);
						Controls[8].Location = new Point(small.Width * 3, small.Height * 2);
						Controls[9].Location = new Point(small.Width * 3, small.Height * 3);
						break;
					}
				#endregion
				#region case VideoLayouts.OnePlusTwelve:
				case VideoLayouts.OnePlusTwelve:
					{
						Size small = new Size(Size.Width / 4, Size.Height / 4);
						Controls[0].Size = new Size(small.Width * 2, small.Height * 2);
						for (int i = 1; i < Controls.Count; i++)
							Controls[i].Size = small;

						Controls[0].Location = new Point(small.Width, small.Height);
						for (int i = 1; i < 5; i++)
							Controls[i].Location = new Point(small.Width * (i - 1), 0);
						for (int i = 5; i < 9; i++)
							Controls[i].Location = new Point(((i - 5) % 2) * small.Width * 3, ((i - 3) / 2) * small.Height);
						for (int i = 8; i < 13; i++)
							Controls[i].Location = new Point(small.Width * (i - 8), small.Height * 3);
						break;
					}
				#endregion
				#region case VideoLayouts.FourByFour:
				case VideoLayouts.FourByFour:
					{
						Size size = new Size(Size.Width / 4, Size.Height / 4);
						foreach (Control control in Controls)
							control.Size = size;

						for (int i = 0; i < Controls.Count; i++)
							Controls[i].Location = new Point((i % 4) * size.Width, (i / 4) * size.Height);
						break;
					}
				#endregion
				#region case VideoLayouts.FiveByFive:
				case VideoLayouts.FiveByFive:
					{
						Size size = new Size(Size.Width / 5, Size.Height / 5);
						foreach (Control control in Controls)
							control.Size = size;

						for (int i = 0; i < Controls.Count; i++)
							Controls[i].Location = new Point((i % 5) * size.Width, (i / 5) * size.Height);
						break;
					}
				#endregion
				#region case VideoLayouts.SixBySix:
				case VideoLayouts.SixBySix:
					{
						Size size = new Size(Size.Width / 6, Size.Height / 6);
						foreach (Control control in Controls)
							control.Size = size;

						for (int i = 0; i < Controls.Count; i++)
							Controls[i].Location = new Point((i % 6) * size.Width, (i / 6) * size.Height);
						break;
					}
					#endregion
			}

			foreach (Control control in Controls)
				control.Invalidate();
		}
		#endregion

		#region public VideoControl CreateVideoControl()
		public VideoControl CreateVideoControl()
		{
			VideoControl videoControl = new VideoControl(this);
			videoControl.Init();
			videoControl.ProcessGetAlarms += OnProcessGetAlarms;
			videoControl.Exception += OnException;
			return videoControl;
		}
		#endregion

		#region private void OnException( object sender, ErrorArgs e )
		private void OnException( object sender, ErrorArgs e )
		{
			Exception( sender, e );
		}
		#endregion
		#region private void OnProcessGetAlarms( object sender, GetAlarmsArgs e )
		private void OnProcessGetAlarms( object sender, GetAlarmsArgs e )
		{
			if( ProcessGetAlarms != null )
				ProcessGetAlarms( sender, e );
		}
		#endregion
		#region public void CloseVideoControl( VideoControl videoControl )
		public void CloseVideoControl(VideoControl videoControl)
		{
			videoControl.Disconnect();
		}
		#endregion
		#region public void ClearVideoControls()
		public void ClearVideoControls()
		{
			foreach( VideoControl control in Controls )
				control.CameraInfo = control.NoCamera;
		}
		#endregion

		internal void OnSetupChanged( object cameraInfo )	{ SetupChanged( this, new SetupChangedArgs( cameraInfo ) ); }

		#region protected override void Dispose( bool disposing )
		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		/// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
		protected override void Dispose(bool disposing)
		{
			if (disposing && (components != null))
			{
				components.Dispose();
			}
			base.Dispose(disposing);
		}
		#endregion

		#region public void DisconnectAll()
		public void DisconnectAll()
		{
            DisconnectAll(false);
		}

        public void DisconnectAll(bool dereferenceCameras)
        {
            foreach(VideoControl control in Controls)
            {
                control.Disconnect(dereferenceCameras);
            }
        }

		#endregion

		internal void ShowSettingsContextMenu(Point point)
		{
			bool isFullScreen = FindForm().FormBorderStyle == FormBorderStyle.None;
			fullScreenMenuItem.Text = isFullScreen ? "Exit Full Screen" : "Full Screen";
			fullScreenMenuItem.Enabled = !DisableFullScreen;
			settingsMenuStrip.Show(point);
		}

		private void OnOneWindow(object sender, EventArgs e) { SetLayout(VideoLayouts.One); }
		private void OnTwoByTwo(object sender, EventArgs e) { SetLayout(VideoLayouts.TwoByTwo); }
		private void OnOnePlusSeven(object sender, EventArgs e) { SetLayout(VideoLayouts.OnePlusSeven); }
		private void OnThreeByThree(object sender, EventArgs e) { SetLayout(VideoLayouts.ThreeByThree); }
		private void OnTwoPlusEight(object sender, EventArgs e) { SetLayout(VideoLayouts.TwoPlusEight); }
		private void OnOnePlusTwelve(object sender, EventArgs e) { SetLayout(VideoLayouts.OnePlusTwelve); }
		private void OnFourByFour(object sender, EventArgs e) { SetLayout(VideoLayouts.FourByFour); }
		private void OnFiveByFive(object sender, EventArgs e) { SetLayout(VideoLayouts.FiveByFive); }
		private void OnSixBySix(object sender, EventArgs e) { SetLayout(VideoLayouts.SixBySix); }

		#region protected override bool IsInputKey( Keys keyData )
		protected override bool IsInputKey( Keys keyData )
		{
			switch( keyData )
			{
				case Keys.Escape:
				case Keys.Right:
				case Keys.Left:
				case Keys.Up:
				case Keys.Down:
				case Keys.Shift | Keys.Right:
				case Keys.Shift | Keys.Left:
				case Keys.Shift | Keys.Up:
				case Keys.Shift | Keys.Down:
				return true;
			}

			return base.IsInputKey( keyData );
		}
		#endregion
		#region private void OnKeyDown( object sender, KeyEventArgs e )
		private void OnKeyDown( object sender, KeyEventArgs e )
		{	
			if( e.KeyCode == Keys.Escape && FindForm().FormBorderStyle == FormBorderStyle.None )
				OnFullScreen( sender, new EventArgs() );
		}
		#endregion
		#region private void OnParentChanged( object sender, EventArgs e )
		private void OnParentChanged( object sender, EventArgs e )
		{
			if( FullScreenForm == null )
				ParentControl = Parent;
		}
		#endregion

		#region internal void OnFullScreen(object sender, EventArgs e)
		internal void OnFullScreen(object sender, EventArgs e)
		{
			if( FullScreenForm == null )
			{
				FullScreenForm                 = new Form();
				FullScreenForm.WindowState     = FormWindowState.Normal;
				FullScreenForm.FormBorderStyle = FormBorderStyle.None;
				FullScreenForm.Bounds          = Screen.PrimaryScreen.Bounds;

				ParentControl.Controls.Remove( this );
				FullScreenForm.Controls.Add( this );
				FullScreenForm.Show();
			}
			else
			{
				FullScreenForm.Controls.Remove( this );
				ParentControl.Controls.Add( this );
				FullScreenForm.Close();
				FullScreenForm = null;
			}
		}
		#endregion

		private void OnAddConfiguration(object sender, EventArgs e)
		{

		}

		private ContextMenuStrip settingsMenuStrip;
		private ToolStripMenuItem oneMenuItem;
		private ToolStripMenuItem twoByTwoMenuItem;
		private ToolStripMenuItem onePlusSevenMenuItem;
		private ToolStripMenuItem threeByThreeMenuItem;
		private ToolStripMenuItem twoPlusEightMenuItem;
		private ToolStripMenuItem onePlusTwelveMenuItem;
		private ToolStripMenuItem fourByFourMenuItem;
		private ToolStripMenuItem fiveByFiveMenuItem;
		private ToolStripMenuItem sixBySixMenuItem;
		private ToolStripMenuItem fullScreenMenuItem;
		private ToolStripMenuItem savedConfigurationsMenuItem;
		private ToolStripMenuItem addConfigurationMenuItem;
		private System.ComponentModel.IContainer components;

		#region private void InitializeComponent()
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.settingsMenuStrip = new System.Windows.Forms.ContextMenuStrip(this.components);
			this.oneMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.twoByTwoMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.onePlusSevenMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.threeByThreeMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.twoPlusEightMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.onePlusTwelveMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.fourByFourMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.fiveByFiveMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.sixBySixMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.fullScreenMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.savedConfigurationsMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.addConfigurationMenuItem = new System.Windows.Forms.ToolStripMenuItem();
			this.settingsMenuStrip.SuspendLayout();
			this.SuspendLayout();
			// 
			// settingsMenuStrip
			// 
			this.settingsMenuStrip.Items.AddRange(new System.Windows.Forms.ToolStripItem[] {
			this.fullScreenMenuItem,
			this.savedConfigurationsMenuItem,
			this.oneMenuItem,
			this.twoByTwoMenuItem,
			this.onePlusSevenMenuItem,
			this.threeByThreeMenuItem,
			this.twoPlusEightMenuItem,
			this.onePlusTwelveMenuItem,
			this.fourByFourMenuItem,
			this.fiveByFiveMenuItem,
			this.sixBySixMenuItem});
			this.settingsMenuStrip.Name = "settingsMenuStrip";
			this.settingsMenuStrip.Size = new System.Drawing.Size(188, 246);
			// 
			// oneMenuItem
			// 
			this.oneMenuItem.Name = "oneMenuItem";
			this.oneMenuItem.Size = new System.Drawing.Size(187, 22);
			this.oneMenuItem.Text = "1 Window";
			this.oneMenuItem.Click += new System.EventHandler(this.OnOneWindow);
			// 
			// twoByTwoMenuItem
			// 
			this.twoByTwoMenuItem.Name = "twoByTwoMenuItem";
			this.twoByTwoMenuItem.Size = new System.Drawing.Size(187, 22);
			this.twoByTwoMenuItem.Text = "4 Windows";
			this.twoByTwoMenuItem.Click += new System.EventHandler(this.OnTwoByTwo);
			// 
			// onePlusSevenMenuItem
			// 
			this.onePlusSevenMenuItem.Name = "onePlusSevenMenuItem";
			this.onePlusSevenMenuItem.Size = new System.Drawing.Size(187, 22);
			this.onePlusSevenMenuItem.Text = "8 Windows";
			this.onePlusSevenMenuItem.Click += new System.EventHandler(this.OnOnePlusSeven);
			// 
			// threeByThreeMenuItem
			// 
			this.threeByThreeMenuItem.Name = "threeByThreeMenuItem";
			this.threeByThreeMenuItem.Size = new System.Drawing.Size(187, 22);
			this.threeByThreeMenuItem.Text = "9 Windows";
			this.threeByThreeMenuItem.Click += new System.EventHandler(this.OnThreeByThree);
			// 
			// twoPlusEightMenuItem
			// 
			this.twoPlusEightMenuItem.Name = "twoPlusEightMenuItem";
			this.twoPlusEightMenuItem.Size = new System.Drawing.Size(187, 22);
			this.twoPlusEightMenuItem.Text = "10 Windows";
			this.twoPlusEightMenuItem.Click += new System.EventHandler(this.OnTwoPlusEight);
			// 
			// onePlusTwelveMenuItem
			// 
			this.onePlusTwelveMenuItem.Name = "onePlusTwelveMenuItem";
			this.onePlusTwelveMenuItem.Size = new System.Drawing.Size(187, 22);
			this.onePlusTwelveMenuItem.Text = "13 Windows";
			this.onePlusTwelveMenuItem.Click += new System.EventHandler(this.OnOnePlusTwelve);
			// 
			// fourByFourMenuItem
			// 
			this.fourByFourMenuItem.Name = "fourByFourMenuItem";
			this.fourByFourMenuItem.Size = new System.Drawing.Size(187, 22);
			this.fourByFourMenuItem.Text = "16 Windows";
			this.fourByFourMenuItem.Click += new System.EventHandler(this.OnFourByFour);
			// 
			// fiveByFiveMenuItem
			// 
			this.fiveByFiveMenuItem.Name = "fiveByFiveMenuItem";
			this.fiveByFiveMenuItem.Size = new System.Drawing.Size(187, 22);
			this.fiveByFiveMenuItem.Text = "25 Window";
			this.fiveByFiveMenuItem.Click += new System.EventHandler(this.OnFiveByFive);
			// 
			// sixBySixMenuItem
			// 
			this.sixBySixMenuItem.Name = "sixBySixMenuItem";
			this.sixBySixMenuItem.Size = new System.Drawing.Size(187, 22);
			this.sixBySixMenuItem.Text = "36 Windows";
			this.sixBySixMenuItem.Click += new System.EventHandler(this.OnSixBySix);
			// 
			// fullScreenMenuItem
			// 
			this.fullScreenMenuItem.Enabled = false;
			this.fullScreenMenuItem.Name = "fullScreenMenuItem";
			this.fullScreenMenuItem.Size = new System.Drawing.Size(187, 22);
			this.fullScreenMenuItem.Text = "Full Screen";
			this.fullScreenMenuItem.Click += new System.EventHandler(this.OnFullScreen);
			// 
			// savedConfigurationsMenuItem
			// 
			this.savedConfigurationsMenuItem.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
			this.addConfigurationMenuItem});
			this.savedConfigurationsMenuItem.Name = "savedConfigurationsMenuItem";
			this.savedConfigurationsMenuItem.Size = new System.Drawing.Size(187, 22);
			this.savedConfigurationsMenuItem.Text = "Saved Configurations";
			// 
			// addConfigurationMenuItem
			// 
			this.addConfigurationMenuItem.Enabled = false;
			this.addConfigurationMenuItem.Name = "addConfigurationMenuItem";
			this.addConfigurationMenuItem.Size = new System.Drawing.Size(173, 22);
			this.addConfigurationMenuItem.Text = "Add Configuration";
			this.addConfigurationMenuItem.Click += new System.EventHandler(this.OnAddConfiguration);
			// 
			// VideoView
			// 
			this.BackColor = System.Drawing.Color.Black;
			this.settingsMenuStrip.ResumeLayout(false);
			this.ResumeLayout(false);

		}
		#endregion
	}
}
